using namespace std;

#include "annotation.h"
#include "gmv.h"
#include "gbmisc.h"
#include "izfstream.h"

bool check_overwrap(int af, int at, int bf, int bt){
  // true if af-at and bg-bt overwrap or include
  if( (bf<=af && af<=bt) || // Afrom is inside B
      (bf<=at && at<=bt) || // Ato   is inside B
      (bf<=af && at<=bt) || // B includes A
      (af<=bf && bt<=at) ) // A includes B
    return true;
  else 
    return false;
}

int check_overwrap_int(int af, int at, int bf, int bt){
  // a is anchor or something smaller, b is gene or something longer
 
  if( bf<=af && at<=bt ) return abs(at-af)+1; // B includes A
  if( af<=bf && bt<=at ) return abs(bt-bf)+1; // A includes B
  
  if( bf<=af && af<=bt ) return abs(bt-af)+1; // Afrom is inside B
  if( bf<=at && at<=bt ) return abs(at-bf)+1; // Ato   is inside B

  return 0;
}

double check_overwrap_ratio(int af, int at, int bf, int bt){
  int len = abs(at-af)+1;
  return( (double)check_overwrap_int(af,at,bf,bt)/len );
}

// ----------------------------------------------------------------------
// annotation markings

rectangle::rectangle(double x1, double y1, double x2, double y2){
  if( x1 <= x2 ){
    p1.x = x1; p2.x = x2;}
  else {
    p1.x = x2; p2.x = x1;}
  
  if( y1 <= y2 ){
    p1.y = y1; p2.y = y2;}
  else {
    p1.y = y2; p2.y = y1;}
}

bool rectangle::includes(double x1, double y1) const{
  return ( p1.x <= x1 && x1 <= p2.x && p1.y <= y1 && y1 <= p2.y );
}

bool operator<(const xypoint a, const xypoint b){
  return (a.x + 100000*a.y < b.x + 100000*b.y);
}

bool operator<(const rectangle a, const rectangle b){
  return (a.p1<b.p1 && a.p2<b.p2);
}
// ----------------------------------------------------------------------
// class annotationIndex

annotation annotationIndex::getAnnotation() const {
  return (mr.rat(mr.annotations[seq].at(track)->at(no)));
}

// ----------------------------------------------------------------------
// class annotationTrack


void annotationTrack::readstream(IFSTREAM& anno, string streamname, int seqno, 
				 results& r,
				 void(*setfrac)(double)){
  // zero:     progbar position at start
  // workload: relative progbar value at end
  
  bound_seq = seqno;

  string file, extension;

  // extract multi fasta index (for GFF/GTF mapping)
  map<string,seqlen_t> mfmap;
  for(unsigned i=0; i<mr.sources_size(seqno); i++){
    seq_source src = mr.get_source(seqno, i);
    mfmap[src.organism] = src.begin;
  }

  // remove directory path 
  file = streamname;
  if( streamname.find_last_of("/") != string::npos )
    file = streamname.substr(streamname.find_last_of("/")+1);
  
  name = file;

  // extract extension
  extension = "";
  if( file.find_last_of(".") != string::npos )
    extension = file.substr(file.find_last_of(".")+1);

  if(extension == "gz" || extension == "bz2" ||
     extension == "GZ" || extension == "BZ2"){
    string tmpname = file.substr(0, file.find_last_of("."));
    extension = tmpname.substr(tmpname.find_last_of(".")+1);
  }

  // read annotation file
  anno.seekg(0, std::ios::end);
  double annolen = anno.tellg();
  anno.seekg(0, std::ios::beg);

  filterActive(false);
  init();

#define ANNO_GBK 0
#define ANNO_DAS 1
#define ANNO_GFF 2
#define ANNO_CDS 3

  int annotype = ( ( extension=="xml" || extension=="XML" ) ? ANNO_DAS :
		   ( extension=="gff" || extension=="GFF" ||
		     extension=="gtf" || extension=="GTF" ) ? ANNO_GFF :
		   ( extension=="cds" || extension=="CDS" ) ? ANNO_CDS :
		   ANNO_GBK );

  // ------------------------------------------------------------
  // GenBank
  if( annotype == ANNO_GBK ){

#define GENBANK_DETECT_NONE    0
#define GENBANK_DETECT_FEATURE 1
#define GENBANK_DETECT_GENE    2
#define GENBANK_DETECT_CDS     3
#define GENBANK_DETECT_SOURCE  4

    seqlen_t f, t;

    string organism;
    string chromosome;
    string plasmid;

    annotation anno_gene;
    annotation anno_feature;

    f=t=0; // initialize (safety net)

    string b; // read buffer

    // skip until "FEATURES" section appears
    while ( getline(anno, b) != NULL ){
      //      cerr << ">" << b  << "<\n";
      if(b.substr(0,8) == "FEATURES") break;
    }
    
    if(getline(anno, b) != NULL){
      do {
	string type_field = b.substr(0,21);

	// break if "ORIGIN"
	if(type_field.find("ORIGIN") != string::npos) break; 

	// ------------------------------
	// found something
	if(type_field.find_last_not_of(" ") != string::npos){

	  if(type_field.find("  source   ") != string::npos){
	    // source 
	    organism=""; chromosome=""; plasmid="";

	    pair<seqlen_t,seqlen_t> r = gb_detect_range(b);
	    f = r.first;
	    t = r.second;

	    while(getline(anno, b) != NULL){
	      if(b.substr(0,21).find_last_not_of(" ") != string::npos) break;
	      
	      size_t fr = b.find_first_of("\"")+1;
	      size_t to = b.find_last_of("\"");
	      size_t npos = string::npos;
	      
	      if(b.find("/plasmid=")!=npos)    plasmid = b.substr(fr, to-fr);
	      if(b.find("/organism=")!=npos)   organism = b.substr(fr, to-fr);
	      if(b.find("/chromosome")!=npos)  chromosome = b.substr(fr, to-fr);
	    }

	    mr.add_source(seqno, 
			  seq_source(f, t, organism, chromosome, plasmid));
	    continue;

	  } else if ( type_field[0] == ' ' /* skip unknown "SECTION NAME"s */){
	    bool feat_is_gene = false;
	    annotation* feat;
	    if (type_field.find("   gene    ") != string::npos){
	      annotation new_anno;
	      anno_gene = new_anno;
	      feat_is_gene = true;
	      feat = &anno_gene;
	    } else {
	      anno_feature = anno_gene;
	      feat = &anno_feature;
	    }
		      
	    // type
	    size_t atype_f = type_field.find_first_not_of(" ");
	    size_t atype_t = type_field.find_last_not_of(" ");
	    feat->atype = type_field.substr( atype_f, atype_t+1 - atype_f );

	    // range
	    pair<seqlen_t,seqlen_t> r = gb_detect_range(b);
	    feat->begin = r.first;
	    feat->end   = r.second;
	    
	    // reverse
	    feat->reverse = (b.find("complement") != string::npos);

	    // read feature entries
	    while(getline(anno, b) != NULL){
	      if(b.substr(0,21).find_last_not_of(" ") != string::npos) break;

	      size_t fr = b.find_first_of("\"")+1;
	      size_t to = b.find_last_of("\"");
	      size_t npos = string::npos;

	      // entry may spanning on multiple lines, 
	      // translation should contain no spaces
	      string tmpstr = b.substr(fr, to-fr);
	      string spacer = (b.find("/translation")==npos) ? " " : "";
	      if(to<fr){ // should read more line(s)
		string tmpstr2;
		while(getline(anno, tmpstr2) != NULL){
		  tmpstr = tmpstr + spacer + tmpstr2.substr(21);
		  if(tmpstr2.find("\"")!=npos) break;
		}
	      }
	
	      // remove quotation at the end of line
	      to=tmpstr.find_last_of("\""); 
	      tmpstr=tmpstr.substr(0, to);

	      if(b.find("/gene=")!=npos)       feat->name = tmpstr;
	      if(b.find("/product=")!=npos)    feat->product = tmpstr;
	      if(b.find("/db_xref")!=npos)     feat->dbref.insert(tmpstr);
	      if(b.find("/locus_tag")!=npos)   feat->locus = tmpstr;
	      if(b.find("/note")!=npos)        feat->note = tmpstr;
	      if(b.find("/translation")!=npos) feat->translation = tmpstr;
	      if(b.find("/EC_number")!=npos){  feat->ecnumber = tmpstr;
		feat->dbref.insert("EC:" + tmpstr);
	      }
	      if(b.find("/function")!=npos)    feat->function = tmpstr;
	    }

	    if(!feat_is_gene) push_back(*feat);
	    continue; // go next without reading the stream
	  }
	}

	if(setfrac!=NULL){
	  if(annolen!=-1) setfrac( anno.tellg()/annolen );
	  else setfrac(-1);
	}
	if(getline(anno, b) == NULL) break;
      } while( true );
    }

    isTraceable(true);
    hasScore(false);
  } 

  // ------------------------------------------------------------
  // GFF

  if (annotype == ANNO_GFF){
    string buf;
    map<string, int> groups;

    bool traceable=false;
    bool hasscore =true;
    while(getline(anno, buf)){
      stringstream ss;
      string buf2;
      string feature;
      string gname, gnote, id, gff_id;
      seqlen_t begin, end;
      double score;
      bool reverse=false;
      bool has_id = false;
      bool has_parent = false;
      string parent;

      size_t pos1, pos2;

      // skip comment lines
      if(buf.find("##FASTA") != string::npos) break;
      if(buf.substr(buf.find_first_not_of(" \t"), 1)=="#") continue;

      buf = decodecharcode(buf); // parse hex code like %20

      pos1 = 0;
      pos2 = buf.find_first_of("\t");
      if(pos2==string::npos) continue; // wrongly formatted line

      buf2 = buf.substr(pos1, pos2); // seqname
      buf  = buf.substr(buf.find_first_not_of("\t", pos2)); // trim

      seqlen_t gene_offset = 0;
      map<string,seqlen_t>::iterator offset_ite = mfmap.find(buf2);
      if(offset_ite != mfmap.end()) gene_offset = offset_ite->second;

      // source
      pos2 = buf.find_first_of("\t");
      buf2 = buf.substr(pos1, pos2); 
      buf  = buf.substr(buf.find_first_not_of("\t", pos2));

      // feature
      pos2 = buf.find_first_of("\t");
      feature = buf.substr(pos1, pos2); 
      buf  = buf.substr(buf.find_first_not_of("\t", pos2));

      gchar* fstr = g_ascii_strdown(feature.c_str(), feature.length());
      string feature_d=fstr;
      free(fstr);

      if(feature_d=="contig" || feature_d=="chromosome") continue;

      // begin & end
      ss << buf;
      ss >> begin;
      ss >> end;

      // score 
      ss >> buf2; score = atof(buf2.c_str()); // score: buf2 may "." 
      if(buf2==".") hasscore = false;

      // strand
      ss >> buf2; reverse = (buf2 == "-"); 
      ss >> buf2; // frame: just get it away

      getline(ss, buf);

      buf2 = buf.substr(buf.find_first_not_of(" \t"));// attr
      
      if(buf2.substr(0, 6)=="/gene="){ // GenBank type annotation in GFF
	pos1=buf2.find_first_of("\"")+1;
	pos2=buf2.find_first_of("\"", pos1);
	gname=buf2.substr(pos1, pos2-pos1);
	traceable = true;
      } else if(buf2.find_first_of("=")!= string::npos){ // try GFF3 parse
	pos1=0;
	pos2=0;
	while( (pos2=buf2.find_first_of("=", pos1))!= string::npos){
	  // first of all, skip extra semicolons
	  size_t pos_semicolon = buf2.find_first_of(";", pos1);
	  if(pos_semicolon != string::npos){  
	    if(pos_semicolon < pos2){
	      pos1 = pos_semicolon+1;
	      continue;
	    }
	  }

	  string key = buf2.substr(pos1, pos2-pos1);
	  pos1 = pos2+1;
	  pos2=buf2.find_first_of(";", pos1);
	  if(pos2 == string::npos) pos2=buf2.length()+1; // end of line
	  string value = buf2.substr(pos1, pos2-pos1);

	  gchar* keystr = g_ascii_strdown(key.c_str(), key.length());
	  key=keystr;
	  free(keystr);
	  
	  if(key=="name" || key=="id"){
	    if(key=="name"){ gname=gff_id=value; traceable=true; }
	    if(key=="id"){ id=gff_id=value; } // is that true?
	    has_id = true;
	  }

	  if(key=="note" || key=="description") gnote=value;

	  if(key=="parent"){
	    parent = value;
	    has_parent = true;
	  }
	  pos1 = pos2+1;
	}

      } else { // This may incorrect, but to read Ciona (Hoya) annotation
	if(buf2.substr(0, 6)=="name \""){
	  size_t pos1=buf2.find_first_of("\"")+1;
	  size_t pos2=buf2.find_first_of("\"", pos1);
	  gname=buf2.substr(pos1, pos2-pos1);
	  traceable = true;
	}
      }

      begin--; // adjust because gmv is 0 origin
      if(begin > end) swap(begin, end);

      bool register_this = (!has_id) && (!has_parent);
      if (has_id){
	//cout << "gff_id: " << gff_id << endl;
	if(has_parent && groups.find(parent) != groups.end()){
	  groups[gff_id] = groups[parent];
	  //cout << "register with parent: " << parent << " at " << groups[parent] << endl;
	} else {
	  groups[gff_id] = (int)anno_array.size();
	  register_this = true;
	  //cout << "register new at: " << groups[gff_id] << endl;
	}
      }

      if(register_this)
	push_back(annotation(begin + gene_offset, end + gene_offset,
			     reverse, feature, score, gname, gnote, id));

      if (has_parent){
	if(groups.find(parent) != groups.end())
	  anno_array[groups[parent]].segments.
	    push_back(annotation::anno_segment(begin + gene_offset, 
					       end + gene_offset, feature));
      }

      if(setfrac!=NULL) setfrac( anno.tellg()/annolen );
    }
    isTraceable(traceable);
    hasScore(hasscore);
  }

  // ----------------------------------------------------------------------
  // DAS
  if(annotype == ANNO_DAS){
    string b;
    string das_data;
    while( getline(anno, b)!=NULL ){ 
      das_data += b;
      if(setfrac!=NULL) setfrac( anno.tellg()/annolen );
    }

    das_featureParser fp;
    vector<das_feature> fvec;
    if(fp.parse(das_data, fvec)){ // parse may need progbar...
      map<string, int> groups;
      
      for(unsigned i=0; i<fvec.size(); i++){
	string group = fvec[i].group;

	if(group == "")
	  push_back(annotation(fvec[i]));
	else {
	  if(groups.find(group) == groups.end()){
	    groups[group] = (int)anno_array.size();
	    push_back(annotation(fvec[i]));
	  }

	  int group_index = groups[group];
	  seqlen_t b = fvec[i].start-1;
	  seqlen_t e = fvec[i].stop;

	  anno_array[group_index].segments.
	    push_back(annotation::anno_segment(b, e));
	  if(b < anno_array[group_index].begin) anno_array[group_index].begin=b;
	  if(anno_array[group_index].end < e) anno_array[group_index].end=e;
	}
      }

      isTraceable(true);
      hasScore(false); // FIXME!
    } // else parse error...
  }

  // ----------------------------------------------------------------------
  // CDS

  if (annotype == ANNO_CDS){
    string buf;

    while(getline(anno, buf) != NULL){
      size_t f = 0;

      string gname, locus;
      seqlen_t start=0, stop=0;
      bool reverse=false;

      for(unsigned field=0; field<5; field++){
	size_t t = buf.find_first_of("\t", f);
	if(field!=4 && t==string::npos) break; // broken line

	string s = buf.substr(f, t-f);
	switch(field){
	case 0: gname = s; break;
	case 1: start = atoll(s.c_str()); break;
	case 2: stop  = atoll(s.c_str()); break;
	case 3: reverse = (s=="-"); break;
	case 4: locus = s; 
	  if(stop<start) swap(start,stop);
	  push_back(annotation(start, stop, reverse, gname, locus));
	}
	
	f = buf.find_first_not_of("\t", t);
      }
      
    }
    isTraceable(true);
    hasScore(false);
  }


  // ----------------------------------------------------------------------
  // Finalize
  if(size() != 0){
    fixate();
    assert(cout	 << "read_annotations(): " 
	   <<  size() << " genes loaded for seq #" << seqno << "  / "
	   << mr.annotations[seqno].size() << endl);
  }

};

void annotationTrack::push_back(const annotation& a){ 
  segments.insert(a.begin, a.end, (int)anno_array.size());
  features.insert(a.atype);
  anno_array.push_back(a); 
};

void annotationTrack::fixate(){
  segments.build_tree();

  // list of features
  for(iterator ite=anno_array.begin(); ite!=anno_array.end(); ite++){
    for(vector<annotation::anno_segment>::const_iterator iite=ite->segments.begin();
	iite!=ite->segments.end(); iite++){
      if(iite->type != "") features.insert(iite->type);
    }
  }

  // dump the features (for debug)
//   for(set<string>::iterator ite=features.begin(); ite!=features.end(); ite++)
//     cout << ":" << *ite << endl;
};



// compute tf-idf based gene scores & create anchor-gene map
void annotationTrack::calcAnnotScores(int s, void(*setfrac)(double)){
  // s is geom#
  if(scored_by_anchors) return;

  bool score_loaded = !mr.tfidf.empty();

  unsigned i=0;
  for(vector<annotation>::iterator ite = anno_array.begin();
      ite != anno_array.end(); ite++){
    seqlen_t gf=ite->begin;
    seqlen_t gt=ite->end;
    // bool grev=ite->reverse;
    
    ite->tfidf_min = ite->tfidf_max = ite->tfidf_avg = 0;
    ite->num_ancs = 0;

    bool first = true;
    double mins, maxs;
    mins = maxs = 0;

    set<int> matches;
    mr.lookup_anchor_segments(s, gf, gt, matches);

#ifndef NDEBUG
    cout << "# of matches: " << matches.size() << endl;
    for(set<int>::iterator a=matches.begin(); a!=matches.end(); a++)
      cout << *a << endl;
#endif

    for(set<int>::iterator a=matches.begin(); a!=matches.end(); a++){
      double ascore = score_loaded ? mr.tfidf.at(*a) : 0;

      // no direction for this filter
      if(first){ 
	mins = maxs = ascore; 
	first = false;
      } else {
	if(ascore < mins) mins = ascore;
	if(maxs < ascore) maxs = ascore;
      }

      ite->tfidf_avg += ascore;
    }
    ite->num_ancs = matches.size();

    if(ite->num_ancs != 0) ite->tfidf_avg /= ite->num_ancs;
    ite->tfidf_min = mins;
    ite->tfidf_max = maxs;

    if(setfrac!=NULL) setfrac( i/anno_array.size() );
    i++;
  }
  scored_by_anchors = true;
}

bool annotationTrack::checkAnnotExpScores(int s){ // s is geom seq
  if(!hasScore()) return true;

  if(annot_exp_scores.empty()) return false;

  rvector<annotationTrack*>::iterator aite;
  unsigned size = 0;
  for(aite =mr.annotations[status.seq_lookup(s)].begin_all(); 
      aite!=mr.annotations[status.seq_lookup(s)].end_all(); aite++){
    if(annot_exp_scores.find(*aite) == annot_exp_scores.end())
      return false; // not found
    size++;
  }
  if(size == annot_exp_scores.size()) return true; else return false;
}

// compute expression based gene scores
void annotationTrack::calcAnnotExpScores(int s, void(*setfrac)(double)){ // s is geom seq#
  if(checkAnnotExpScores(s)) return;
  annot_exp_scores.clear();
  annot_exp_scores_indivisual.clear();

  rvector<annotationTrack*>::iterator tIte;
  unsigned tracks = mr.annotations[status.seq_lookup(s)].realsize();
  unsigned t = 0;
  for(tIte =mr.annotations[status.seq_lookup(s)].begin_all(); 
      tIte!=mr.annotations[status.seq_lookup(s)].end_all(); tIte++){

    annotationTrack* target = (annotationTrack*)*tIte;
    vector<double> scores;
    annot_exp_scores[target] = scores;
    if(target == this) continue; // no self score

    //    cout << this->get_name() << " based scores on " << target->get_name() << endl;

    annotationTrack::iterator aIte;
    int a=0;
      
    for(aIte = target->begin(); aIte != target->end(); aIte++){
      seqlen_t tf = aIte->begin;
      seqlen_t tt = aIte->end;
      bool trev = aIte->reverse;
      if(tf>tt) swap(tf, tt);

      vector<double> v;
      annot_exp_scores_indivisual[target].push_back(v);

      double sum_score = 0; 
      int overwraps = 0;

      // this filter checks direction
      set<int> matches;
      lookup_segment(tf, tt, matches);
      for(set<int>::iterator i=matches.begin(); i!=matches.end(); i++){
	if(anno_array[*i].reverse == trev){
	  sum_score += anno_array[*i].score;
	  annot_exp_scores_indivisual[target].at(a).push_back(anno_array[*i].score);
	  overwraps++;
	}
      }

      if(overwraps!=0)
	annot_exp_scores[target].push_back(sum_score/overwraps);
      else
	annot_exp_scores[target].push_back(0);
      
      if(setfrac!=NULL) setfrac((t + ((double)a/target->size()))/tracks);
      a++;
    }
    t++;
  }
}

// apply expression scores on anchors
void annotationTrack::calcAncScores(int s, void(*setfrac)(double)){ // s is geom seq#
  if(has_anc_scores) return;
  anc_scores.clear();
  anc_scores_indivisual.clear();
  
  //  cout << "calcAncScores(): go for " << get_name() << endl;

  for(unsigned r=0; r<mr.arows(); r++){
    seqlen_t af = abs(mr.raa(r, s*3));
    seqlen_t at = abs(mr.raa(r, s*3+1));
    int rev = (mr.raa(r, s*3+2)==0) ? true : false;
    if(at<af) swap(at, af);

    double sum_score = 0; 
    int overwraps = 0;
    vector<double> scores;

    // this filter checks direction
    set<int> matches;
    lookup_segment(af, at, matches);

    for(set<int>::iterator i=matches.begin(); i!=matches.end(); i++){
      if(anno_array[*i].reverse == rev){
	//	cout << "anc #" << r << " matched! " << anno_array[*i].score<< endl;

	sum_score += anno_array[*i].score;
	scores.push_back(anno_array[*i].score);
	overwraps++;
      }
    }

    anc_scores_indivisual.push_back(scores);
    if(overwraps!=0)
      anc_scores.push_back(sum_score/overwraps);
    else
      anc_scores.push_back(0);

    if(setfrac!=NULL) setfrac( r/mr.arows() );
  }
  has_anc_scores = true;
  // cout << "scored: " << anc_scores.size() << endl;
}

bool annotationTrack::checkAncFilter(int ancIndex) const {
  if(!filterAnchor()) return true;
  if(!has_anc_scores) return true; // filter isn't ready
  return (range_filter_min <= anc_scores.at(ancIndex) &&
	  anc_scores.at(ancIndex) <= range_filter_max );
}

bool annotationTrack::checkGenFilter(annotationTrack* t, int g) const {
  if(!filterGene()) return true;

  if(filter_policy == FILTER_POLICY_RANGE && 
     range_filter_policy == RANGE_FILTER_POLICY_AVG){
    map<annotationTrack*, vector<double> >::const_iterator ite;
    ite = annot_exp_scores.find(t);
    if(ite == annot_exp_scores.end()) return true; // filter isn't ready
    if(ite->second.empty()) return true; // because it should be myself

    double score = ite->second.at(g);
    if(range_filter_min<=score && score<=range_filter_max) return true;
    return false;
  }

  if(filter_policy == FILTER_POLICY_COUNT ||
     (filter_policy == FILTER_POLICY_RANGE && 
      range_filter_policy != RANGE_FILTER_POLICY_AVG) ){
    map<annotationTrack*, vector<vector<double> > >::const_iterator ite;
    ite = annot_exp_scores_indivisual.find(t);
    if(ite == annot_exp_scores_indivisual.end()) return true; //filter not ready
    if(ite->second.empty()) return true; // because it should be myself

    const vector<double>* scorevec = &(ite->second.at(g));
    
    double min=0, max=0;
    int morecount=0;
    int lesscount=0;
    for(vector<double>::const_iterator s=scorevec->begin(); 
	s!=scorevec->end(); s++){
      if(s==scorevec->begin()){ min = *s; max = *s; }
      if(*s < min) min = *s;
      if(max < *s) max = *s;
      if(*s >= count_filter_threshold) morecount++;
      if(*s <= count_filter_threshold) lesscount++;
    }

    if(filter_policy == FILTER_POLICY_RANGE)
      return (range_filter_min <= min && max <= range_filter_max);
    else { // count filter
      if(count_filter_policy == COUNT_FILTER_POLICY_COUNT){
	if(!count_filter_less){
	  if(!count_filter_lower)
	    return morecount >= count_filter_count;
	  else
	    return lesscount >= count_filter_count;
	} else {
	  if(!count_filter_lower)
	    return morecount <= count_filter_count;
	  else 
	    return lesscount <= count_filter_count;
	}
      } else { // COUNT_FILTER_POLICY_PERCENTAGE
	double moreratio = (double)morecount*100/scorevec->size();
	double lessratio = (double)morecount*100/scorevec->size();
	if(!count_filter_less){
	  if(!count_filter_lower)
	    return moreratio >= count_filter_percentage;
	  else
	    return lessratio >= count_filter_percentage;
	} else {
	  if(!count_filter_lower)
	    return moreratio <= count_filter_percentage;
	  else 
	    return lessratio <= count_filter_percentage;
	}
      } 
    }
  }

  // safety net
  return true;
}

// ----------------------------------------------------------------------
// class annotation

annotation::annotation(const das_feature &d){
  name = d.label;
  for(vector<string>::const_iterator i=d.url.begin(); i!=d.url.end(); i++)
    dbref.insert(*i);
  
  locus = d.id;
  atype = d.type;
  begin = abs(d.start);
  end = abs(d.stop);
  reverse = d.reverse;
  score = d.score;
  if(end>begin) swap(begin, end);
}


bool annotation::match(
#ifndef USE_GLIB_REGEX
		       regex_t* reg,
#else
		       GPatternSpec* reg, 
#endif
		       bool caseinSensitive, int strand,
		       bool name_e, bool product_e, bool locus_e, 
		       bool note_e, bool function_e){
  // strand 0: both, 1: fwd, 2: bwd

  bool found = false;

#ifndef USE_GLIB_REGEX
  regmatch_t match;
#endif

  if( (strand==1 && reverse) || (strand==2 && !reverse) ) return false;

  string target = ( ( name_e ? name : "" ) +
		    ( product_e ? product : "" ) +
		    ( locus_e ? locus : "" ) +
		    ( note_e ? note : "" ) +
		    ( function_e ? function : "" ) );
		    
  if (caseinSensitive) transform(target.begin(), target.end(), target.begin(),
				 (int (*)(int))std::toupper);

#ifndef USE_GLIB_REGEX
  found = ( regexec(reg, target.c_str(), 1, &match, 0) == 0 );
#else 
  found = g_pattern_match_string(reg, target.c_str());
#endif  

  return found;
}
