using namespace std;

#ifndef _ANNOTATION_H_
#define _ANNOTATION_H_

#include <assert.h>

#include <set>
#include <vector>
#include <string>
#include <map>
#include <iostream>

#include "segment_tree.h"
#include "dasclient.h"
#include "basestat.h"
#include "izfstream.h"

#ifndef USE_GLIB_REGEX
 #include <regex.h>
#endif

#ifdef USE_GLIB_REGEX
# include <glib.h>
#endif

#include "gmv_defs.h"

class annotationTrack;
typedef pair<const annotationTrack*, int> atIte;

// ----------------------------------------------------------------------
// DAS stuff

struct das_import_result {
  string server;
  string source;
  string segment;
  string features;
  bool separate_track;
  vector<das_feature> fvec;
};

// ----------------------------------------------------------------------
// annotation

class annotation{
 public:
  class anno_segment{
  public:
    seqlen_t begin, end;
    string type;
    anno_segment(seqlen_t b, seqlen_t e) : begin(b), end(e) {};
    anno_segment(seqlen_t b, seqlen_t e, string t) : 
      begin(b), end(e), type(t) {};
  };

  seqlen_t begin;
  seqlen_t end;
  bool reverse;
  string atype;
  string name;
  string locus;
  string product;
  string note;
  double score;
  string ecnumber;
  string function;
  string translation;
  set<string> dbref;
  vector<anno_segment> segments;

  double tfidf_min, tfidf_max, tfidf_avg;
  int num_ancs;

  annotation(){
    tfidf_min = tfidf_max = tfidf_avg = 0;
    num_ancs = 0;
    segments.clear();
  };

  annotation(const das_feature &d);

  // GenBank
  annotation(seqlen_t b, seqlen_t e, bool r, string t, string n, set<string> d, 
	     string l, string p, string nn, string f, string ec, string tr){
    annotation();
    begin=b; end=e; reverse=r; 
    atype=t; name=n; dbref=d; locus=l; product=p; note=nn;
    function=f; ecnumber=ec; translation=tr; };
  
  // GFF
  annotation(seqlen_t b, seqlen_t e, bool r, string f, double s, string n,
	     string nn, string id){
    annotation();
    atype="GFF";
    
    begin=b; end=e; reverse=r; atype=f; score=s; name=n; note=nn; locus=id;};


  // CDS
  annotation(seqlen_t b, seqlen_t e, bool r, string n, string l){
    annotation(); begin=b; end=e; reverse=r; name=n; locus=l; };

  string getname(int policy){
    switch(policy){
    case NAME_BY_GNAME: return name;
    case NAME_BY_LOCUS: return locus; 
    case NAME_BY_PRODUCT: return product;
    }
    return getname_auto();
  };

  bool match(
#ifndef USE_GLIB_REGEX
	     regex_t* reg,
#else
	     GPatternSpec* reg,
#endif
	     bool caseinSensitive, int strand,
	     bool name_e, bool product_e, bool locus_e, 
	     bool note_e, bool function_e);

  string getname_auto(){
    string n;
    n = name;
    if(n=="") n=locus;
    if(n=="") n=product;
    return n;
  };
};

class results;
class seq_source;
class annotationTrack {
  friend class results;

public:
  int bandRequest;
  
  typedef vector <annotation>::iterator iterator;
  set <string> features, hidden_features;
  
  ~annotationTrack(){ 
    assert(cout << "annotationTrack(): delete" << endl);
  };
  
  bool is_hidden_feature(const string& f) const {
    return hidden_features.find(f) != hidden_features.end();
  }

  void init(){ 
    hidden_features.insert("EST_match");

    assert(cout << "annotationTrack(): init" << endl);
    avg_score = sd_score = max_score = min_score = 0; 
    isTraceable(true); hasScore(false); bandRequest=1;
    has_statistics = false; 
    has_anc_scores = false;
    anc_scores.clear();
    anc_scores_indivisual.clear();
    scored_by_anchors = false;
    annot_exp_scores.clear();
    annot_exp_scores_indivisual.clear();

    filter_active = false;
    filter_policy = FILTER_POLICY_RANGE;
    range_filter_policy = RANGE_FILTER_POLICY_AVG;
    count_filter_policy = COUNT_FILTER_POLICY_COUNT;
    count_filter_count = 0;
    count_filter_percentage = 0;
    count_filter_less = false;
    count_filter_lower = false;

    filter_anchor = true;
    filter_gene = true;

    show_name = false;
    emphasize = false;
    plot_style = 0;
  };

  void fixate();

  annotationTrack(){ init(); };
  annotationTrack(string n){ init(); name=n; };

//   annotation& at(unsigned i){ return anno_array.at(i); };
//   annotation& operator[](int i) { return anno_array[i]; };
  atIte at(unsigned i)    const { return atIte(this, i);};
  atIte operator[](int i) const { return atIte(this, i);};

  void readstream(IFSTREAM&, string, int, 
		  results&, void(*)(double));

  void push_back(const annotation& a);

  unsigned int size() const { return anno_array.size(); };

  void lookup_segment(seqlen_t a, seqlen_t b, set<int> &res){ segments.query(a, b, res); };

  size_t expScores_size(annotationTrack* t, int a) const {
    return annot_exp_scores_indivisual.find(t)->second.at(a).size();
  };
  vector<double>::const_iterator expScores_begin(annotationTrack* t, int a) const { 
    return annot_exp_scores_indivisual.find(t)->second.at(a).begin(); };
  vector<double>::const_iterator expScores_end(annotationTrack* t, int a) const { 
    return annot_exp_scores_indivisual.find(t)->second.at(a).end(); };

  bool isTraceable(bool t) { traceable = t; return t; };
  bool isTraceable() const { return traceable; };
  bool isFilterable() const { return isTraceable(); }; // can be filtered = true
  bool hasScore(bool t){has_score = t; return t;}; // have its own score or not
  bool hasScore() const{ return has_score; };

  double maxScore(){ if(!has_statistics) calcStat(); return max_score; };
  double minScore(){ if(!has_statistics) calcStat(); return min_score; };

  double avgScore(){ if(!has_statistics) calcStat(); return avg_score; };
  double sdScore(){  if(!has_statistics) calcStat(); return sd_score; };

  string get_name(){ return name; };

  bool hasAncScores(){ return has_anc_scores; };
  void calcAncScores(int s, void(*)(double)); // s is geom seq#

  bool hasAnnotScores(){ return scored_by_anchors; }; // annotation tfidf score 
  void calcAnnotScores(int s, void(*)(double)); // s is geom seq#

  bool checkAnnotExpScores(int s); // expression score is up to date or not
  void calcAnnotExpScores(int s, void(*)(double)); // s is geom seq#

  bool checkAncFilter(int ancIndex) const;
  bool checkGenFilter(annotationTrack* t, int g) const;

  size_t ancScoreIndivisual_size(int a) const { 
    return anc_scores_indivisual.at(a).size(); };
  vector<double>::const_iterator ancScoreIndivisual_begin(int a) const { 
    return anc_scores_indivisual.at(a).begin(); };
  vector<double>::const_iterator ancScoreIndivisual_end(int a) const { 
    return anc_scores_indivisual.at(a).end(); };

  double ancScore(int a) const { return anc_scores.at(a); };
  double genScore(annotationTrack *t, int a) const { 
    if(t==this) return anno_array.at(a).score;
    
    map<annotationTrack*, vector<double> >::const_iterator ite;
    ite = annot_exp_scores.find(t);
    if(ite == annot_exp_scores.end()) return 0; // filter isn't ready
    if(ite->second.empty()) return 0; // because it should be myself
    return(ite->second.at(a));
  };

  bool hasGenScoreForTrack(annotationTrack *t){
    map<annotationTrack*, vector<double> >::const_iterator ite;
    ite = annot_exp_scores.find(t);
    return( ite != annot_exp_scores.end() ); 
  };

  bool filterAnchor(bool a) { filter_anchor=a; return a; };
  bool filterAnchor() const { return filter_anchor; };
  bool filterGene(bool a) { filter_gene=a; return a; };
  bool filterGene() const { return filter_gene; };

  bool filterActive(bool a) { filter_active=a; return a; };
  bool filterActive() const { return filter_active; };
  int filterPolicy(int i){ filter_policy=i; return i; };
  int filterPolicy() const{ return filter_policy; };

  bool showName(bool s) { show_name=s; return s; };
  bool showName() const { return show_name; };
  bool emphasizePlot(bool e) { emphasize=e; return e; };
  bool emphasizePlot() const { return emphasize; };
  int plotStyle(int s)  { plot_style=s; return s; };
  int plotStyle() const { return plot_style; };

  int rangeFilterPolicy(int i){ range_filter_policy=i; return i; };
  int rangeFilterPolicy() const{ return range_filter_policy; };
  double filterMax(double a){ range_filter_max = a; return a; };
  double filterMax() const  { return range_filter_max; };
  double filterMin(double a){ range_filter_min = a; return a; };
  double filterMin() const  { return range_filter_min; };

  bool countFilterLower(bool b){ count_filter_lower=b; return b; };
  bool countFilterLower() const{ return count_filter_lower; };
  double countFilterThreshold(double d){ count_filter_threshold=d; return d;};
  double countFilterThreshold() const{ return count_filter_threshold; };
  bool countFilterLess(bool b){ count_filter_less=b; return b; };
  bool countFilterLess() const{ return count_filter_less; };
  int countFilterPolicy(int i){ count_filter_policy=i; return i; };
  int countFilterPolicy() const{ return count_filter_policy; };
  int countFilterCount(int i){ count_filter_count=i; return i; };
  int countFilterCount() const{ return count_filter_count; };
  double countFilterPercentage(double d){ count_filter_percentage=d; return d;};
  double countFilterPercentage() const{ return count_filter_percentage; };
  int bound_to() const{ return bound_seq; };

 private:
  int bound_seq;

  bool scored_by_anchors;
  bool has_statistics;
  bool traceable;
  bool has_score;
  string name;
  double min_score, max_score, avg_score, sd_score;

  bool   filter_anchor;
  bool   filter_gene;

  int    filter_policy;
  bool   filter_active;

  double range_filter_min, range_filter_max;
  int    range_filter_policy; // avg, min+max

  double count_filter_threshold;
  int    count_filter_policy;  // count, percentage
  int    count_filter_count;
  double count_filter_percentage;
  bool   count_filter_less;   // less than minimum count
  bool   count_filter_lower;  // lower than threshold

  bool show_name;
  bool emphasize;
  int  plot_style;  // 0: default, 1: block style

  iterator begin(){ return anno_array.begin(); };
  iterator end(){ return anno_array.end(); };

  vector <annotation> anno_array;
  segment_tree<int> segments;

  bool has_anc_scores;
  vector <double> anc_scores; // average exp scores of anchors
  vector<vector<double> > anc_scores_indivisual;

  map<annotationTrack*, vector<double> > annot_exp_scores;
  map<annotationTrack*, vector< vector<double> > > annot_exp_scores_indivisual;

  void calcStat(){
    avg_score = sd_score = 0; 
    for(unsigned i=0; i<anno_array.size(); i++){
      double score = anno_array[i].score;
      avg_score+=score;
      if(i==0){
	min_score = score;
	max_score = score;
      } else {
	if (min_score > score) min_score = score;
	if (max_score < score) max_score = score;
      }
    }
    if(min_score == max_score){ // they mustn't be same even if all scores are
      min_score -= 1;
      max_score += 1;
    }
    avg_score /= anno_array.size();

    for(unsigned i=0; i<anno_array.size(); i++){
      double d = anno_array[i].score-avg_score;
      sd_score += d*d;
    }
    sd_score /= anno_array.size();
    sd_score = sqrt(sd_score);

    has_statistics = true;

    filterMin(min_score);
    filterMax(max_score);

    countFilterThreshold(min_score);
  };
};

class annotationIndex{
public:
  int seq;  // should be native seq#
  int track;
  int no;
  annotationIndex(){ seq=0; track=0; no=0; };
  annotationIndex(int s, int t, int n){ seq=s; track=t; no=n; };
  annotationIndex(const annotationIndex &ai){ seq=ai.seq; track=ai.track; no=ai.no;}
  annotation getAnnotation() const;

  friend bool operator<(const annotationIndex a, const annotationIndex b){
    if ( a.seq > b.seq ) return false;
    if ( a.seq < b.seq ) return true;

    // now a.seq == b.seq
    if ( a.track > b.track ) return false;
    if ( a.track < b.track ) return true;

    //
    if ( a.no >= b.no ) return false;
    return true;
  };

  friend bool operator==(const annotationIndex a, const annotationIndex b){
    return (a.seq == b.seq && a.track == b.track && a.no == b.no );
  };
};

// ----------------------------------------------------------------------
// sequence source

class seq_source{
public:
  seqlen_t begin, end, last_reference;
  string organism;
  string chromosome;
  string plasmid;

  seq_source(seqlen_t b, seqlen_t e, string org, string chr, string pla){
    begin=b; end=e;
    organism=org; chromosome=chr; plasmid=pla;
    last_reference = 0;
  };

  string getstr() const {
    string s;
    s = organism +
      string(string(organism!="" && (chromosome+plasmid!="") ? " / " : "") +
	     string((chromosome!="") ? string("Chr "+chromosome) : "") +
	     string(plasmid!="" && (organism+chromosome!="") ? " / " : "") +
	     string((plasmid!="") ? string("Plasmid "+plasmid) : "")
	     );
    return s;
  };

  string getstr_short() const {
    if(chromosome!="") return chromosome;
    return plasmid;
  }

  string getstr_domestic() const {
    string s;
    s = 
      string(string((chromosome!="") ? string("Chr "+chromosome) : "") +
	     string(plasmid!="" && chromosome!="" ? " / " : "") +
	     string((plasmid!="") ? string("Plasmid "+plasmid) : "")
	     );
    return s;
  }
};

// ----------------------------------------------------------------------
// annotation markings

class xypoint{
 public:
  double x, y;
  xypoint(int a, int b){ x=a, y=b; };
  xypoint(){ x=0; y=0; };
  xypoint(const xypoint &p){ x=p.x; y=p.y; };
  friend bool operator<(const xypoint a, const xypoint b);
};

class rectangle{
 public:
  xypoint p1, p2;
  bool includes(double x1, double y1) const;
  rectangle(){ p1=xypoint(0,0); p2=xypoint(0,0); };
  rectangle(double x1, double y1, double x2, double y2);
  rectangle(const rectangle &r){ p1=r.p1; p2=r.p2; };
  friend bool operator<(const rectangle a, const rectangle b);
};

class geneMarkerRecord{
public:
  rectangle rect;
  annotationIndex ai;
  
  geneMarkerRecord(rectangle r, annotationIndex a){ rect=r; ai=a;};
  friend bool operator<(const geneMarkerRecord a, const geneMarkerRecord b){
    return (a.ai < b.ai);
  }
};

#endif
