using namespace std;

#ifndef _DAS_CLIENT_H_
#define _DAS_CLIENT_H_

#include <glib.h>
#include "gmv_defs.h"

string url_encode(string);
bool get_by_http(string* result, const char* host, int port, const char* path);
bool parseURL(const string URL, string* host, int *port, string* path);

// ----------------------------------------------------------------------
// abstracted XML parser

void xml_parser_start_element(GMarkupParseContext* context,
			      const gchar* elem, 
			      const gchar** attr_names, 
			      const gchar** attr_vals,
			      gpointer parser, GError** gerr);

void xml_parser_end_element(GMarkupParseContext* context,
			    const gchar* elem, 
			    gpointer parser, GError** gerr);

void xml_parser_text(GMarkupParseContext* context, 
		     const gchar* text, 
		     gsize text_len, 
		     gpointer parser, GError** gerr);

class xmlParser{
public:
  virtual void start_element(GMarkupParseContext*,
			     const gchar*, const gchar**, const gchar**,
			     GError**) = 0;
  virtual void end_element(GMarkupParseContext*,
			   const gchar*, GError**) = 0;
  virtual void text(GMarkupParseContext*, const gchar*, 
		    gsize, GError**) = 0;
  virtual ~xmlParser() = 0;
  xmlParser();

protected:
  GMarkupParser p;
  GMarkupParseContext* pc;
};

// ----------------------------------------------------------------------
// DSN

class das_dsn{
public:
  string source;
  string desc;
  string mapmaster;
};

class das_dsnParser : public xmlParser{
public:
  map<string,das_dsn>* dsnmap;
  das_dsn current_dsn;
  string current_id;

  void start_element(GMarkupParseContext* context, const gchar* elem, 
		     const gchar** attr_names, const gchar** attr_vals,
		     GError** gerr);

  void end_element(GMarkupParseContext* context, const gchar* elem, 
		   GError** gerr);

  void text(GMarkupParseContext* context, const gchar* text, 
	    gsize text_len, GError** gerr);

  bool parse(const string& dsn_string, map<string,das_dsn>& ds);
};

// ----------------------------------------------------------------------
// entry_points parser

class das_entrypoint{
public:
  unsigned start;
  unsigned stop;
  bool reverse; // true if orientation is '-'
  bool subparts;
  string segment;
};


class das_entrypointParser : public xmlParser{
public:
  map<string,das_entrypoint>* epmap;
  das_entrypoint current_ep;
  string current_id;

  void start_element(GMarkupParseContext* context, const gchar* elem, 
		     const gchar** attr_names, const gchar** attr_vals,
		     GError** gerr);
  void end_element(GMarkupParseContext* context, const gchar* elem, 
		   GError** gerr);
  void text(GMarkupParseContext* context, const gchar* text, 
	    gsize text_len, GError** gerr);
  bool parse(const string& dsn_string, map<string,das_entrypoint> &m);
};


// ----------------------------------------------------------------------
// types parser

class das_type{
public:
  string category;
  string type;
};

class das_typeParser : public xmlParser{
public:
  map<string,das_type>* dtmap;
  das_type current_dt;
  string current_id;

  void start_element(GMarkupParseContext* context, const gchar* elem, 
		     const gchar** attr_names, const gchar** attr_vals,
		     GError** gerr);
  void end_element(GMarkupParseContext* context, const gchar* elem, 
		   GError** gerr);
  void text(GMarkupParseContext* context, const gchar* text, 
	    gsize text_len, GError** gerr);
  bool parse(const string& dsn_string, map<string,das_type> &m);
};

// ----------------------------------------------------------------------
// features parser

class das_feature{
public:               // map to gmv's "annotation" class
  string label;       // name
  vector<string> url; // dbref

  // required ones
  string id;          // locus
  string type;        // atype
  seqlen_t start;     // begin
  seqlen_t stop;      // end
  bool reverse;       // reverse
  double score;       // score
  
  // optional stuff
  string group;

  void clear(){ // clear only optional entries
    label=""; 
    group="";
    url.clear();
  };
};

class das_featureParser : public xmlParser{
public:
  vector<das_feature>* fvec;
  das_feature current_f;
  string current_id;

  void start_element(GMarkupParseContext* context, const gchar* elem, 
		     const gchar** attr_names, const gchar** attr_vals,
		     GError** gerr);
  void end_element(GMarkupParseContext* context, const gchar* elem, 
		   GError** gerr);
  void text(GMarkupParseContext* context, const gchar* text, 
	    gsize text_len, GError** gerr);
  bool parse(const string& dsn_string, vector<das_feature> &m);
};


// ----------------------------------------------------------------------
// parser sample 

/*
class basicParser : public xmlParser{
public:

  void start_element(GMarkupParseContext* context, const gchar* elem, 
		     const gchar** attr_names, const gchar** attr_vals,
		     GError** gerr){

    cout << "elem: " << elem << endl;

    gchar** an = (gchar**)attr_names;
    gchar** av = (gchar**)attr_vals;
    while(*an!=NULL){
      cout << " attr: " << *an << " = " << *av << endl;
      an++; av++;
    }
  };

  void end_element(GMarkupParseContext* context, const gchar* elem, 
		   GError** gerr){

    cout << "/elem: " << elem << endl;
  };


  void text(GMarkupParseContext* context, const gchar* text, 
	    gsize text_len, GError** gerr){
    string t = text;
    if(t.find_first_not_of(" \n\r\t") == string::npos) return;
    cout << "text: \"" << t << "\"" << endl;
  };

  bool parse(const string& dsn_string){
    GError *gerr = NULL;
    g_markup_parse_context_parse(pc, dsn_string.c_str(),
				 dsn_string.size(), &gerr);
    if(gerr != NULL) return false;
    return true;
  };
};
*/

#endif
