using namespace std;

#include <iomanip>
#include <iostream>
#include <sstream>
#include <gtk/gtk.h>

#include "basestat.h"
#include "gmv_base.h"
#include "segment_tree.h"
#include "util.h"

#include "sys/types.h"
#include "sys/stat.h"

#include "gmv.h"
#ifdef __MINGW32__
#include "io.h"
#endif

#define STITCH_GAP 10 // gap inserted by chromstitch

// ----------------------------------------------------------------------
//  progress control by status

void pbar_set_fraction(double fraction){
  status.prog->fraction(fraction);
}

void pbar_set_fraction(int thread, double fraction){
  status.prog->fraction(thread, fraction);
}

void pbar_set_task(double b, double e){
  status.prog->set_task(b, e);
}

void pbar_reset(string s){
  status.prog->reset(s);
  pbar_set_task(0, 1);
}

void pbar_set_threads(unsigned t){
  status.prog->set_threads(t);
}

// ----------------------------------------------------------------------
//  progress

progress::progress(){
  assert(cout << "progress::progress() called" << endl);
  m = g_mutex_new();
  reset();
}

void progress::reset(){
  reset("");
}

void progress::reset(string s){
  g_mutex_lock(m);
  j = s;   // current job
  f = 0;   // current fraction
  job_changed = true;
  fvec.clear();
  g_mutex_unlock(m);
}

void progress::set_threads(unsigned t){
  g_mutex_lock(m);
  fvec.clear();
  for(unsigned i=0; i<t; i++) fvec.push_back(0);
  g_mutex_unlock(m);
}

double progress::fraction(){
  double d;

  g_mutex_lock(m);
  d = f;
  for(vector<double>::iterator i=fvec.begin(); i!=fvec.end(); i++) d += *i;
  g_mutex_unlock(m);

  return d;
}

double progress::fraction(int thread, double d){
  if(g_mutex_trylock(m)){
    fvec[thread]=d;
    g_mutex_unlock(m);
  }
  return d;
}

double progress::fraction(double d){
  double dd = d*(e-b) + b;
  if(g_mutex_trylock(m)){
    f = dd;
    g_mutex_unlock(m);
  }
  return dd;
}

string progress::jobtitle(){
  string t="";
  g_mutex_lock(m);
  if(job_changed) t=j;
  job_changed=false;      
  g_mutex_unlock(m);

  return t;
}

// ----------------------------------------------------------------------
// score 

void score::clear(){
  lowest = 0;
  highest = 0;
  average = 0;
  score_array.clear();
}

void score::push_back(const double s){
  score_array.push_back(s);
  if(score_array.size() == 1){ lowest=s; highest=s; average=s; } 
  else { 
    if(lowest>s) lowest=s;
    if(highest<s) highest=s;
    average += s;
  }
}

void score::setval(const size_t i, const double s){
  average -= s;
  score_array[i]=s;
  if(lowest>s) lowest=s;
  if(highest<s) highest=s;
  average += s;
}

void score::fixate(int histomode){
  average = average/score_array.size();
  histo.init(histomode, highest, lowest, score_array);
}

// ----------------------------------------------------------------------
//  results

bool results::set_tmurasakix(const string& basedir, const string& ntseqs){
  is_tmurasakix = true; 
  ifstream ntifs;
  ntifs.open(ntseqs.c_str());
  string fastafile;
  cout << "NTSEQS: " << ntseqs << endl;
  while(getline(ntifs, fastafile)){
    cout << "FASTAFILE: " << fastafile << endl;

    ifstream fasta;
    string buf;
    fasta.open(fastafile.c_str());
    if(!fasta) // second chance
      fasta.open(string(basedir+fastafile).c_str());
    if(!fasta && status.use_x11){ // third chance: interactive
      GtkWidget *dialog;
      dialog = gtk_message_dialog_new(GTK_WINDOW(status.mainwindow),
				      GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_MESSAGE_ERROR,
				      GTK_BUTTONS_YES_NO,
				      "Error loading sequence file: %s\nSubstitute sequence file (or open another anchor file) ?\n\
Note: you're about to load tmurasakix result. If not, please check your files.",
				      fastafile.c_str());
      gint r = gtk_dialog_run(GTK_DIALOG(dialog));
      gtk_widget_destroy(dialog);

      if(r==GTK_RESPONSE_YES){
	size_t pos = fastafile.find_last_of("/", string::npos);
	if(pos==string::npos) pos=0; else pos++;
	string fastafile2 = fastafile.substr(pos);

	dialog = gtk_file_chooser_dialog_new ("Substitute Sequence File",
					      GTK_WINDOW(status.mainwindow),
					      GTK_FILE_CHOOSER_ACTION_OPEN,
					      GTK_STOCK_CANCEL, 
					      GTK_RESPONSE_CANCEL,
					      GTK_STOCK_OPEN,
					      GTK_RESPONSE_ACCEPT,
					      (gpointer)NULL);
  
	GtkFileFilter *filter = gtk_file_filter_new();
	gtk_file_filter_add_pattern(filter, fastafile2.c_str());
	gtk_file_chooser_set_filter(GTK_FILE_CHOOSER(dialog), filter);

	bool file_selected =
	  (gtk_dialog_run(GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT);

	char *filename;
	filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
	gtk_widget_destroy(dialog);
	if(!file_selected) return false; // no substitution
	
	// windows compatibility hack
	char *fn = filename;
	while(*fn != 0){
	  if(*fn=='\\') *fn='/';
	  fn++;
	}
	// end windows compatibility hack
	
	fastafile2 = string(filename);
	g_free(filename);
	fasta.open(fastafile2.c_str());
      } else return false; // choose another anchor file
    }
    if(!fasta){ // failed on third chance!
      cerr << "Couldn't open sequence file: " << fastafile << endl;
      return false;  // choose another anchor file
    }
    getline(fasta, buf);
    ntlen.push_back(atoll(buf.substr(1).c_str()));
    fasta.close();
  }
  ntifs.close();
  return true;
}

results::results(){
  for(int i=0; i<MAXSEQ; i++) bs[i] = NULL;
  order_seq = NULL;
}

seq_source results::get_source(int seq, unsigned i) const {
  int s = seq;
  seq_source src = sources[s].at(i);
  seqlen_t offset = get_segment_offset(s, src.begin, src.end);
  src.begin += offset;
  src.end   += offset;
  return src;
}

void results::rebuild_segment_offset_index(){
  int seq = 0;
  basestat *b;
  while((b = getbs_direct(seq)) != NULL){
    if(segment_offsets[seq].empty()){ // default is no offset
      segment_offsets[seq].push_back(0);
      segment_offsets[seq].push_back(b->totalbases());
      segment_offsets[seq].push_back(0);
    }
    
    for(size_t i=0; i<segment_offsets[seq].size(); i+=3){
      segment_offset_index[seq].insert(segment_offsets[seq].at(i),
				       segment_offsets[seq].at(i+1),
				       segment_offsets[seq].at(i+2));
    }
    segment_offset_index[seq].build_tree(NULL);
    seq++;
  }
}

seqlen_t results::get_segment_offset(int seq, seqlen_t pos) const {
  set<seqlen_t> so;
  segment_offset_index[seq].query(pos, so);
  if(so.empty()) return 0;
  return *so.begin();
}

seqlen_t results::get_segment_offset(int seq, seqlen_t pos1, seqlen_t pos2) const { // this one should be more robust, but will be slower
  set<seqlen_t> so1, so2;
  vector<seqlen_t> so;
  vector<seqlen_t>::iterator so_ite;
  seqlen_t ret = 0;

  segment_offset_index[seq].query(pos1, so1);
  segment_offset_index[seq].query(pos2, so2);
  if(! (so1.empty() && so2.empty())){
    so.reserve(so1.size() + so2.size());
    
    so_ite = set_intersection(so1.begin(), so1.end(), 
			      so2.begin(), so2.end(), so.begin());
    if (so_ite != so.begin()) ret = so[0];
    else if (!so1.empty()) ret = *so1.begin();
    else if (!so2.empty()) ret =  *so2.begin();
  }
  return ret;
}

void results::sort_sequence_sources(int along, vector<int>& sort_seqs,
				    int method, bool keep_target_order,
				    bool apply_anchor_filter,
				    void(*prog)(double)){

  for(unsigned i=0; i<sort_seqs.size(); i++){
    int seq = sort_seqs.at(i);
    multimap<seqlen_t, unsigned> source_map; // averaged src position and src #

    // sort sources 
    // (averaged based policy: suitable for unsymmetric cases such as 
    //  assembled chroms vs contigs)
    for(unsigned j=0; j<sources_size(seq); j++){
      multimap<seqlen_t, int> anclen_map; // list of matched anchors' length
      seq_source src = sources[seq].at(j); // get_source returns src with offset
      set<int> ancs;
      anchor_segments[seq].query(src.begin, src.end, ancs);

      double mean = 0;
      double anctotal = 0;
      int anc_count = 0;
      for(set<int>::iterator k=ancs.begin(); k!=ancs.end(); k++){
	if(apply_anchor_filter && !mr.anchorFilter(*k)) continue;
	seqlen_t anclen =
	  keep_target_order ? abs(raa_physical(*k, along*3+1)-
				  raa_physical(*k, along*3)) :
	  abs(anchors[*k].at(along*3+1) - anchors[*k].at(along*3));

	switch(method){
	case SORT_SOURCES_BY_WEIGHTED_AVERAGE:
	case SORT_SOURCES_BY_AVERAGE:
	  mean += 
	  (method == SORT_SOURCES_BY_WEIGHTED_AVERAGE ? anclen : 1) *
	    ( keep_target_order ? (raa_physical(*k, along*3)+
				   raa_physical(*k, along*3+1))/2 :
	      (anchors[*k].at(along*3) + anchors[*k].at(along*3+1))/2 );
	  anctotal += anclen;
	  anc_count++;
	  break;

	case SORT_SOURCES_BY_MEDIAN:
	case SORT_SOURCES_BY_LONGEST_ANCHOR:
	  anclen_map.insert(pair<seqlen_t, int>(anclen, *k));
	  break;
	}
      }

      multimap<seqlen_t, int>::iterator anclen_ite  = anclen_map.begin();
      multimap<seqlen_t, int>::iterator anclen_ite2 = anclen_map.end();
      int ta; // target anchor for median/longest
      switch(method){
      case SORT_SOURCES_BY_WEIGHTED_AVERAGE:
      case SORT_SOURCES_BY_AVERAGE:
	if(anc_count != 0) 
	  mean /= ( method == SORT_SOURCES_BY_WEIGHTED_AVERAGE ? anctotal 
		    : ancs.size());
	break;

      case SORT_SOURCES_BY_MEDIAN:
      case SORT_SOURCES_BY_LONGEST_ANCHOR:
	if(anclen_ite == anclen_ite2){ mean = 0; break; }
	if(method==SORT_SOURCES_BY_MEDIAN){
	  bool itesel = false;
	  do {
	    if(itesel) anclen_ite++; else anclen_ite2--;
	    itesel = (!itesel);
	  } while (anclen_ite != anclen_ite2);
	} else { // SORT_SOURCES_BY_LONGEST_ANCHOR
	  anclen_ite2--; 
	  if(anclen_ite != anclen_ite2) anclen_ite = anclen_ite2;
	}
	ta = anclen_ite2->second;
	mean = (anclen_ite == anclen_ite2) ?
	  ( keep_target_order ? (raa_physical(ta, along*3)+
				 raa_physical(ta, along*3+1))/2: 
	    (anchors[ta].at(along*3)+anchors[ta].at(along*3+1))/2 )
	  : 0;
	break;
      }
      source_map.insert(pair<seqlen_t, unsigned>((seqlen_t)mean, j));
    }

    // compute source offsets
    if(!( seq == along && keep_target_order ))
      relocate_segment_offsets(seq, source_map);

  } // for (i=0; i<sort_seqs.size(); i++)
  rebuild_segment_offset_index();
}


void results::sort_sequence_sources(map<int, multimap<seqlen_t, unsigned> >&
				    srcmap){

  for(map<int, multimap<seqlen_t, unsigned> >::iterator i=srcmap.begin();
      i!=srcmap.end(); i++){
    relocate_segment_offsets(i->first, i->second);
  }
  rebuild_segment_offset_index();
}


void results::relocate_segment_offsets(int seq, 
				       multimap<seqlen_t, unsigned>& srcmap){

  seqlen_t source_gap = // assume it's constant
    sources[seq].at(1).begin - sources[seq].at(0).end;

  segment_offsets[seq].clear();
  seqlen_t startpoint = 0;
  for(multimap<seqlen_t,unsigned>::iterator srcite=srcmap.begin();
      srcite!=srcmap.end(); srcite++){
    if(srcite->first == 0) continue;
    seq_source* src = &sources[seq].at(srcite->second);
    segment_offsets[seq].push_back(src->begin);
    segment_offsets[seq].push_back(src->end);
    segment_offsets[seq].push_back(startpoint - src->begin);
    src->last_reference = srcite->first;

    startpoint += src->end - src->begin + source_gap;
  }

  // non-positioned (or positioned at 0) segments are placed at end
  for(multimap<seqlen_t,unsigned>::iterator srcite=srcmap.lower_bound(0);
      srcite!=srcmap.upper_bound(0); srcite++){
    seq_source* src = &sources[seq].at(srcite->second);
    segment_offsets[seq].push_back(src->begin);
    segment_offsets[seq].push_back(src->end);
    segment_offsets[seq].push_back(startpoint - src->begin);
    src->last_reference = srcite->first;

    startpoint += src->end - src->begin + source_gap;
  }
}

void results::clear(){
  for(int i=0; i<MAXSEQ; i++){
    if(order_seq != NULL) order_seq[i]=i;
    for(unsigned j=0; j<annotations[i].size(); j++)
      delete annotations[i].at(j);
    if(bs[i] != NULL){ delete bs[i]; bs[i]=NULL; };
    annotations[i].clear();
    sources[i].clear();
    segment_offset_index[i].clear();
    segment_offsets[i].clear();
  }
  clear_offset();

  atseqmap.clear();

  seqfiles.clear();
  seqnames.clear();

  is_tmurasakix = false;
  ntlen.clear();
  anchors.clear();
  anchor_segments.clear();
  
  tfidf.clear();

  anchor_ranges.clear();
  anchor_ranges_f.clear();
  anchor_ranges_b.clear();

  rift_filter_enabled = false;
  for(int i=0; i<MAXSEQ; i++) rift_filter_policy[i] = 0;  

  excepted = -1;
  has_detailed_statistics = false;
  has_sequence_statistics = false;
  annotationsChanged = false;

  order_seq=NULL;
  max_left_offset = 0;
  excepted = -1;
  has_detailed_statistics = false;
  has_sequence_statistics = false;

  has_cog = false;
  cogs.clear();
  gene_cog.clear();
  cog_specie.clear();
  gene_specie.clear();
  specie_gene.clear();

  rep_ptn.clear();
  rep_pos.clear();

  for(int i=0; i<MAXSEQ; i++)
    seq_offset[i]=0;

  annotationsChanged = false;
}

void results::chk_order_seq(){
  bool tsoc[MAXSEQ];
  int seqs=0;
  basestat** bs = mr.bs;
  while(bs[seqs] != NULL) seqs++;

  for(int i=0; i<MAXSEQ; i++) tsoc[i]=false;

  for(int i=0; i<seqs; i++){
    int smallesti=MAXSEQ+1;
    int smallest =MAXSEQ+1;
    for(int j=0; j<seqs; j++)
      if(order_seq[j]<smallest && tsoc[j]==false)
	{ smallesti=j; smallest=order_seq[j]; }
    tsoc[smallesti] = true;
    order_seq[smallesti] = i;
  }
}

// returns absolute location, with sequence order lookup
seqlen_t results::raa_native(int row, int col) const{
  int s = col/3;
  int cc = status.seq_lookup(s)*3;
  int c = cc + col%3;

  return anchors.at(row).at(c);
}

// returns absolute location with segment offset, with sequence order lookup
seqlen_t results::raa_physical(int row, int col) const{
  int s = col/3;
  int cc = status.seq_lookup(s)*3;
  int c = cc + col%3;

  seqlen_t ret = anchors.at(row).at(c);
 
  if(col%3 < 2){ // apply segment offset
    
    ret += mr.get_segment_offset(status.seq_lookup(s),
				 anchors.at(row).at(cc),
				 anchors.at(row).at(cc+1));
  }

  return ret;
}

// returns offseted location, with sequence order lookup
seqlen_t results::raa(int row, int col) const{
  int s = col/3;
  int cc = status.seq_lookup(s)*3;
  int c = cc + col%3;

  //  cout << "raa s=" << s << " slkup=" << status.seq_lookup(s) << " r=" << row << " col=" << col << " c=" << c << endl;

  seqlen_t ret = abs(anchors.at(row).at(c));

  if(col%3 < 2){ // apply sequence / segment offset
    ret += get_segment_offset(status.seq_lookup(s), 
			      anchors.at(row).at(cc),
			      anchors.at(row).at(cc+1));

    if(ret < 0) ret -= (seq_offset[status.seq_lookup(s)]+max_left_offset);
    else        ret += (seq_offset[status.seq_lookup(s)]+max_left_offset);
  }

  return ret;
}

// read annotation track: returns annotation with segment remapping
annotation results::rat(atIte i){
    annotation a = i.first->anno_array.at(i.second);
    seqlen_t offset = get_segment_offset(atseq_lookup((annotationTrack*)(i.first)), a.begin, a.end);
    a.begin += offset;
    a.end += offset;
    return a;
}

// read annotation track: returns pointer to an annotaion
// USE WITH CARE: the result from this func comes without segment remapping
annotation* results::rat_native(atIte i){
    return &((annotationTrack*)(i.first))->anno_array[i.second]; 
}

// list anchors between 'a' and 'b' on seq 's' in 'res'
void results::lookup_anchor_segments_direct(int s, seqlen_t a, seqlen_t b, set<int> &res){
  seqlen_t A=abs(a);
  seqlen_t B=abs(b);
  seqlen_t C;
  if(A>B){ C=A; A=B; B=C; }
  anchor_segments[s].query(A, B, res);
} 

void results::lookup_anchor_segments(int s, seqlen_t a, seqlen_t b, set<int> &res){
  seqlen_t A=abs(a);
  seqlen_t B=abs(b);
  seqlen_t C;
  if(A>B){ C=A; A=B; B=C; }
  anchor_segments[status.seq_lookup(s)].query(A, B, res);
} 

pair<int, int> results::hotspot_range_by_pattern(unsigned i) const { 
  return pair<int, int>(max_rep_p, min_rep_p); 
}

pair<int, int> results::hotspot_range_by_combination(unsigned i) const { 
  return pair<int, int>(max_rep_c, min_rep_c); 
}

pair<hspotIte, hspotIte> results::hotspot_pos_lookup(int s, int p) const {
  return rep_pos[s].equal_range(p); 
}

pair<bool, seqlen_t> results::get_hotspot_pos(int s, hspotIte i) const {
  bool reverse = i->second < 0;
  seqlen_t pos = abs(i->second);
  pos += get_segment_offset(s, pos);
  pos += max_left_offset + seq_offset[s]; // not good for dot-plot mode
  return pair<bool, seqlen_t>(reverse, pos);
}

void results::clear_offset(){
  for(int i=0; i<MAXSEQ; i++) seq_offset[i]=0;
  max_left_offset=0;
}

void results::set_offset(int s, seqlen_t offset){ 
  seq_offset[s]=offset; 

  if(offset<0){
    if(offset < -max_left_offset) max_left_offset = -offset;
  }
};

seqlen_t results::maxbases_physical(bool* toggle_seq){
  seqlen_t longest=0;
  int seq=0;

  while(bs[seq]!=NULL){
    if(toggle_seq[seq]){
      if(bs[seq]->totalbases() > longest) longest=bs[seq]->totalbases();
      //assert(cout<<"results::maxbases(): check seq #" << seq << endl);
    }
    seq++;
  }
  return longest;
}

seqlen_t results::maxbases(bool* toggle_seq){
  seqlen_t longest=0;
  int seq=0;

  while(bs[seq]!=NULL){
    if(toggle_seq[seq] && seq != excepted){
      seqlen_t l=bs[seq]->totalbases();
      l+=seq_offset[seq];
      l+=max_left_offset;
      if (l > longest) longest=l;
      //assert(cout<<"results::maxbases(): check seq #" << seq << endl);
    }
    seq++;
  }
  return longest;
}

void results::except(int s){
  excepted = status.seq_lookup(s); // need seq_lookup()?
  //  cout << "excepted: " << excepted << endl;
  bak_max_left_offset = max_left_offset;
  bak_seq_offset = seq_offset[excepted];
  seq_offset[excepted]=0;

  max_left_offset = 0;
  for(int i=0; i<MAXSEQ; i++)
    if(seq_offset[i] < -max_left_offset) max_left_offset = -seq_offset[i];
};

void results::restore_except(){
  max_left_offset = bak_max_left_offset;
  seq_offset[excepted] = bak_seq_offset;
  for(int i=0; i<MAXSEQ; i++)
    if(seq_offset[i] < -max_left_offset) max_left_offset = -seq_offset[i];
  excepted = -1;
};

void results::justify(int pos){
  clear_offset();
  seqlen_t l=maxbases_physical(status.toggle_seq);

  for(unsigned int i=0; i<acols()/3; i++){
    switch(pos){
    case JUSTIFY_LEFT: set_offset(i,0); break;
    case JUSTIFY_CENTER: set_offset(i, abs(l-getbs(i)->totalbases())/2); break;
    case JUSTIFY_RIGHT:  set_offset(i, abs(l-getbs(i)->totalbases())); break;
    }
  }
}

rvector<annotationTrack*>& results::annoTracks(int s){
  return(annotations[status.seq_lookup(s)]);
}

seqlen_t results::get_offset(int s) const { 
  return seq_offset[status.seq_lookup(s)]; }

seqlen_t results::get_offset_t(int s) const { 
  if(order_seq[s] != excepted) 
    return seq_offset[status.seq_lookup(s)]+max_left_offset;
  else
    return 0;
}

basestat* results::getbs_direct(int seq) const {
  return bs[seq];
}

basestat* results::getbs(int seq) const {
  return bs[status.seq_lookup(seq)];
}

bool results::anchorFilter(int anc) const {
  // rift filter
  if(rift_filter_enabled){
    for(unsigned i=0; i<acols()/3; i++){
      if(rift_filter_policy[i] == RIFT_FILTER_DONT_CARE) continue;
      if((rift_filter_policy[i] == RIFT_FILTER_MUST_HAVE_RIFT &&
	  anchors.at(anc).at(i*3  ) != -1 &&
	  anchors.at(anc).at(i*3+1) != -1 ) ||
	 (rift_filter_policy[i] == RIFT_FILTER_MUST_NOT_HAVE_RIFT &&
	  anchors.at(anc).at(i*3)   == -1 &&
	  anchors.at(anc).at(i*3+1) == -1 )) return false;
    }
  }

  // score filter
  bool score_loaded = (!tfidf.empty());
  if(score_loaded){
    if(!tfidf.inrange(anc)) return false;

    for(vector<annotationTrack*>::const_iterator at=filterTracks.begin();
	at!=filterTracks.end(); at++){
      if( ! (*at)->checkAncFilter(anc) ) return false;
    }
  }
 
  return true;
}

bool results::geneFilter(annotationTrack* thistrack, int a){
  // a= annot #

  bool score_loaded = !tfidf.empty();

  if( (status.hide_anchored_gene || status.hide_unanchored_gene) &&
      thistrack->isFilterable()){
    bool unanchored = (rat(thistrack->at(a)).num_ancs == 0);
    if(score_loaded){
      unanchored = unanchored ||
	rat(thistrack->at(a)).tfidf_max < tfidf.getmin();
      unanchored = unanchored || 
	((!tfidf.empty()) &&
	 rat(thistrack->at(a)).tfidf_min > tfidf.getmax());
    }

    if( unanchored && status.hide_unanchored_gene) return false;
    if(!unanchored && status.hide_anchored_gene  ) return false;
  }

  for(unsigned t=0; t<filterTracks.size(); t++)
    if (! filterTracks[t]->checkGenFilter(thistrack, a)) return false;
  
  return true;
}

void results::load_repeat_file(const string& repeatfile_n, int seqno){
  IFSTREAM repeatfile;

  pbar_reset("Loading repeats...");
  pbar_set_task(0,1);

  repeatfile.open(repeatfile_n.c_str());
  repeatfile.seekg(0, std::ios::end);
  double repeatfilelen = repeatfile.tellg();
  repeatfile.seekg(0, std::ios::beg);

  if(repeatfilelen > 100000000){
    status.infobar.push_back(string("Loading .repeats file is canceled because it's too large."));
    repeatfile.close();
    return;
  }

  rep_ptn.clear(); 
  rep_pos.clear();
  rep_c.clear();
  rep_p.clear();

  int r_cnt = 0;
  do {
    string r_pat;
    vector<string> r_pos;
    string buf;

    getline(repeatfile, r_pat);
    if (repeatfile.eof()) break;
    do {
      getline(repeatfile, buf);
      if( buf!="" ) r_pos.push_back(buf);
    } while( buf!="" && !repeatfile.eof());

    assert(cout << r_cnt << endl);
    rep_ptn.push_back(r_pat.substr(3));
    rep_pos.push_back(multimap<int,seqlen_t>());
    for(vector<string>::iterator r_pos_i=r_pos.begin(); r_pos_i!=r_pos.end(); 
	r_pos_i++){
      int s = atoi( r_pos_i->substr(0, r_pos_i->find_first_of(":")).c_str() );
      seqlen_t nl = is_tmurasakix ? ntlen[s] : 0;

      size_t p = r_pos_i->find_first_of(":")+2;
      while(p!=string::npos){
	size_t p2 = r_pos_i->find_first_of(" ", p);
	seqlen_t pos = atoll(r_pos_i->substr(p, p2-p).c_str());
	if(is_tmurasakix){
	  if(pos < 3*nl){
	    pos = pos%(nl-1);
	  } else {
	    pos = nl-(nl-bs[s]->totalbases()+1) - ((pos-3*nl)%(nl-1));
	  }
	}
	assert(cout << "[" << pos << "] ");
	rep_pos.at(r_cnt).insert(pair<int,seqlen_t>(s, pos));
	p = r_pos_i->find_first_not_of(" ", p2);
      }
      assert(cout << endl);
    }

    // statistics
    int combinations=0, totalpoints=0;
    for(int s=0; s<seqno; s++){
      int points = (int)rep_pos.at(r_cnt).count(s);
      if(s==0) combinations = points; else combinations *= points;
      totalpoints += points;
    }

    if(r_cnt==0){
      max_rep_c = min_rep_c = combinations;
      max_rep_p = min_rep_p = totalpoints;
    } else {
      if( combinations > max_rep_c ) max_rep_c = combinations;
      if( combinations < min_rep_c ) min_rep_c = combinations;
      if( totalpoints > max_rep_p ) max_rep_p = totalpoints;
      if( totalpoints < min_rep_p ) min_rep_p = totalpoints;
    }

    rep_c.push_back(combinations);
    rep_p.push_back(totalpoints);

    // go next step 
    r_cnt++;
    pbar_set_fraction( repeatfile.tellg()/repeatfilelen );
  } while (true);
}

string results::disabledFeatures(){
  string f;
  for(unsigned i=0; i<acols()/3; i++){
    for(rvector<annotationTrack*>::iterator j=annotations[i].begin_all();
	j!=annotations[i].end_all(); j++){
      annotationTrack* t = (annotationTrack*)*j;
      for(set<string>::iterator ite=t->hidden_features.begin();
	  ite!=t->hidden_features.end(); ite++){
	f += *ite + ",";
      }
      f += "|";
    }
  }
  //  cout << f << endl;
  return f;
}

void results::disabledFeaturesImport(const string& s){
  vector<set<string> > hidden_features;

  // extract the string
  size_t seq_pos0, seq_pos1;
  size_t pos0, pos1;
  seq_pos0 = seq_pos1 = 0;
  while( (seq_pos1 = s.find_first_of("|", seq_pos0)) != string::npos){
    string features = s.substr(seq_pos0, seq_pos1-seq_pos0);
    //    cout << "+" << features << endl;
    pos0 = pos1 = 0;
    set<string> featureset;
    while( (pos1 = features.find_first_of(",", pos0)) != string::npos){
      string feature = features.substr(pos0, pos1-pos0);
      //      cout << feature << endl;
      featureset.insert(feature);
      pos0 = pos1+1;
    }
    hidden_features.push_back(featureset);
    seq_pos0 = seq_pos1+1;
  }

  // then set the status
  unsigned fset = 0;
  for(unsigned i=0; i<acols()/3; i++){
    for(rvector<annotationTrack*>::iterator j=annotations[i].begin_all();
	j!=annotations[i].end_all(); j++){
      annotationTrack* t = (annotationTrack*)*j;
      t->hidden_features = hidden_features[fset];
      if(++fset == hidden_features.size()) break; // this shouldn't happen
    }
  }
}

string results::annotationProfile(){
  string profile="";
  for(unsigned i=0; i<acols()/3; i++){
    // rvector profile: order and toggle status
    profile = profile + annotations[i].profileString() + "|";

    // append track names
    for(rvector<annotationTrack*>::iterator j=annotations[i].begin_all();
	j!=annotations[i].end_all(); j++){
      annotationTrack* t = (annotationTrack*)*j;
      profile = profile + t->get_name() + "|";
    }
    profile = profile + "|";

    // append score filter configurations
    for(rvector<annotationTrack*>::iterator j=annotations[i].begin_all();
	j!=annotations[i].end_all(); j++){
      annotationTrack* t = (annotationTrack*)*j;

      profile = ( profile+
		  (t->filterAnchor() ? "1" : "0")+"|"+
		  (t->filterGene() ? "1" : "0")+"|"+
		  (t->filterActive() ? "1" : "0")+"|"+
		  itos(t->filterPolicy())+"|"+
		  //
		  itos(t->rangeFilterPolicy())+"|"+
		  itos(t->filterMin())+"|"+
		  itos(t->filterMax())+"|"+
		  //
		  (t->countFilterLower() ? "1" : "0")+"|"+
		  itos(t->countFilterThreshold())+"|"+
		  (t->countFilterLess() ? "1" : "0")+"|"+
		  itos(t->countFilterPolicy())+"|"+
		  itos(t->countFilterCount())+"|"+
		  itos(t->countFilterPercentage())+"|"+

		  (t->emphasizePlot() ? "1" : "0")+"|"+
		  itos(t->plotStyle())+"|"+
		  (t->showName() ? "1" : "0")+"|"
		  );
    }    
    profile = profile + "|";
  }

  return profile;
}

void results::annotationProfileImport(string profile){
  int o[MAXSEQ];
  bool a[MAXSEQ];

  vector<string> prof_tokens = split(profile, "|");

  unsigned pos = 0;
  unsigned seq = 0;
  do {
    unsigned pos2=pos;
    while(prof_tokens[pos2] != "") pos2++;
    unsigned tracks = (pos2-pos)/2;

    for(unsigned t=0; t<tracks; t++){
      int start=pos+tracks*3+2+t*16; // 15 is # of properties

      o[t] = atoi(prof_tokens[pos+t].c_str());
      a[t] = atoi(prof_tokens[pos+tracks+t].c_str());

      annotations[seq].at_direct(t)->
	filterAnchor(atoi(prof_tokens[start++].c_str()));
      annotations[seq].at_direct(t)->
	filterGene(atoi(prof_tokens[start++].c_str()));
      annotations[seq].at_direct(t)->
	filterActive(atoi(prof_tokens[start++].c_str()));
      annotations[seq].at_direct(t)->
	filterPolicy(atoi(prof_tokens[start++].c_str()));

      annotations[seq].at_direct(t)->
	rangeFilterPolicy(atoi(prof_tokens[start++].c_str()));
      annotations[seq].at_direct(t)->
	filterMin(atof(prof_tokens[start++].c_str()));
      annotations[seq].at_direct(t)->
	filterMax(atof(prof_tokens[start++].c_str()));
      
      annotations[seq].at_direct(t)->
	countFilterLower(atoi(prof_tokens[start++].c_str()));
      annotations[seq].at_direct(t)->
	countFilterThreshold(atof(prof_tokens[start++].c_str()));
      annotations[seq].at_direct(t)->
	countFilterLess(atoi(prof_tokens[start++].c_str()));
      annotations[seq].at_direct(t)->
	countFilterPolicy(atoi(prof_tokens[start++].c_str()));
      annotations[seq].at_direct(t)->
	countFilterCount(atoi(prof_tokens[start++].c_str()));
      annotations[seq].at_direct(t)->
	countFilterPercentage(atof(prof_tokens[start++].c_str()));

      annotations[seq].at_direct(t)->
	emphasizePlot(atoi(prof_tokens[start++].c_str()));
      annotations[seq].at_direct(t)->
	plotStyle(atoi(prof_tokens[start++].c_str()));
      annotations[seq].at_direct(t)->
	showName(atoi(prof_tokens[start++].c_str()));
    }

    annotations[seq].load_actives(a);
    annotations[seq].load_order(o);
    
    seq++;
    pos += tracks*(2+1+16)+3; // order,toggle + name + scorefilter + separators
  } while(pos<prof_tokens.size());
}

int results::load_seq_files(bool loadseq, bool loadannot){
  string buf;

  int seqno = -1;

  assert(cout << "load_seq_files: will load " << seqfiles.size() << " seqs" << endl);

  for( vector<string>::iterator fn=seqfiles.begin(); 
       fn!=seqfiles.end(); fn++ ){
    
    assert(cout << "results::load_seq_files(): load " << *fn << endl);

    IFSTREAM fasta;
    IFSTREAM lengthfile;
    bool has_lengthfile = false;
    bool reading = false; // reached to the sequence body
    seqlen_t l=0;

    // if hoge.seq.length is newer than hoge.seq, believe hoge.seq.length
    if(check_timestamp(*fn, string(*fn+".len"))){
      lengthfile.open(string(*fn+".len").c_str());
      if(lengthfile){
	lengthfile >> l;
	lengthfile.close();
	has_lengthfile = true;
      }
    }

    fasta.open(fn->c_str());
      
    if(!fasta){ cout << "results::load_seq_files() : " << fn->c_str()
		     << " couldn't be open" << endl; exit(-1);	}

    fasta.seekg(0, std::ios::end);
    double fastalen = fasta.tellg();
    fasta.seekg(0, std::ios::beg);

    bool format_is_fasta   = false;
    bool format_is_genbank = false;
    bool format_is_stitch  = fn->substr(fn->find_last_of(".")) == ".stitch";
    assert(cout << ( format_is_stitch ? "this sequence file is stitch.\n" :""));

    // prepare basestat
    if(seqno != -1){
      bs[seqno]->finishup();
      assert(bs[seqno]->disp_summary()); // show stat summary
    }

    seqno++;
    if (seqno == MAXSEQ) { cerr<< "too many seqs." << endl; exit(-1);}

    //    delete bs[seqno];
    bs[seqno]=new basestat;
    bs[seqno]->init("default name");
    
    // ------------------------------
    // load sequence!

    // loading counters
    int line=0;

    // multifasta tweaks
    int mf_count=0;
    seqlen_t mf_from=0, mf_to=0;
    string mf_name;
    vector<seqlen_t> mf_l_list;
    vector<string> mf_n_list;
    
    // go!
    while ( getline(fasta, buf) != NULL ){
      if (format_is_stitch ||
	  buf[0] == '>' ||
	  buf.substr(0, 5) == "LOCUS" ||
	  buf.substr(0, 6) == "ORIGIN"){

	// check format
	if(buf.substr(0, 1) == ">") format_is_fasta = true;
	if(buf.substr(0, 5) == "LOCUS" || 
	   buf.substr(0, 6) == "ORIGIN") format_is_genbank = true;

	// genbank  quick load: get sequence length in l
	if(buf.substr(0, 5) == "LOCUS"){
	  stringstream ss;
	  string useless;
	  ss << buf;
	  ss >> useless;  // LOCUS
	  ss >> useless;  // ACCESSION
	  ss >> l;
	}
	assert(cout << "length = " << l << endl);

	assert(cout << "reading sequence..." << endl);
	reading = true;

	// filename + (fasta sequence name) for basestat
	string seq_name = seqnames.at(seqno);
	seq_name = seq_name.substr(seq_name.find_last_of("/")+1);
	bs[seqno]->name= seq_name;

	// load
	if(format_is_fasta){
	  assert(cout << "it's fasta !" <<endl);

	  // if .len is available, skip all but read the cache
	  if(has_lengthfile && !loadseq){
	    IFSTREAM lenfile;
	    lenfile.open(string(*fn+".len").c_str());
	    if(lenfile){
	      seqlen_t tmp_l;
	      lenfile >> tmp_l; // total length to throw away
	      lenfile >> tmp_l;
	      while(!lenfile.eof()){

		mf_count++;
		getline(lenfile, mf_name);
		size_t mf_name_start = mf_name.find_first_not_of(" \t");
		mf_name = ((mf_name_start == string::npos ) ? "(null)"
			   : mf_name.substr(mf_name_start));

		mf_to = mf_from + tmp_l;
		add_source(seqno, seq_source(mf_from, mf_to, mf_name,
						       itos(mf_count), ""));
		mf_from = mf_to + 10;
		lenfile >> tmp_l;
	      }
	      bs[seqno]->addfake(mf_to);
	      break;
	    }
	  }

	  if(mf_count != 0){
	    mf_to = bs[seqno]->totalbases();
	    //	    cout << "MF #" << mf_count << ": " << mf_name << " : " << mf_from << "-" << mf_to << endl;
	    add_source(seqno, seq_source(mf_from, mf_to, mf_name,
					 itos(mf_count), ""));

	    mf_n_list.push_back(mf_name);
	    mf_l_list.push_back(mf_to-mf_from);

	    bs[seqno]->addseq((char*)"NNNNNNNNNN");
	    mf_from = bs[seqno]->totalbases();
	  }

	  size_t p = buf.substr(1).find_first_not_of(" \t");
	  if(p==string::npos) p=0;
	  mf_name = buf.substr(1).substr(p);

	  mf_count++;
	}
	  
	// ------------------------------
	// Let's load!
	pbar_reset("Loading " + bs[seqno]->name);
	assert(cout << "Loading " << seqnames[seqno] << endl);

	// genbank: load annotation
	if(format_is_genbank){
	  if(loadannot){
	    pbar_set_task(0, 1);
	    annotationTrack* at = new annotationTrack;
	    at->readstream(fasta, seqnames[seqno], seqno, mr,
			   pbar_set_fraction);
	    annotations[seqno].push_back(at);
	  } else {
	    while ( getline(fasta, buf) != NULL ){
	      if(buf.substr(0, 6)=="ORIGIN") break;
	    }
	  }
	}

	// stitch: load sequence list (and make dummy sequence statistics)
	if(format_is_stitch){
	  stringstream ss;
	  string sub_name;
	  seqlen_t sub_length;
	  seqlen_t sub_f, sub_t;

	  ss << buf;
	  ss >> sub_name; 
	  ss >> sub_length;
	  ss >> sub_f;
	  ss >> sub_t;

	  assert(cout << "stitch subsequence " << itos_comma(sub_length) << " assigned in "
		 << itos_comma(sub_f) << "-" << itos_comma(sub_t) << endl);

	  add_source(seqno, seq_source(sub_f, sub_t, sub_name,
						 itos(line+1), ""));

	  bs[seqno]->addfake(sub_length + STITCH_GAP);
	  has_sequence_statistics = true;
	}
      } else {
	// sequence body
	if(reading){
	  has_sequence_statistics = false;

	  // if length is already known, skip reading!
	  if(l!=0 && !loadseq){ bs[seqno]->addfake(l); break; }

	  // otherwise, work hard...
	  has_sequence_statistics = true;
	  bs[seqno]->addseq((char*)buf.c_str());
	  if(fastalen!=-1) pbar_set_fraction( fasta.tellg()/fastalen );
	  else pbar_set_fraction(-1);
	}
      }
      line++;
    } // end reading
    if(format_is_stitch) bs[seqno]->addfake(1-STITCH_GAP); // no gap at end
    if(format_is_fasta && !has_lengthfile){
      mf_to = bs[seqno]->totalbases();
      //      cout << "MF #" << mf_count << ": " << mf_name << " : " << mf_from << "-" << mf_to << endl;
      add_source(seqno, 
		 seq_source(mf_from, mf_to, mf_name, itos(mf_count), ""));
      mf_n_list.push_back(mf_name);
      mf_l_list.push_back(mf_to-mf_from);

      // generate cache
      ofstream l;
      l.open(string(*fn+".len").c_str());
      if(l){
	l << bs[seqno]->totalbases() << endl;
	for(unsigned i=0; i<mf_n_list.size(); i++)
	  l << mf_l_list[i] << "\t" << mf_n_list[i] << endl;
	l.close();
      }
    } 

    fasta.close();
    if(format_is_fasta || format_is_stitch)
      status.annotation_files.push_back(pair<int,string>(seqno, string(*fn+".cds")));

    // rename if alternative is specified
    if(seqno < (int)status.sequence_names.size())
      bs[seqno]->name = status.sequence_names[seqno];

  }

  bs[seqno]->finishup();
  assert(bs[seqno]->disp_summary());

  seqno++;
  bs[seqno] = NULL;
  return seqno;
}

void results::prepare_anchor_segments(int n){
  anchor_segments.clear();
  for(int i=0; i<n; i++) anchor_segments.push_back(segment_tree<int>());
}

void results::insert_anchor_row(vector<seqlen_t>& a){
  if(is_tmurasakix){
    for(unsigned i=0; i<a.size(); i++){
      seqlen_t nl = ntlen[i/3];
      if((i+1)%3 != 0){
	if(a[i] < 3*nl){
	  a[i] = a[i]%(nl-1);
	} else {
	  a[i] = nl-(nl-bs[i/3]->totalbases()+1) - ((a[i]-3*nl)%(nl-1));
	}
      }
    }
  }
  
  for(unsigned c=0; c<a.size()/3; c++)
    anchor_segments[c].insert(a[c*3], a[c*3+1], anchors.size());
  anchors.push_back(a);
}
  
void results::build_anchor_segment_tree(int s, void(*setfrac)(double)){
  anchor_segments[s].build_tree(setfrac);
}

void results::loadcog(string cogpath){
  pbar_reset("Loading COG orthlog list...");
  IFSTREAM cog;
  cog.open(cogpath.c_str());
  if(!cog){
    // COG file not found...
    return; 
  }

  cog.seekg(0, std::ios::end);
  double coglen = cog.tellg();
  cog.seekg(0, std::ios::beg);

  string buf;
  string cogid, cogname;
  string locus;
  string last_specie = "";

  while( getline(cog, buf) != NULL ){
    if(buf[0]=='['){
      size_t f, t;
      f = buf.find_first_of(" ") + 1;
      t = buf.find_first_of(" ", f);
      cogid = buf.substr(f, t-f);
      cogname = buf.substr(t+1);
      cogs[cogid]=cogname;
    }

    if(buf.substr(0,2)=="  "){
	size_t f, t;
      string specie;
      if(buf.find(":")!=string::npos){
	
	// extract specie
	f = buf.find_first_not_of(" ");
	t = buf.find(":");
	specie = buf.substr(f, t-f);
	last_specie = specie;
      } else specie = last_specie;
      
      // extract locus tags
      f = buf.find(":")+1;
      while( (f = buf.find_first_not_of(" ", f)) != string::npos){
	t = buf.find(" ", f);
	locus = buf.substr(f, t-f);
	gene_cog[locus]=cogid;
	cog_specie.insert(pair<string, string>(cogid, specie));
	gene_specie[locus]=specie;
	specie_gene.insert(pair<string,string>(specie, locus));
	f = t;
      }
    } else last_specie="";
    pbar_set_fraction( cog.tellg()/coglen );
  }

  has_cog = true;

  cog.close();
}

string results::summarize_in_text(){
  int i=0;
  ostringstream stats;
  while(bs[i]!=NULL){
    bs[i]->summary();
    seqlen_t tf = anchor_ranges_f.at(i).total_length();
    seqlen_t tb = anchor_ranges_b.at(i).total_length();
    seqlen_t tt = anchor_ranges.at(i).total_length();
    seqlen_t len = bs[i]->totalbases();
    
    int af = anchors_num_f[i];
    int ab = anchors_num_b[i];
    int aa = af + ab;
    
    stats << "[ Sequence #" << i+1 << ": " << bs[i]->name << " ]" << endl
	  << endl
	  
	  << "Anchor coverage: " << endl
	  << "                      Length (bp) "
	  << "           # of anchors" << endl
	  << "---------------------------------------"
	  << "---------------------------------------" << endl
	  << "Forward  " << setw(14) << itos_comma(tf) << " ("
	  << setw(6) << itos((double)tf*100/len,2) << "%)"
	  << setw(14) << itos_comma(af) 
	  << " (" << setw(6) << itos((double)af*100/aa,2) << "%)" << endl

	  << "Backward " << setw(14) << itos_comma(tb) << " ("
	  << setw(6) << itos((double)tb*100/len,2) << "%)" 
	  << setw(14) << itos_comma(ab) 
	  << " (" << setw(6) << itos((double)ab*100/aa,2) << "%)" << endl

	  << "Total    " << setw(14) << itos_comma(tt) << " ("
	  << setw(6) << itos((double)tt*100/len,2) << "%)"
	  << setw(14) << itos_comma(aa) 
	  << " (" << setw(6) << itos((double)aa*100/aa,2) << "%)" << endl
	  << "---------------------------------------"
	  << "---------------------------------------" << endl
	     
	  << endl
	  << "Base statistics:" << endl
	  << bs[i]->summary()<< endl;
    i++;
  }

  return stats.str();
}

string results::summarize_in_html(){
  string html = "";
  char buf[4097];

  snprintf(buf, 4096, 
	   "\
<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\">\n\
<html>\n\
  <head>\n\
    <title>Base/Anchor Statistics Summary: %s</title>\n\
    <style type=\"text/css\"><!--\n\
      TABLE { border-collapse: collapse; }\n\
      TH, TD { border: 1px solid black; }\n\
     -->\n\
    </style>\n\
  </head>\n\
\n\
  <body>\n\
    <h1>Base/Anchor Statistics Summary: %s</h1>\n",
	   string(status.basename+status.anchorfile_extension).c_str(),
	   string(status.basename+status.anchorfile_extension).c_str());
  
  html = html + buf;
  
  int seqno=0;
  while(seqno < MAXSEQ && bs[seqno] != NULL){
    basestat *b = bs[seqno];

    snprintf(buf, 4096, 
	     "<h2>Sequence #%d: %s</h2>\n\n", seqno+1, b->name.c_str());
    html = html + buf;

    if(has_detailed_statistics){
      seqlen_t anchorlen = anchor_ranges.at(seqno).total_length();
      seqlen_t tf = anchor_ranges_f.at(seqno).total_length();
      seqlen_t tb = anchor_ranges_b.at(seqno).total_length();
      
      int af = anchors_num_f[seqno];
      int ab = anchors_num_b[seqno];

      snprintf(buf, 4096,
	       "\
    <h3>Anchor Coverage</h3>\n\
    <table>\n\
      <thead>\n\
	<tr>\n\
	  <td>\n\
	  <td colspan=\"2\">Length (bp)\n\
	  <td colspan=\"2\"># of Anchors\n\
      </thead>\n\
      <tbody>\n\
	<tr><td>Forward  <td>%lld <td>%6.2lf <td>%d <td>%6.2lf \n\
	<tr><td>Backward <td>%lld <td>%6.2lf <td>%d <td>%6.2lf \n\
	<tr><td>Total    <td>%lld <td>%6.2lf <td>%d <td>%6.2lf \n\
      </tbody>\n\
    </table>\n\n",
	       tf, (double)tf*100/b->totalbases(), af, (double)af*100/(af+ab),
	       tb, (double)tb*100/b->totalbases(), ab, (double)ab*100/(af+ab),
	       anchorlen, (double)anchorlen*100/b->totalbases(), 
	       af+ab, (double)(af+ab)*100/(af+ab)
	       );

      html = html+buf;
    }

    if(has_sequence_statistics){
      seqlen_t tb = b->totalbases();

      snprintf(buf, 4096,
	       "\
    <h3>Base Statistics</h3>\n\
    <table>\n\
      <thead>\n\
	<tr>\n\
	  <td>\n\
	  <td colspan=\"2\">Non-repeat\n\
	  <td colspan=\"2\">Repeat\n\
	  <td colspan=\"2\">Total\n\
      </thead>\n\
      <tbody>\n\
	<tr><td>A <td>%lld <td>%6.2lf <td>%lld <td>%6.2lf <td>%lld <td>%6.2lf \n\
	<tr><td>T <td>%lld <td>%6.2lf <td>%lld <td>%6.2lf <td>%lld <td>%6.2lf \n\
	<tr><td>C <td>%lld <td>%6.2lf <td>%lld <td>%6.2lf <td>%lld <td>%6.2lf \n\
	<tr><td>G <td>%lld <td>%6.2lf <td>%lld <td>%6.2lf <td>%lld <td>%6.2lf \n\
	<tr><td>N <td>%lld <td>%6.2lf <td>%lld <td>%6.2lf <td>%lld <td>%6.2lf \n\
	<tr><td>Total <td>%lld <td>%6.2lf <td>%lld <td>%6.2lf <td>%lld <td>%6.2lf \n\
      </tbody>\n\
    </table>\n\n",
	       b->A, (double)b->A*100/tb, b->a, (double)b->a*100/tb, 
	       b->A+b->a, (double)(b->A+b->a)*100/tb, 
	       b->T, (double)b->T*100/tb, b->t, (double)b->t*100/tb, 
	       b->T+b->t, (double)(b->T+b->t)*100/tb, 
	       b->C, (double)b->C*100/tb, b->c, (double)b->c*100/tb, 
	       b->C+b->c, (double)(b->C+b->c)*100/tb, 
	       b->G, (double)b->G*100/tb, b->g, (double)b->g*100/tb, 
	       b->G+b->g, (double)(b->G+b->g)*100/tb, 
	       b->N, (double)b->N*100/tb, b->n, (double)b->n*100/tb, 
	       b->N+b->n, (double)(b->N+b->n)*100/tb,
	       b->nonrepeats(), (double)b->nonrepeats()*100/tb,
	       b->repeats(), (double)b->repeats()*100/tb,
	       tb, (double)100);
      html = html+buf;
    }
    seqno++;
  }

  snprintf(buf, 4096,
	   "\
  </body>\n\
</html>\n");
  
  html = html+buf;

  return html;
}

// ----------------------------------------------------------------------
//  mv_status 

int mv_status::seq_rlookup(int ns) const { // native # to geom #
  unsigned s, ss;
  unsigned skip=0;
  for(s=0; s<mr.acols()/3; s++) if(order_seq[s]==ns) break;

  for(ss=0; ss<=s; ss++) if(!toggle_seq[order_seq[ss]]) skip++;
  s -= skip;

  return s;
}

int mv_status::seq_lookup(int gs) const { // geometrycal # to native #
  int seqcount = 0;
  unsigned ns = mr.acols()/3;

  // search visible sequences
  for(unsigned int i=0; i<ns; i++){
    if(toggle_seq[order_seq[i]]){
      if(gs==seqcount) return order_seq[i];
      seqcount++;
    }
  }

  // search invisible sequences
  for(unsigned int i=0; i<ns; i++){
    if(!toggle_seq[order_seq[i]]){
      if(gs==seqcount) return order_seq[i];
      seqcount++;
    }
  }

  // something was wrong...
  return -1;
}

void mv_status::chk_toggle_seq(){
  basestat** bs = mr.bs;

  int c=0; // count 
  int s=0; // seq 

   while(bs[s] != NULL){
     if(status.toggle_seq[s]) c++;
     s++;
   }

   if(c==0){
     s=0;
     while(bs[s] != NULL){
       status.toggle_seq[s]=true;
       s++;
     }
   } else {
     if(c==1){
       s=0;
       while(bs[s] != NULL){
	 if(!status.toggle_seq[s]){ status.toggle_seq[s]=true; break; }
	 s++;
       }
     }
   }

//   status.toggle_seq[0]=true;

//   basestat** bs = mr.bs;
//   int s=0;
//   bool orall=false;
//   while(bs[s] != NULL){
//     if (s != 0) orall = orall || status.toggle_seq[s];
//     s++;
//   }
//   if (!orall) status.toggle_seq[1] = true;
}

void mv_status::backup(){ // backup cmdline opts
  if(status_bkup != NULL) return;
  assert(cout << "mv_status::backup() going on." << endl);
  status_bkup = new mv_status;

  status_bkup->copy(status);
  // sequence_names; // isn't saved
}

void mv_status::restore(){ // restore cmdline opts
  if(status_bkup == NULL) return;
  assert(cout << "mv_status::restore() going on." << endl);

  copy(*status_bkup);
}

bool mv_status::set_filenames(){
  // CAUTION: don't call this function from subthread.

  assert(cout << "mv_status::set_filenames() is called" << endl);

  // let's begin
  char* bn;

  if(! anchorfile_fromGUI )
    bn = strdup(argv[0]); // specified by cmdline 
  else
    bn = strdup(anchorfile.c_str()); // specified by GUI

  string orig_extension;
  
  char *b = bn;
  while(*b != 0){
    if(*b=='\\') *b='/'; b++;
  }

  for(int i=strlen(bn)-1; i>0; i--) 
    if(bn[i]=='.'){bn[i]=0; orig_extension = (string)&bn[i+1]; break;}
  basename = bn;

  int c=0;
  for(int i=strlen(bn)-1; i>0; i--)
    if(bn[i]=='/' /*G_DIR_SEPARATOR*/ ){ c++; if(c==2){ bn[i+1]=0; break;} }
  if(c==1) bn[0]=0;
  basedir = bn;

  char *cwd=getcwd(NULL, 1024);

  if(!g_path_is_absolute(basename.c_str())){
    basename = (string)cwd + "/" + basename;
  }

  // windows compatibility hack
  char *cw = cwd;
  while(*cw != 0){
    if(*cw=='\\') *cw='/';
    cw++;
  }
  // end windows compatibility hack

  // append slash if necessary
  if(status.basedir=="") status.basedir = (string)cwd + "/";

  //  cout << "basedir=" << status.basedir << endl;

  free(cwd);
  free(bn);

  // check whether tmurasakix or not
  string ntseqs = basename+".ntseqs";
  string seqs_ext = ".seqs";
  ifstream ntifs;
  ntifs.open(ntseqs.c_str());
  if(ntifs){
    seqs_ext = ".ntseqs";
    ntifs.close();
    if(!mr.set_tmurasakix(basedir, string(basename+".seqs"))) return false;
  }


  // try to find XXX.anchors and XXX.seqs:
  //  .seqs, .anchors, .anchors.details will be loaded without timestamp check.
  string extension = ".anchors";
  if(orig_extension=="block"){ extension=".block"; }
  string a = basename+extension;
  string s = basename+seqs_ext;
  string ad = basename+extension+".details";
  string tfidf = basename+".stats.tfidf";
  string repeat = basename+".repeats";

  IFSTREAM ifs;
  bool exist=true;

  // if stats.tfidf file is available, load it.
  ifs.open(tfidf.c_str());
  if(ifs && check_timestamp(a, tfidf)){
    tfidffile = tfidf;
  } else {
    ifs.close();
    tfidf = basename+".anchors.stats.tfidf";
    ifs.open(tfidf.c_str());
    if(ifs && check_timestamp(a, tfidf)){
      tfidffile = tfidf;
    } else {
    // or, if anchors.details (or blocks.details) available, use it.
    ifs.close();
    ifs.open(ad.c_str());
    if(ifs) extension+=".details";
    }
  }
  ifs.close();

  // test whether .repeats is available or not
  ifs.open(repeat.c_str());
  if(ifs && check_timestamp(a, repeat)){
    repeatfile = repeat;
  }
  ifs.close();

  // check existence of anchor file
  ifs.open(a.c_str());
  if(!ifs){
    exist=false;
    //    cout << "set_filenames(): no anc:" << a << endl; 
    
    if(use_x11){
      GtkWidget *dialog;
      dialog = gtk_message_dialog_new(GTK_WINDOW(status.mainwindow),
				      GTK_DIALOG_DESTROY_WITH_PARENT,
				      GTK_MESSAGE_ERROR,
				      GTK_BUTTONS_CLOSE,
				      "Error loading anchor file: %s",
				      a.c_str() );
      gtk_dialog_run(GTK_DIALOG(dialog));
      gtk_widget_destroy(dialog);
    }
    return false;  // absolutely abort: will be retry in GUI
  }
  ifs.close();

  // check existence of seqs file
  ifs.open(s.c_str());
  if(!ifs){ 
    exist=false; 
    //    cout << "set_filenames(): no seq:" << s << endl; 
    if(use_x11){
      // prepare for filter: allow only <name>.seqs with <name>.anchors
      size_t pos = s.find_last_of("/", string::npos);
      if(pos==string::npos) pos=0; else pos++;
      string seqlistname=s.substr(pos, string::npos);

      while(!ifs){
	ifs.close();

	GtkWidget *dialog;
	dialog = gtk_message_dialog_new(GTK_WINDOW(status.mainwindow),
					GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_MESSAGE_ERROR,
					GTK_BUTTONS_YES_NO,
					"Error loading sequence list: %s\nSubstitute sequence list (or open another anchor file) ?",
					s.c_str() );
	gint r = gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);
	if (r != GTK_RESPONSE_YES) return false; // don't substitute

	// substitute dialog
	dialog = gtk_file_chooser_dialog_new ("Substitute Sequence List",
					      GTK_WINDOW(status.mainwindow),
					      GTK_FILE_CHOOSER_ACTION_OPEN,
					      GTK_STOCK_CANCEL, 
					      GTK_RESPONSE_CANCEL,
					      GTK_STOCK_OPEN,
					      GTK_RESPONSE_ACCEPT,
					      (gpointer)NULL);
  
	GtkFileFilter *filter = gtk_file_filter_new();
	gtk_file_filter_add_pattern(filter, seqlistname.c_str());
	gtk_file_chooser_set_filter(GTK_FILE_CHOOSER(dialog), filter);
	
	bool file_selected =
	  (gtk_dialog_run(GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT);

	char *filename;
	filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
	gtk_widget_destroy(dialog);
	if(!file_selected) return false; // no substitution
	
	// windows compatibility hack
	char *fn = filename;
	while(*fn != 0){
	  if(*fn=='\\') *fn='/';
	  fn++;
	}
	// end windows compatibility hack
	
	g_free(filename);
	s = filename;
	ifs.open(s.c_str());
      }
      exist=true;  // found it! yeah!
    }
  }
  seqlist = s;
  // ifs.close(); // this stream will be closed later

  // check existence of sequence files (by reading sequence list)
  char fastaname[256];
  mr.seqfiles.clear();
  mr.seqnames.clear();
  while( ifs.getline(fastaname, 255) != NULL ){
    IFSTREAM fasta;
    assert(cout << "initialize(): open fasta " << fastaname << endl);

    if((string)fastaname=="") continue; // skip blank lines

    // try to open: first 
    fasta.open(fastaname);
    mr.seqfiles.push_back(fastaname);
    mr.seqnames.push_back(string(fastaname));

    // try to open: second chance
    if (!fasta){
      fasta.clear(); // clear status bits (errors etc.)
      assert(cout << "couldn't open " << fastaname 
	     << " trying with basedir..." << endl
	     << " (" << (status.basedir+(string)fastaname).c_str() <<")"<< endl);
      fasta.open((status.basedir+(string)fastaname).c_str());
      mr.seqfiles.at(mr.seqfiles.size()-1) = status.basedir+(string)fastaname;
    }
    while (!fasta){ 
      // failed on second challenge :(

      // prepare for filter: filename must be preserved
      size_t pos = string(fastaname).find_last_of("/", string::npos);
      if(pos==string::npos) pos=0; else pos++;
      string seqlistname=string(fastaname).substr(pos, string::npos);

      if (use_x11){
	string newfasta;
	GtkWidget *dialog;
	dialog = gtk_message_dialog_new(GTK_WINDOW(status.mainwindow),
					GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_MESSAGE_ERROR,
					GTK_BUTTONS_YES_NO,
					"Error loading sequence file: %s\nSubstitute sequence file (or open another anchor file) ?",
					fastaname);
	gint r = gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);
	if (r != GTK_RESPONSE_YES) return false; // don't substitute

	// substitute dialog
	dialog = gtk_file_chooser_dialog_new ("Substitute Sequence File",
					      GTK_WINDOW(status.mainwindow),
					      GTK_FILE_CHOOSER_ACTION_OPEN,
					      GTK_STOCK_CANCEL, 
					      GTK_RESPONSE_CANCEL,
					      GTK_STOCK_OPEN,
					      GTK_RESPONSE_ACCEPT,
					      (gpointer)NULL);
  
	GtkFileFilter *filter = gtk_file_filter_new();
	gtk_file_filter_add_pattern(filter, seqlistname.c_str());
	gtk_file_chooser_set_filter(GTK_FILE_CHOOSER(dialog), filter);
	
	bool file_selected =
	  (gtk_dialog_run(GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT);

	char *filename;
	filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
	gtk_widget_destroy(dialog);
	if(!file_selected) return false; // no substitution
	
	// windows compatibility hack
	char *fn = filename;
	while(*fn != 0){
	  if(*fn=='\\') *fn='/';
	  fn++;
	}
	// end windows compatibility hack
	
	newfasta = string(filename);
	g_free(filename);
	fasta.open(newfasta.c_str());
	mr.seqfiles.at(mr.seqfiles.size()-1) = newfasta;
      } else {
	cerr << "couldn't open " << status.basedir+fastaname << endl; 
	return false;
	// exit(-1);
      }
    }
    fasta.close();
  }

  ifs.close();  // close hogehoge.seq

  anchorfile_extension = extension;

  assert(cout << "mv_status::set_filenames(): " << endl
	 << "  basename =" << basename << endl
	 << "  basedir  =" << status.basedir << endl
	 << "  extension=" << extension << endl);

  assert(cout << "mv_status::set_filenames(): try to find'em :" << exist << endl);

  return exist;
}

string mv_status::sequenceProfile(){
  string prof;

  prof="";

  for(unsigned i=0; i<mr.acols()/3; i++){
    prof = prof + "|" + itos(order_seq[i]) + "|" + (toggle_seq[i] ? "1" : "0");
  }
  prof += "|";

  return prof;
}

void mv_status::check_ranges(){
  assert(cout << "mv_status::check_ranges()" << endl);

  fixrange<int>(&default_width,  min_size, max_size);
  fixrange<int>(&default_height, min_size, max_size);
  fixrange<int>(&max_size, MAX_SIZE_MIN, MAX_SIZE_MAX);
  fixrange<int>(&min_size, MIN_SIZE_MIN, max_size);
  fixrange<double>(&zoom_factor, ZOOM_FACTOR_MIN, ZOOM_FACTOR_MAX);

  fixrange<double>(&alpha, ALPHA_MIN, ALPHA_MAX);
  fixrange<double>(&alpha_low, ALPHA_MIN, ALPHA_MAX);
  fixrange<double>(&saturation, SATURATION_MIN, SATURATION_MAX);
  fixrange<double>(&fwdvalue, FWDVALUE_MIN, FWDVALUE_MAX);
  fixrange<double>(&bwdvalue, BWDVALUE_MIN, BWDVALUE_MAX);
  fixrange<double>(&bgvalue,  BGVALUE_MIN, BGVALUE_MAX);
  fixrange<double>(&huediff_fb, HUEDIFF_FB_MIN, HUEDIFF_FB_MAX);
  fixrange<double>(&fill_stroke_ratio, FILL_STROKE_RATIO_MIN, FILL_STROKE_RATIO_MAX);

  // fonts
  fixrange<int>(&font_size_general, FONT_SIZE_MIN, FONT_SIZE_MAX);
  fixrange<int>(&font_size_gene,    FONT_SIZE_MIN, FONT_SIZE_MAX);

  // plotting
  fixrange<double>(&sequence_thickness, 
		   SEQUENCE_THICKNESS_MIN, SEQUENCE_THICKNESS_MAX);
  fixrange<double>(&orf_thickness, 
		   SEQUENCE_THICKNESS_MIN, SEQUENCE_THICKNESS_MAX);
  fixrange<double>(&min_thickness, MIN_THICKNESS_MIN, MIN_THICKNESS_MAX);
  fixrange<double>(&max_thickness, min_thickness, MAX_THICKNESS_MAX);
  fixrange<double>(&min_thickness, 0, max_thickness);

  // chrom link specific params
  fixrange<double>(&chrlink_endcap, CHRLINK_ENDCAP_MIN, CHRLINK_ENDCAP_MAX);
  fixrange<double>(&chrlink_endcap_na, CHRLINK_ENDCAP_MIN, CHRLINK_ENDCAP_MAX);

  // dot plot specific params
  fixrange<double>(&dot_radius, DOT_RADIUS_MIN, DOT_RADIUS_MAX);
  fixrange<double>(&dot_radius_end, DOT_RADIUS_END_MIN, DOT_RADIUS_END_MAX);

  // superzoom cursor
  fixrange<int>(&zoomcursor_width_px, 
		ZOOMCURSOR_WIDTH_PX_MIN, ZOOMCURSOR_WIDTH_PX_MAX);
  fixrange<int>(&zoomcursor_width_bp, 
		ZOOMCURSOR_WIDTH_BP_MIN, ZOOMCURSOR_WIDTH_BP_MAX);

  // annotation
  fixrange<int>(&show_gene_name_threshold, 
		GENE_NAME_THRESHOLD_MAX, GENE_NAME_THRESHOLD_MIN);

  // system settings
  fixrange<int>(&pbcache_cap, PBCACHE_CAP_MIN, PBCACHE_CAP_MAX);
}

void mv_status::set_defaults(){
  assert(cout << "mv_status::set_defaults()" << endl);

  // --------------------------------------------------
  // some initialization
  prog = new progress;
  using_temp_pixbuf = false;

  for(int i=0; i<MAXSEQ; i++){
    order_seq[i]=i; 
    toggle_seq[i]=true;
  }

  gene_markers.clear();
  annotation_files.clear();
  das_tracks.clear();
  cog_file="";

  // --------------------------------------------------
  // Defaults
  pbcache_cap =    DEFAULT_PBCACHE_CAP;

  mainwindow_w = DEFAULT_WIN_WIDTH;
  mainwindow_h = DEFAULT_WIN_HEIGHT;

  // canvas
  min_size =       DEFAULT_MIN_SIZE;
  max_size =       DEFAULT_MAX_SIZE;
  default_width =  DEFAULT_WIDTH;
  default_height = DEFAULT_HEIGHT;

  zoom_factor =    DEFAULT_ZOOM_FACTOR;

  // colors
  color_by =       0;
  color_by_default = 0;
  alpha_by_default = 0;
  saturation = DEFAULT_SATURATION;
  fwdvalue    = DEFAULT_VALUE;
  bwdvalue    = DEFAULT_VALUE;
  bgvalue    = DEFAULT_VALUE;

  for (unsigned i=0; i<3; i++) rift0_col[i] = 0.3;
  rift0_col_for_all = false;
  rift0_use_alt_seq = false;

  for (unsigned i=0; i<10; i++) anno_col[i] = RGBA(0, 0, 0, 0.5);

  // fonts
  font_size_general = DEFAULT_FONT_SIZE_GENERAL;
  font_size_gene    = DEFAULT_FONT_SIZE_GENE;

  // plots
//   min_anchor_score  = 0;
//   max_anchor_score  = 0;

  name_by =        0;
  name_by_default = 0;
  sequence_thickness = DEFAULT_SEQUENCE_THICKNESS;
  orf_thickness = DEFAULT_ORF_THICKNESS;
  min_thickness = DEFAULT_MIN_THICKNESS;
  max_thickness = DEFAULT_MAX_THICKNESS; 
  fill_stroke_ratio = DEFAULT_FILL_STROKE_RATIO;
  chrlink_endcap = DEFAULT_CHRLINK_ENDCAP;
  chrlink_endcap_na = DEFAULT_CHRLINK_ENDCAP;

  min_orf = DEFAULT_MINORF;

  generate_pdf = DEFAULT_PDF;
  generate_png = DEFAULT_PNG;

  hide_unanchored_gene = false;
  hide_anchored_gene   = false;

  plot_type    = PLOT_CHROM_LINKS;
  plot_all     = false;

  cmdopt_forward  = false;
  cmdopt_backward = false;

  show_sidebar_schemes = true;
  show_sidebar_toggles = true;

  anchorfile_fromGUI = false;
  anchorfile = "";
  tfidffile  = "";
  repeatfile = "";

  flushAtRedraw = false;
  loadSeqAtRedraw = false;

  zoomcursor_width_bp = DEFAULT_ZOOMCURSOR_WIDTH_BP;
  zoomcursor_width_px = DEFAULT_ZOOMCURSOR_WIDTH_PX;

  show_gene_marker    = true;
  show_gene_name_threshold = DEFAULT_GENE_NAME_THRESHOLD;
  gene_name_onoff_method = 0;
  gene_name_onoff_method_default = 0;

  show_hotspots       = true;
  hotspot_by = 0;
  hotspot_by_default = 0;

  genewidth   = DEFAULT_GENE_WIDTH;
  genepadding = DEFAULT_GENE_PADDING;
  genemargin  = DEFAULT_GENE_MARGIN;
  justify_by_anchor = false;

  // --------------------------------------------------
  // Defaults (but currently no configuration options)
  max_width_gn = DEFAULT_MAX_WIDTH_GN;
  margin_gn    = DEFAULT_MARGIN_GN;
  margin_lr    = DEFAULT_MARGIN_LR;
  margin_tb    = DEFAULT_MARGIN_TB;

  // defaults (in config menu and used)

  plot_U = DEFAULT_CHR_NMASK;
  plot_R = DEFAULT_CHR_RMASK;
  plot_O = DEFAULT_CHR_ORF;

  plot_fwd = DEFAULT_PLOT_FWD;
  plot_bwd = DEFAULT_PLOT_BWD;

  rift_stencil = false;
  rift_pattern_fill = true;

  clip_hard = true;

  dot_radius = DEFAULT_DOT_RADIUS;
  dot_radius_end = DEFAULT_DOT_RADIUS_END;

  disp_seqno = true;
  disp_range = true;
  disp_seqticks = true;
  disp_source = true;

  opt_stat_html = false;
  opt_stat_text = false;

  auto_order = false;

  antialias = DEFAULT_ANTIALIAS;

  huediff_fb   = DEFAULT_HUEDIFF;

  alpha_low = DEFAULT_ALPHA_LOW;
  alpha =     DEFAULT_ALPHA;

  opt_load_seq = false;
  opt_load_annot = true;

  opt_min_tfidf = 0;

  load_prefs();

  restore(); // apply command line opts
}

void mv_status::zoom_normal(){
  if(plot_type == PLOT_DOT) return;

  plot_type = PLOT_CHROM_LINKS;
  superzoom_from = 0;
  superzoom_to   = mr.maxbases(toggle_seq);
};

void mv_status::zoom_in(){
  if(plot_type == PLOT_DOT) return;
  plot_type = PLOT_SUPERZOOM;
  
  seqlen_t center = (superzoom_from + superzoom_to)/2;
  seqlen_t range  = (superzoom_to - superzoom_from);

  if (range/4<1) return;
  superzoom_to   = center + range/4;
  superzoom_from = center - range/4;
};

void mv_status::zoom_out(){
  if(plot_type == PLOT_DOT) return;
  plot_type = PLOT_SUPERZOOM;

  seqlen_t center = (superzoom_from + superzoom_to)/2;
  seqlen_t range  = (superzoom_to - superzoom_from);

  superzoom_from = center - range;
  if (superzoom_from < 0) superzoom_from = 0;

  superzoom_to = center + range;
  if (superzoom_to > mr.maxbases(toggle_seq)) superzoom_to = mr.maxbases(toggle_seq);
};

void mv_status::go_fwd(){
  if(plot_type != PLOT_SUPERZOOM) return;

  seqlen_t range  = (superzoom_to - superzoom_from);
  superzoom_to += (range*3)/4;
  if (superzoom_to > mr.maxbases(toggle_seq)) superzoom_to = mr.maxbases(toggle_seq);

  superzoom_from = superzoom_to - range;
}

void mv_status::go_bwd(){
  if(plot_type != PLOT_SUPERZOOM) return;

  seqlen_t range  = (superzoom_to - superzoom_from);
  superzoom_from -= (range*3)/4;
  if(superzoom_from<0) superzoom_from=0;

  superzoom_to = superzoom_from + range;
}

// auto reorder
void mv_status::auto_reorder(){
  bool seq_ordered[MAXSEQ];
  for(int i=0; i<MAXSEQ; i++) seq_ordered[i]=false;

  assert(cout << "automatic ordering..." << endl);
  int firstseq = status.order_seq[0];
  int seqs = mr.acols()/3;
      
  double corrector[MAXSEQ];
  for(int s=0; s<seqs; s++) 
    corrector[s]=(double)mr.maxbases(status.toggle_seq)/
      mr.bs[s]->totalbases();
      
  mr.order_seq[0]=firstseq;
  seq_ordered[0]=true;

  int candidate=0;
  int current=firstseq;
  unsigned long long int distance;
  unsigned long long int candidate_distance=0;

  for(int i=1; i<seqs; i++){
    candidate_distance=0;
    candidate_distance--;
    for(int ss=0; ss<seqs; ss++){
      if(seq_ordered[ss] || ss==current) continue;
      distance=0;

      for(unsigned int a=0; a<mr.arows(); a++){
	seqlen_t a1=mr.raa_physical(a,ss*3);
	seqlen_t a2=mr.raa_physical(a,current*3);
	    
	a1 = (seqlen_t)(a1*corrector[ss]);
	a2 = (seqlen_t)(a2*corrector[current]);

	distance+=abs(a2-a1);
      }

      if (distance<candidate_distance){
	candidate = ss;
	candidate_distance = distance; }
    }
    current=candidate;
    mr.order_seq[i]=candidate;
    seq_ordered[candidate]=true;
  }

  //  for(int i=0; i<seqs; i++) cout << i << ":" << mr.order_seq[i] << endl;

}

void mv_status::save_prefs(){
  pref p;
  p.load(preffile());

  p.setpref(default_width,      "default_width");
  p.setpref(default_height,     "default_height");
  p.setpref(sequence_thickness, "sequence_thickness");
  p.setpref(orf_thickness, "orf_thickness");
  p.setpref(saturation, "saturation");
  p.setpref(fwdvalue, "fwdvalue");
  p.setpref(bwdvalue, "bwdvalue");
  p.setpref(bgvalue, "bgvalue");
  p.setpref(bgvalue, "bgvalue");
  p.setpref(alpha, "alpha");
  p.setpref(alpha_low, "alpha_low");
  p.setpref(disp_seqno, "disp_seqno");
  p.setpref(font_size_general, "font_size_general");
  p.setpref(min_orf, "min_orf");
  p.setpref(min_size, "min_size");
  p.setpref(max_size, "max_size");
  p.setpref(zoom_factor, "zoom_factor");
  p.setpref(antialias, "antialias");
  p.setpref(pbcache_cap, "pbcache_cap");
  p.setpref(min_thickness, "min_thickness");
  p.setpref(max_thickness, "max_thickness");
  p.setpref(zoomcursor_width_bp, "zoomcursor_width_bp");
  p.setpref(zoomcursor_width_px, "zoomcursor_width_px");
  p.setpref(fill_stroke_ratio, "fill_stroke_ratio");
  p.setpref(chrlink_endcap, "chrlink_endcap");
  p.setpref(disp_range, "disp_range");
  p.setpref(disp_seqticks, "disp_seqticks");
  p.setpref(disp_source, "disp_source");
  p.setpref(show_gene_marker, "show_gene_marker");
  p.setpref(show_hotspots, "show_hotspots");
  p.setpref(show_gene_name_threshold, "show_gene_name_threshold");
  p.setpref(color_by, "color_by");
  p.setpref(alpha_by, "alpha_by");
  p.setpref(name_by, "name_by");
  p.setpref(gene_name_onoff_method, "gene_name_onoff_method");
  p.setpref(hotspot_by, "hotspot_by");
  p.setpref(font_size_gene, "font_size_gene");
  p.setpref(genewidth, "genewidth");
  p.setpref(genepadding, "genepadding");
  p.setpref(genemargin, "genemargin");
  p.setpref(dot_radius, "dot_radius");
  p.setpref(dot_radius_end, "dot_radius_end");
  p.setpref(huediff_fb, "huediff_fb");

  p.setpref(rift0_col, "rift_color");
  p.setpref(rift0_col_for_all, "rift_col_for_all");
  p.setpref(rift0_use_alt_seq, "rift_use_alt_seq");

  p.setpref(anno_col,  "anno_color");

  p.save(preffile());
}

void mv_status::load_prefs(){
  pref p;

  if(p.load(preffile())){

    bool autosave = true;
    p.getprefa(autosave, "autosave_prefs");

    p.getpref(default_width,      "default_width");
    p.getpref(default_height,     "default_height");
    p.getpref(sequence_thickness, "sequence_thickness");
    p.getpref(orf_thickness,      "orf_thickness");
    p.getpref(saturation, "saturation");
    p.getpref(fwdvalue, "fwdvalue");
    p.getpref(bwdvalue, "bwdvalue");
    p.getpref(bgvalue, "bgvalue");
    p.getpref(bgvalue, "bgvalue");
    p.getpref(alpha_low, "alpha_low");
    p.getpref(alpha, "alpha");
    p.getpref(disp_seqno, "disp_seqno");
    p.getpref(disp_seqticks, "disp_seqticks");
    p.getpref(disp_source, "disp_source");
    p.getpref(font_size_general, "font_size_general");
    p.getpref(min_orf, "min_orf");
    p.getpref(min_size, "min_size");
    p.getpref(max_size, "max_size");
    p.getpref(zoom_factor, "zoom_factor");
    p.getpref(antialias, "antialias");
    p.getpref(pbcache_cap, "pbcache_cap");
    p.getpref(min_thickness, "min_thickness");
    p.getpref(max_thickness, "max_thickness");
    p.getpref(zoomcursor_width_bp, "zoomcursor_width_bp");
    p.getpref(zoomcursor_width_px, "zoomcursor_width_px");
    p.getpref(fill_stroke_ratio, "fill_stroke_ratio");
    p.getpref(chrlink_endcap, "chrlink_endcap");
    p.getpref(disp_range, "disp_range");
    p.getpref(show_gene_marker, "show_gene_marker");
    p.getpref(show_hotspots, "show_hotspots");
    p.getpref(show_gene_name_threshold, "show_gene_name_threshold");
    p.getpref(alpha_by_default, "alpha_by");
    p.getpref(color_by_default, "color_by");
    p.getpref(name_by_default, "name_by");
    p.getpref(gene_name_onoff_method_default, "gene_name_onoff_method");
    p.getpref(hotspot_by_default, "hotspot_by");
    p.getpref(font_size_gene, "font_size_gene");
    p.getpref(genewidth, "genewidth");
    p.getpref(genepadding, "genepadding");
    p.getpref(genemargin, "genemargin");
    p.getpref(dot_radius, "dot_radius");
    p.getpref(dot_radius_end, "dot_radius_end");
    p.getpref(huediff_fb, "huediff_fb");

    vector<double> tempcol; p.getpref(tempcol, "rift_color");
    if(tempcol.size() == 3) rift0_col = tempcol;

    p.getpref(anno_col, "anno_color");
    if(anno_col.size() == 10) anno_col.resize(10);

    p.getpref(rift0_col_for_all, "rift_col_for_all");
    p.getpref(rift0_use_alt_seq, "rift_use_alt_seq");

    p.getprefa(mainwindow_w, "window_width");
    p.getprefa(mainwindow_h, "window_height");

    bool sb_visible = true;
    p.getprefa(sb_visible, "sidebar_visible");
    lsb.set_visible_silent(sb_visible);

    sb_visible = true;
    p.getprefa(sb_visible, "thumbbar_visible");
    rsb.set_visible_silent(sb_visible);

    p.getprefa(show_sidebar_schemes, "sidebar_schemes_visible");
    p.getprefa(show_sidebar_toggles, "sidebar_toggles_visible");

    vector<string> dn, du;
    p.getprefa(dn, "DAS_names");
    p.getprefa(du, "DAS_URLs");
    if(dn.size() == du.size() /* size must match */){  
      das_server_list.clear();
      for(unsigned i=0; i<dn.size(); i++){
	das_server_list.push_back(pair<string,string>(dn[i], du[i]));
      }
      dn.clear();
      du.clear();
    }

//     // fix this!!
//     char **nl, **ul, **nlp, **ulp;
//     gsize useless;
//     GError *e=NULL;
//     nl = g_key_file_get_string_list(prefs, "gmv_auto", "DAS_names", &useless, &e);
//     e = NULL;
//     ul = g_key_file_get_string_list(prefs, "gmv_auto", "DAS_URLs",  &useless, &e);
    
//     if(nl!=NULL){
//       nlp = nl; ulp = ul;
//       das_server_list.clear();
//       while(*nlp!=NULL){
// 	if(*ulp == NULL){ das_server_list.clear(); break; } // broken ~/.gmv
// 	das_server_list.push_back(pair<string,string>(*nlp, *ulp));
// 	nlp++; ulp++;
//       }
//       g_strfreev(nl);
//       g_strfreev(ul);
//     }

    das_prev_server = das_prev_source = das_prev_segment = -1;
    p.getprefa(das_prev_server,  "DAS_server");
    p.getprefa(das_prev_source,  "DAS_source");
    p.getprefa(das_prev_segment, "DAS_segment");

    status.gid.import_prefs(&p);
    status.ins.import_prefs(&p);
  }
}
