using namespace std;

#ifndef _GMV_BASE_H_
#define _GMV_BASE_H_

#include "mainwindow.h"
#include "range_list.h"
#include "confwin.h"
#include "basestat.h"
#include "gmv_defs.h"
#include "pixbuf_cache.h"
#include "mainwindow_handlers.h" 
#include "dialogs.h"
#include "inspector.h"
#include "rvector.h"
#include "segment_tree.h"
#include "histogram.h"

#include <gtk/gtk.h>
#include <map>

typedef pair<annotationTrack*, int> filterTrack;
typedef multimap<int, seqlen_t>::const_iterator hspotIte;

template <typename T> T fixrange(T* now, T min, T max){
  //  assert(cout << "fixrange(): " << *now << ", " << min << ", " << max);
  if (*now > max) *now=max;
  if (*now < min) *now=min;
  return *now;
}

class RGBA {
protected:
  double R, G, B, A;


public:
  RGBA() : 
    R(0), G(0), B(0), A(0) {};
  RGBA(double rr, double gg, double bb, double aa) : 
    R(rr), G(gg), B(bb), A(aa) {};
  RGBA(const vector<double>& v, size_t i) :
    R(v[i]), G(v[i+1]), B(v[i+2]), A(v[i+3]) {};
  
  bool operator==(const RGBA& x) const { 
    return R==x.R && G==x.G && B==x.B && A==x.A;  }
  bool operator<(const RGBA& x) const { 
    return R<x.R && G<x.G && B<x.B && A<x.A;  }

  RGBA& operator=(const RGBA& x){
    R=x.R; G=x.G; B=x.B; A=x.A;
    return *this;
  };

  double r() const { return R; };
  double g() const { return G; };
  double b() const { return B; };
  double a() const { return A; };

  
};

class score {
public:
  void clear();
  void push_back(double);
  void fixate(int histomode);
  bool empty() const { return score_array.empty(); }; 
  double getmin() const { return lowest; };
  double getmax() const { return highest; };
  double getavg() const { return average; };
  void setval(const size_t, const double);
  double at(unsigned i) const { return score_array[i]; };
  double operator[](unsigned i) const { return score_array[i]; };
  bool inrange(unsigned i) const { return histo.inrange(score_array[i]); };

  histogram histo;

private:
  vector <double> score_array;
  double lowest, highest, average;
};

class mv_status;
class results {
 public:
#define JUSTIFY_LEFT 0
#define JUSTIFY_CENTER 1
#define JUSTIFY_RIGHT 2

  basestat* bs[MAXSEQ];
  vector< string > seqfiles;
  vector< string > seqnames;

  score tfidf;
  double tfidf_color_scoremax;

  bool rift_filter_enabled;
  int  rift_filter_policy[MAXSEQ];

  // statistics+
  bool has_detailed_statistics;
  bool has_sequence_statistics;

  // generated by opt_anchor_statistics()
  vector< range_list > anchor_ranges;  
  vector< range_list > anchor_ranges_f;  
  vector< range_list > anchor_ranges_b;  
  int anchors_num_f[MAXSEQ], anchors_num_b[MAXSEQ];

  rvector<annotationTrack*> &annoTracks(int);
  rvector<annotationTrack*> annotations[MAXSEQ];
  bool annotationsChanged;

  int* order_seq;
  seqlen_t max_left_offset;
  seqlen_t seq_offset[MAXSEQ];

  vector<annotationTrack*> filterTracks;

  // segment offset I/O
  vector<seqlen_t> segment_offsets[MAXSEQ];

  // COG
  bool has_cog;
  
  map<string, string> cogs;  // COG-Id : Name
  map<string, string> gene_cog; // Locus tag : COG-Id
  multiset< pair<string, string> > cog_specie; // COG-Id : specie
  map<string, string> gene_specie; // Locus tag : specie
  multimap<string, string> specie_gene;   // specie : locus tag

  void loadcog(string cogpath);

  //
 
  results();
  void clear();
  void chk_order_seq();

  int load_seq_files(bool, bool);
  void load_repeat_file(const string&, int);

  // anchor loader
  void prepare_anchor_segments(int);
  void insert_anchor_row(vector<seqlen_t>&);
  void build_anchor_segment_tree(int, void(*)(double));

  // ------------------------------
  // accessors
  basestat* getbs(int seq) const;         // with seq lookup
  basestat* getbs_direct(int seq) const;  // without seq lookup

  unsigned int acols() const{ return anchors.at(0).size(); }; 
  unsigned int arows() const{ return anchors.size(); }; 

  unsigned int nseqs(bool* toggle_seq) const{ // visible seqs
    unsigned int hidden_seqs = 0;
    for(unsigned int i=0; i<acols()/3; i++)
      if(!toggle_seq[i]) hidden_seqs++;

    return acols()/3-hidden_seqs;
  }

  seqlen_t  maxbases_physical(bool* toggle_seq); // without offset
  seqlen_t  maxbases(bool* toggle_seq);   // with offset included
  void clear_offset();

  void set_offset(int s, seqlen_t offset);
  seqlen_t  get_offset(int s) const;
  seqlen_t  get_offset_t(int s) const;

  void except(int s);
  void restore_except();

  void justify(int);

  bool set_tmurasakix(const string& basedir, const string& seqs);

  // raa: read anchor array
  seqlen_t raa_native(int row, int col) const; // without any offset
  seqlen_t raa_physical(int row, int col) const;  // with segment offset
  seqlen_t raa(int row, int col) const;  // with segment / seq offset

  // rat: read annotation track
  annotation rat(atIte); 
  annotation* rat_native(atIte); // without segment remapping: use with care

  // anc segments
  void lookup_anchor_segments_direct(int seq, seqlen_t a, seqlen_t b, set<int> &res);  // without seq order lookup
  void lookup_anchor_segments(int seq, seqlen_t a, seqlen_t b, set<int> &res); 

  // repeats (hotspots)
  bool has_hotspots() const {return !rep_ptn.empty();};
  seqlen_t hotspot_pattern_len() const { return rep_ptn[0].length(); };
  unsigned hotspot_patterns() const { return rep_ptn.size(); };
  int get_hotspot_by_pattern(unsigned i) const { return rep_p[i]; };
  int get_hotspot_by_combination(unsigned i) const { return rep_c[i]; };

  pair<int, int> hotspot_range_by_pattern(unsigned i) const;
  pair<int, int> hotspot_range_by_combination(unsigned i) const;
  pair<hspotIte, hspotIte> hotspot_pos_lookup(int s, int p) const;
  pair<bool, seqlen_t> get_hotspot_pos(int s, hspotIte i) const;

  bool anchorFilter(int) const; // filter by anchor #
  bool geneFilter(annotationTrack*, int);

  string annotationProfile();
  void annotationProfileImport(string);
  
  string disabledFeatures();
  void disabledFeaturesImport(const string&);
  
  string summarize_in_html();
  string summarize_in_text();

  void add_source(int seq, seq_source so){ sources[seq].push_back(so); };
  size_t sources_size(int seq) const { return sources[seq].size(); };
  seq_source get_source(int seq, unsigned i) const;
  const seq_source* get_source_native(int seq, unsigned i) const { return &(sources[seq].at(i)); };

  void rebuild_segment_offset_index();
  seqlen_t get_segment_offset(int seq, seqlen_t pos) const;
  seqlen_t get_segment_offset(int seq, seqlen_t pos1, seqlen_t pos2) const;


  void sort_sequence_sources(map<int, multimap<seqlen_t, unsigned> >&);
  void sort_sequence_sources(int, vector<int>&, int, bool, bool, void(*)(double));
  
  int atseq_lookup(annotationTrack* t){
    map<annotationTrack*, int>::iterator ite;
    ite = atseqmap.find(t);
    if(ite != atseqmap.end()) return ite->second; // yeah! it's cached!

    // not found. searching...
    for(unsigned s=0; s<acols()/3; s++){
      for(rvector<annotationTrack*>::iterator i=annotations[s].begin_all();
	  i!=annotations[s].end_all(); i++){
	if(*i == t){
	  atseqmap[t]=s;
	  return s;
	}
      }
    }

    return -1;
  }

 protected:
  void relocate_segment_offsets(int, multimap<seqlen_t, unsigned>&);

  vector< vector<seqlen_t> > anchors;
  vector< segment_tree<int> > anchor_segments;

  vector<seqlen_t> ntlen;
  bool is_tmurasakix;
  
  map <annotationTrack*, int> atseqmap;

  // sequence sources / chromosomes
  vector<seq_source> sources[MAXSEQ];
  segment_tree<seqlen_t> segment_offset_index[MAXSEQ];

  // repeats
  vector< string > rep_ptn;
  vector< multimap<int, seqlen_t> > rep_pos; 
  vector<int> rep_c, rep_p; // # of combinations, # of pattern itself
  int max_rep_c, min_rep_c; // repeat combinations
  int max_rep_p, min_rep_p; // repeat points

  // other secret stuff
  seqlen_t bak_max_left_offset;
  seqlen_t bak_seq_offset;
  int excepted;
};

class progress;

void pbar_set_fraction(double);
void pbar_set_fraction(int, double);
void pbar_set_task(double, double); 
void pbar_reset(string);
void pbar_set_threads(unsigned);

class mv_status{
public:
  // ----------------------------------------------------------------------
  // drawing parameters (can be copied)

  // size setting
  int default_width;
  int default_height;
  int min_size;
  int max_size;
  double zoom_factor;

  int max_width_gn;
  int margin_gn;
  int margin_lr;
  int margin_tb;

  // color setting
  bool antialias;
  double alpha_low;
  double alpha;
  double saturation;
  double fwdvalue;
  double bwdvalue;
  double bgvalue;

  double huediff_fb; // Hue difference between forward/backward anchor in dot-plot
  double fill_stroke_ratio;

  // font size
  int font_size_general;
  int font_size_gene;

  // plotting
  double sequence_thickness;
  double orf_thickness;

  double min_thickness;
  double max_thickness;

  bool disp_seqno;
  bool disp_range;
  bool disp_seqticks;
  bool disp_source;

  bool hide_anchored_gene;
  bool hide_unanchored_gene;

  int min_orf;
  
  // plot type
  int plot_type;

  // common color params
  int alpha_by_default;
  int alpha_by;

  // chrom link specific params
  int color_by_default;
  int color_by;
  int chrom_left;
  int chrom_right;
  double basewidth;
  double chrlink_endcap;
  double chrlink_endcap_na;

  vector<double> rift0_col;
  bool rift0_col_for_all;
  bool rift0_use_alt_seq;

  vector<RGBA> anno_col;
  map<string, int> anno_col_map;

  // dot plot specific params
  double dot_radius;
  double dot_radius_end;

  // super zoomer
  seqlen_t superzoom_from;
  seqlen_t superzoom_to;

  // superzoom cursor
  int zoomcursor_width_bp;
  int zoomcursor_width_px;
  int zcwidth;

  // annotation
  int name_by;
  int name_by_default;
  bool show_gene_marker;
  int  gene_name_onoff_method;
  int  gene_name_onoff_method_default;
  int  show_gene_name_threshold;

  // hotspot
  bool show_hotspots;
  int hotspot_by;
  int hotspot_by_default;

  // quickbar stuff
  bool toggle_seq[MAXSEQ];
  int  order_seq[MAXSEQ];

  bool plot_U; // unknown
  bool plot_R; // repeat
  bool plot_O; // ORF

  bool plot_fwd;
  bool plot_bwd;

  bool rift_stencil;
  bool rift_pattern_fill;

  bool clip_hard;

  // and some configurable stuff
  int pbcache_cap;

  // ----------------------------------------------------------------------
  // facilities (should not copied)

  vector<pair<string,string> > das_server_list;
  int das_prev_server, das_prev_source, das_prev_segment;

  bool flushAtRedraw;
  bool loadSeqAtRedraw;
  bool anchorfile_fromGUI;
  string anchorfile;
  string tfidffile;
  string repeatfile;

  string basename;
  string basedir;   // estimated dir where murasaki was launched
  string anchorfile_extension; // ".anchors" or ".block"

  string seqlist;
  // string fasta[MAXSEQ]; // names of fasta files are stored in mr

  bool blocked;

  int argc;
  char **argv;

  GtkWidget* splashwindow;
  GtkWidget* splash_progbar_backup;

  GtkWidget* mainwindow;
  int mainwindow_w;
  int mainwindow_h;

  vector<string> infobar;
  GtkWidget* infobarbox;

  GtkWidget *prog_hpaned;
  GtkWidget *progbar;
  progress *prog;

  GtkWidget *bmark_menu;

  // for command-line configuration
  bool plot_all;
  bool opt_load_seq;
  bool opt_load_annot;

  bool opt_stat_html;
  bool opt_stat_text;

  double opt_min_tfidf;

  vector<string> sequence_names;
  bool cmdopt_forward;
  bool cmdopt_backward;

  bool show_sidebar_schemes;
  bool show_sidebar_toggles;

  // interthread communication 
  GMutex* itc_mutex;

  volatile bool thread_running;
  volatile bool wait_for_disp;

  // multithread rendering mutex
  GMutex* render_mutex;

  // double click timer
  doubleclicker dc;
  int last_dc_x, last_dc_y; // coordinate of last doubleclick
  set<int> overwrap_ancs;
  size_t overwrap_all;
  int dc_counter;

  // anchor statistics
  bool auto_order;

  // annotation
  vector< pair<int,string> > annotation_files;
  vector< pair<int,das_import_result*> > das_tracks;
  string cog_file;

  set<int> in_range_anchors;
  gene_markers_type gene_markers;

  int genewidth;
  int genepadding;
  int genemargin;

  bool justify_by_anchor;

  // for plot area
  bool using_temp_pixbuf;
  GdkPixbuf *temp_pixbuf;
  vector < pair<int,int> > clickable_y;
  gmv_anchor_context_menu acm;

  // for config window
  mv_status *status_copy;
  vector <config_table> confwin_tables;

  // for set_defaults (we have to remember command line opts!)
  mv_status *status_bkup;

  // configuration stack for undo/redo
  vector <pixbuf_cache_tag> undo_stack;
  vector <pixbuf_cache_tag> redo_stack;

  // file output
  bool generate_png;
  bool generate_pdf;
  bool use_x11;

  // update notifier
  GMutex *vercheck_mutex;
  int latest_release;

  // main window stuff
  mainmenu mb;
  left_sidebar lsb;
  thumb_bar rsb;
  justificationStatus js;
  GtkTooltips *tips;

  // dialogs
  inspector ins;
  geneInfoDialog gid;
  statisticsDialog statd;
  reportDialog repd;
  preferenceDialog prefd;
#ifndef OLD_GTK_COMPAT
  das_server_manage_dialog dasd;
#endif

  bool has_any_modal_dialogs() const {
    return  ( ins.ismodal() || gid.ismodal() || statd.ismodal() 
	      || repd.ismodal() || prefd.ismodal() 
	      || mb.has_any_modal_dialogs()
	      || ins.has_any_modal_dialogs()
#ifndef OLD_GTK_COMPAT
	      || dasd.ismodal()
#endif
	      );
  }

  // tip and so on 
  genetip gt;
  gmhilight gmh;

  // some methods
  mv_status(){ 
    rift0_col.resize(3);
    anno_col.resize(10);
    status_bkup=NULL; 

    anno_col_map.insert(pair<string, int>("CDS",           0));
    anno_col_map.insert(pair<string, int>("RBS",           1));
    anno_col_map.insert(pair<string, int>("rRNA",          2));
    anno_col_map.insert(pair<string, int>("tRNA",          3));
    anno_col_map.insert(pair<string, int>("misc_RNA",      4));
    anno_col_map.insert(pair<string, int>("repeat_region", 5));
    anno_col_map.insert(pair<string, int>("repeat_unit",   6));
    anno_col_map.insert(pair<string, int>("stem_loop",     7));
    anno_col_map.insert(pair<string, int>("misc_feature",  8));
    anno_col_map.insert(pair<string, int>("",              9));
  };

  void save_prefs();
  void load_prefs();

  void apply_defaults(){
    if(alpha_by==0) alpha_by = alpha_by_default;
    if(color_by==0) color_by = color_by_default;
    if(name_by==0)  name_by  = name_by_default;
    if(gene_name_onoff_method==0) gene_name_onoff_method = gene_name_onoff_method_default;
    if(hotspot_by==0) hotspot_by = hotspot_by_default;
  }

  void copy(const mv_status &a){
    // size setting
    default_width = a.default_width;
    default_height = a.default_height;
    min_size = a.min_size;
    max_size = a.max_size;
    zoom_factor = a.zoom_factor;

    max_width_gn = a.max_width_gn;
    margin_gn = a.margin_gn;
    margin_lr = a.margin_lr;
    margin_tb = a.margin_tb;

    // color setting
    antialias = a.antialias;
    alpha = a.alpha;
    alpha_low = a.alpha_low;
    saturation = a.saturation;
    fwdvalue = a.fwdvalue;
    bwdvalue = a.bwdvalue;
    bgvalue = a.bgvalue;

    huediff_fb = a.huediff_fb;
    fill_stroke_ratio = a.fill_stroke_ratio;

    rift0_col = a.rift0_col;
    rift0_col_for_all = a.rift0_col_for_all;
    rift0_use_alt_seq = a.rift0_use_alt_seq;

    anno_col = a.anno_col;

    // font size
    font_size_general = a.font_size_general;
    font_size_gene = a.font_size_gene;

    // plotting
    sequence_thickness = a.sequence_thickness;
    orf_thickness = a.orf_thickness;

    min_thickness = a.min_thickness;
    max_thickness = a.max_thickness;

    disp_seqno = a.disp_seqno;
    disp_range = a.disp_range;
    disp_seqticks = a.disp_seqticks;
    disp_source = a.disp_source;

    hide_anchored_gene   = a.hide_anchored_gene;
    hide_unanchored_gene = a.hide_unanchored_gene;

    min_orf = a.min_orf;

    // plot type
    plot_type = a.plot_type;

    // common color params
    alpha_by = a.alpha_by;
  
    // chrom link specific params
    color_by = a.color_by;
    chrom_left = a.chrom_left;
    chrom_right = a.chrom_right;
    basewidth = a.basewidth;
    chrlink_endcap = a.chrlink_endcap;
    chrlink_endcap_na = a.chrlink_endcap_na;

    // dot plot specific params
    dot_radius = a.dot_radius;
    dot_radius_end = a.dot_radius_end;

    // super zoomer
    superzoom_from = a.superzoom_from;
    superzoom_to = a.superzoom_to;

    // superzoom cursor
    zoomcursor_width_bp = a.zoomcursor_width_bp;
    zoomcursor_width_px = a.zoomcursor_width_px;
    zcwidth = a.zcwidth;

    // annotation
    name_by = a.name_by;
    show_gene_marker = a.show_gene_marker;
    show_hotspots = a.show_hotspots;
    gene_name_onoff_method = a.gene_name_onoff_method;
    hotspot_by = a.hotspot_by;
    show_gene_name_threshold = a.show_gene_name_threshold;

    name_by_default = a.name_by_default;
    alpha_by_default = a.alpha_by_default;
    color_by_default = a.color_by_default;
    gene_name_onoff_method_default = a.gene_name_onoff_method_default;
    hotspot_by_default = a.hotspot_by_default;

    genewidth = a.genewidth;
    genepadding = a.genepadding;
    genemargin = a.genemargin;

    // quickbar stuff
    for(int i=0; i<MAXSEQ; i++){
      toggle_seq[i] = a.toggle_seq[i];
      order_seq[i]  = a.order_seq[i];
    }

    plot_U = a.plot_U;
    plot_R = a.plot_R;
    plot_O = a.plot_O;
    
    plot_fwd = a.plot_fwd;
    plot_bwd = a.plot_bwd;

    rift_stencil = a.rift_stencil;
    rift_pattern_fill = a.rift_pattern_fill;

    clip_hard = a.clip_hard;
    
    // and some configurable stuff
    pbcache_cap = a.pbcache_cap;
  };

  string sequenceProfile();

  int seq_lookup(int gs) const;  // geometrycal # (0 origin) to native #
  int seq_rlookup(int ns) const;  // native # (0 origin) to geom #

  void set_tip(GtkWidget* w, const char* s){
    gtk_tooltips_set_tip(tips, w, s, "");
  };

  void set_defaults();
  void check_ranges();
  void chk_toggle_seq();
  
  void backup();
  void restore();

  void zoom_normal();
  void zoom_in();
  void zoom_out();

  void go_fwd();
  void go_bwd();

  void auto_reorder();

  bool set_filenames();
};

// ----------------------------------------------------------------------

class progress {
public:
  progress();

  void set_task(double begin, double end){ b = begin; e = end; };
  void set_taskname(string);

  double fraction(double);
  double fraction(int, double); // for multithread
  double fraction(); 
  string jobtitle();

  void reset(string);
  void reset();           

  void set_threads(unsigned t);

private:
  double f;
  vector<double> fvec;

  GMutex* m;
  bool job_changed;

  string j;

  double b, e;
};

#endif
