#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <iostream>
#include "histogram.h"

histogram::histogram(){
  histo_raw_image = (unsigned char*)malloc(resolution * height * 3);
  state_change_trig = NULL; 
  button_release_trig = NULL;

  ready = false;
  slider_h = NULL;
  histo_pb = NULL;
}

histogram::~histogram(){
  free(histo_raw_image);
}

bool histogram::check_range(GtkWidget* check_by){
  bool violated = false;
  GtkWidget *another = NULL;

  if(check_by==spinbutton_l || check_by==spinbutton_h){
    double l = gtk_spin_button_get_value(GTK_SPIN_BUTTON(spinbutton_l));
    double h = gtk_spin_button_get_value(GTK_SPIN_BUTTON(spinbutton_h));
    if(l>=h){
      another = check_by==spinbutton_l ? spinbutton_h : spinbutton_l;
      double fixed_val = check_by==spinbutton_l ? l+0.001 : h-0.001;
      gtk_spin_button_set_value(GTK_SPIN_BUTTON(another), fixed_val);
      violated = true;
    }
  }

  if(check_by==slider_l || check_by==slider_h){
    double l = gtk_range_get_value(GTK_RANGE(slider_l));
    double h = gtk_range_get_value(GTK_RANGE(slider_h));
    if(l>=h){
      another = check_by==slider_l ? slider_h : slider_l;
      double fixed_val = check_by==slider_l ? l+0.001 : h-0.001;
      gtk_range_set_value(GTK_RANGE(another), fixed_val);
      violated = true;
    }
  }

  if(violated){
    activate(another);
    return false;
  }

  // no range violation found! let's go home.
  threshold_l = gtk_range_get_value(GTK_RANGE(slider_l));
  threshold_h = gtk_range_get_value(GTK_RANGE(slider_h));
  if(is_logmode()) threshold_l = pow(10, threshold_l)-log_geta;
  if(is_logmode()) threshold_h = pow(10, threshold_h)-log_geta;

  return true; 
}

// activate() and release(): signal handlers
void histogram::activate(GtkWidget *w){
  if(w==spinbutton_l || w==spinbutton_h){
    double sb_val = gtk_spin_button_get_value(GTK_SPIN_BUTTON(w));
    gtk_range_set_value( GTK_RANGE( w==spinbutton_l ? slider_l : slider_h ),
			 is_logmode() ? log10(sb_val+log_geta) : sb_val );
  }

  if(w==slider_l || w==slider_h){
    double slider_val = gtk_range_get_value(GTK_RANGE(w));
    gtk_spin_button_set_value( GTK_SPIN_BUTTON( w==slider_l ? spinbutton_l :
						spinbutton_h ), 
			       is_logmode() ? pow(10, slider_val)-log_geta : slider_val);
  }

  if(check_range(w)) fixhisto();
  if(state_change_trig != NULL) state_change_trig->activate(NULL);
}

void histogram::release(GtkWidget *w){
  // value is handled by "value-changed" signal
  if(button_release_trig != NULL) button_release_trig->release(NULL);
}

// return approx count
int histogram::visible(){
  double thl = ( is_logmode() ? log10(threshold_l+log_geta) - log10(bottom+log_geta) :
		 threshold_l - bottom );

  double thh = ( is_logmode() ? log10(threshold_h+log_geta) - log10(bottom+log_geta) :
		 threshold_h - bottom );

  int tl = (int)(thl/step + 0.49*step); // .49 avoids rounds off
  int th = (int)(thh/step);

  int visible = 0;
  for(int x=0; x<resolution; x++)
    if(x>=tl && x<=th) visible += histo_int[x];

  return visible;
}

// fixhisto(): set thresholds / fix histogram image by its thresholds
void histogram::fixhisto(double l, double h){
  threshold_l = l;
  threshold_h = h;

  double thl = is_logmode() ? log10(l+log_geta) : l ;
  double thh = is_logmode() ? log10(h+log_geta) : h ;

  if(slider_h != NULL){
    gtk_range_set_value( GTK_RANGE(slider_l), thl );
    gtk_range_set_value( GTK_RANGE(slider_h), thh );
  }
  fixhisto();
  return;
}

void histogram::fixhisto(){
  double thl = ( is_logmode() ? log10(threshold_l+log_geta) - log10(bottom+log_geta) :
		 threshold_l - bottom );

  double thh = ( is_logmode() ? log10(threshold_h+log_geta) - log10(bottom+log_geta) :
		 threshold_h - bottom );

  int tl = (int)(thl/step + 0.49*step); // .49 avoids rounds off
  int th = (int)(thh/step);

  if(slider_h==NULL) return;

  // re-color
  for(int x=0; x<resolution; x++){
    for(int y=0; y<height; y++){
      int p = (y*resolution + x) * 3;
      if(x<tl || th<x){ // outer: turn everything gray
	if(histo_raw_image[p] == 255){ memset(&histo_raw_image[p], 127, 3); }
	if(histo_raw_image[p] ==   0){ memset(&histo_raw_image[p],  40, 3); }
      } else { // inner: black & white
	if(histo_raw_image[p] == 127){ memset(&histo_raw_image[p], 255, 3); }
	if(histo_raw_image[p] ==  40){ memset(&histo_raw_image[p],   0, 3); }
      }
    }
  }

  // replace image
  GdkPixbuf* histo_newpb;
  histo_newpb = gdk_pixbuf_new_from_data((const guchar*)histo_raw_image, 
					 GDK_COLORSPACE_RGB, 
					 FALSE, 8, resolution, height,
					 resolution*3, NULL, NULL);

  gtk_image_set_from_pixbuf(GTK_IMAGE(histo_image), histo_newpb);
  g_object_unref(histo_pb);
  histo_pb = histo_newpb;
}


// reinit_widgets(): prepare histogram image & widgets
bool histogram::reinit_widgets(){
  if(!ready) return false;

  // image
  histo_pb = gdk_pixbuf_new_from_data((const guchar*)histo_raw_image,
				      GDK_COLORSPACE_RGB, FALSE, 8,
				      resolution, height, resolution*3,
				      NULL, NULL);
  histo_image = gtk_image_new_from_pixbuf(histo_pb);

  // sliders
  double lo = is_logmode() ? log10(bottom+log_geta) : bottom;
  double hi = is_logmode() ? log10(top+log_geta) : top;
  slider_h = gtk_hscale_new_with_range(lo, hi, 0.001);
  slider_l = gtk_hscale_new_with_range(lo, hi, 0.001);
  gtk_scale_set_draw_value(GTK_SCALE(slider_h), false);
  gtk_scale_set_draw_value(GTK_SCALE(slider_l), false);

  g_signal_connect(slider_h, "destroy", G_CALLBACK(gtk_widget_destroyed),
		   &slider_h);

  // scales
  spinbutton_l = gtk_spin_button_new_with_range(bottom, top, 0.001);
  spinbutton_h = gtk_spin_button_new_with_range(bottom, top, 0.001);

  // events
  gtk_widget_add_events(slider_h, GDK_BUTTON_RELEASE_MASK);
  gtk_widget_add_events(slider_l, GDK_BUTTON_RELEASE_MASK);
  gtk_widget_add_events(spinbutton_l, GDK_BUTTON_RELEASE_MASK);
  gtk_widget_add_events(spinbutton_h, GDK_BUTTON_RELEASE_MASK);

  // set slider positions: set before signals are connected!
  gtk_spin_button_set_value( GTK_SPIN_BUTTON(spinbutton_h), threshold_h );
  gtk_spin_button_set_value( GTK_SPIN_BUTTON(spinbutton_l), threshold_l );
  gtk_range_set_value(GTK_RANGE(slider_h),
		      is_logmode() ? log10(threshold_h+log_geta) : threshold_h);
  gtk_range_set_value(GTK_RANGE(slider_l),
		      is_logmode() ? log10(threshold_l+log_geta) : threshold_l);

  // signals
  g_signal_connect(spinbutton_l, "value-changed",
		   G_CALLBACK(handle_activate), (gpointer)this);
  g_signal_connect(slider_l,     "value-changed",
		   G_CALLBACK(handle_activate), (gpointer)this);
  g_signal_connect(spinbutton_h, "value-changed",
		   G_CALLBACK(handle_activate), (gpointer)this);
  g_signal_connect(slider_h,     "value-changed",
		   G_CALLBACK(handle_activate), (gpointer)this);
  g_signal_connect(slider_h, "button-release-event", 
		   G_CALLBACK(handle_release),  (gpointer)this);
  g_signal_connect(slider_l, "button-release-event", 
		   G_CALLBACK(handle_release),  (gpointer)this);
  g_signal_connect(spinbutton_h, "button-release-event", 
		   G_CALLBACK(handle_release),  (gpointer)this);
  g_signal_connect(spinbutton_l, "button-release-event", 
		   G_CALLBACK(handle_release),  (gpointer)this);

  return true;
}

// init(): build internal histogram
void histogram::init(int mode, double hi, double lo, const vector<double>& data){
  // make some initializations
  histomode = mode;

  log_geta = ( lo >= 1 ) ? 0 : 1 - lo;

  bottom = lo * 0.9999;
  top    = hi * 1.0001;

  step = ( is_logmode() ? (log10(top+1 + log_geta) - log10(bottom + log_geta)) :
	   top-bottom ); // hmm, does this work fine?
  step /= (double)resolution;

  // count'em
  histo_int.resize(resolution);
  for(std::vector<int>::iterator i=histo_int.begin();
      i!=histo_int.end(); i++) *i=0;

  for(std::vector<double>::const_iterator i=data.begin();
      i!=data.end(); i++){
    double index;
    index = ( is_logmode() ? ((log10(*i + log_geta) - log10(bottom + log_geta))
			      / step) :
	      ((*i-bottom) / step) );
    if(index>=resolution || index<0){
      cerr << "index=" << index
	   << " score=" << *i << endl; 
    }
    if(index>=resolution){ index=resolution-1; }
    if(index<0){ index=0; };
    histo_int.at((size_t)index)++;
    // FIXME: progress bar
  }
    
  // normalize histogram
  int histomax=0;
  histo.resize(resolution);
  
  for(int x=0; x<resolution; x++)
    if(histomax < histo_int[x]) histomax = histo_int[x]; 

  for(int x=0; x<resolution; x++){
    histo[x] = ( is_logmode() ? ( log10(histo_int[x]) / log10(histomax) ) :
 		 ( (double)histo_int[x] / histomax) );
    histo[x] *= height;
    histo[x] += 0.9999999;
  }
  
  // prepare image
  for(int x=0; x<resolution; x++){
    for(int y=0; y < height; y ++){
      int p = ((height-1-y)*resolution + x)*3;
      if(y<histo[x]) 
	memset(&histo_raw_image[p], 0, 3); 
      else
	memset(&histo_raw_image[p], 255, 3);
    }
  }

  // 

  slider_h = NULL;
  histo_pb = NULL;

  threshold_l = lo;
  threshold_h = hi;
  ready = true;
}
