#ifndef _IZFSTREAM_H_
#define _IZFSTREAM_H_

#ifdef HAS_GZBZ2

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>

#include <iostream>
#include <fstream>
#include <string>
#include <string.h>

#include <zlib.h>

#ifdef HAS_BZ2
# include <bzlib.h>
#else 
# define BZFILE void
#endif

class zstreambuf : public std::streambuf {
private:
  bool readmode, writemode;
  int format;

  std::string fname;
  char fmode;

  int fd;
  BZFILE *bzfd;
  gzFile gzfd;

  int count, bufsize;
  unsigned char buf[10240];
  
  bool active;
  bool eof;

  bool seek_failed;

  // ----------------------------------------------------------------------
  // read to buffer
  bool refill_buffer(){ 
    if(!active) return false;
    if(count != bufsize) return true; // still have data in buffer

    count = 0;
    switch(format){
#ifdef HAS_BZ2
    case bzip2:
      bufsize = BZ2_bzread(bzfd, (void*) buf, 10240); break;
#endif
    case gzip:
      bufsize = gzread(gzfd, (void*) buf, 10240); break;
    case plain:
      bufsize = read(fd, buf, 10240); break;
    }
    if(bufsize <= 0) active = false;

    return active;
  }

  void init(){
    count = bufsize = 0;
    active = false;
    eof = false;
    readmode = writemode = false;
    format = unknown;;
    fd = -1;
    bzfd = gzfd = NULL;
    seek_failed = false;
  };

  bool is_bzip(){ return format==bzip2; }

public:
  static const int unknown = 0;
  static const int plain = 1;
  static const int gzip  = 2;
  static const int bzip2 = 3;

  zstreambuf(){ 
    init();
  };

  bool isactive(){ return active; };

  // ------------------------------------------------------------
  // open, close ...

  void open(const char* f, char mode){
    fname = f;
    fmode = mode;

    int flags = (mode == 'w') ? O_WRONLY | O_CREAT | O_TRUNC : O_RDONLY;

    readmode  = (mode == 'r') ? true : false;
    writemode = (mode == 'w') ? true : false;

    fd = ::open(fname.c_str(), flags, 0644);
    if(fd != -1) active = true; else return;

    // probe the format
    int size = read(fd, buf, 3);
    if (size!=3){ active = false; return; } // failed to probe
    if (buf[0]=='B' && buf[1]=='Z' && buf[2]=='h') format = bzip2;
    else if (buf[0]==0x1f && buf[1]==0x8b) format=gzip;
    else format=plain;

    ::close(fd);

    //    std::cout << (is_bzip ? "bzip\n" : "gzip") << std::endl;

    // [bg]zopen
    char gzmode[3];
    if(writemode) strcpy(gzmode, "wb");
    if(readmode)  strcpy(gzmode, "rb");

    switch(format){
#ifdef HAS_BZ2
    case bzip2:
      bzfd = BZ2_bzopen(fname.c_str(), gzmode); break;
#endif
    case gzip:
      gzfd = gzopen(fname.c_str(), gzmode); break;
    case plain:
      fd = ::open(fname.c_str(), flags, 0644); break;
    }
    
    if(bzfd==NULL && gzfd==NULL && fd==-1){ active=false; return; }
  };

  void close(){
    switch(format){
#ifdef HAS_BZ2
    case bzip2:
      BZ2_bzclose(bzfd); break;
#endif
    case gzip:
      gzclose(gzfd); break;
    case plain:
      ::close(fd); break;
    }
    
    init();
  };


  // ------------------------------------------------------------
  // streambuf input functions

  std::streamsize showmanyc(){
    //    std::cout << "showmanyc()\n";
    return bufsize-count;
  };

  std::streamsize xsgetn(char* s, std::streamsize n){
    //    std::cout << "xsgetn()\n";
    std::streamsize cnt = 0;
    int ch;
    while( (ch = uflow()) != EOF ){
      s[cnt] = ch;
      cnt++;
      if(cnt > n) break;
    }
    return cnt;
  };

  int uflow(){
    //    std::cout << "uflow()\n";
    int c = underflow();
    count++;
    return c;
  };

  int underflow(){
    //    std::cout << "underflow()\n";
    refill_buffer(); 
    if(!active) return EOF;

    if(count<bufsize) return buf[count]; 
    return EOF;
  };

  // ------------------------------------------------------------
  // streambuf output functions (not available yet :p )

  // ------------------------------------------------------------
  // 

  std::streampos lseek(off_t offset, int whence){
    std::streampos pos = -1;

    bool sf = seek_failed; seek_failed = false;
    if(sf && offset==0 && whence==SEEK_CUR){
      return -1;
    }

    bool seeked = false;

    switch(format){
    case plain:
      pos = ::lseek(fd, offset, whence);
      if(!(offset==0 && whence==SEEK_CUR)) seeked = true;
      break;
    case gzip:
      if(whence==SEEK_END) seek_failed = true;
      else{
	pos = gzseek(gzfd, offset, whence);
	if(!(offset==0 && whence==SEEK_CUR)) seeked = true;
      }
      break;
#ifdef HAS_BZ2
    case bzip2:
      if(whence==SEEK_SET && offset==0){ close(); open(fname.c_str(), fmode); }
      else seek_failed = true;
      break;
#endif
    }
    if(seeked){ bufsize=0; count=0; }

    return pos;
  };

};

class izfstream : public std::istream {
private:
  zstreambuf *zsb;
public:
  ~izfstream(){ delete zsb; };
  izfstream() : std::istream(zsb = new zstreambuf){};

  izfstream& seekg(std::streampos pos){
    zsb->lseek(pos, SEEK_SET);
    return *this;
  };

  izfstream& seekg(std::streamoff off, std::ios_base::seekdir dir){
    zsb->lseek(off, 
	       (dir==std::ios_base::beg) ? SEEK_SET :
	       (dir==std::ios_base::cur) ? SEEK_CUR : SEEK_END);
    return *this;
  };

  std::streampos tellg(){ return zsb->lseek(0, SEEK_CUR); };

  void open(const char *f){
    if(zsb->isactive()){ // double open check
      setstate(std::ios_base::failbit); return;}

    clear();
    zsb->open(f, 'r'); 
    if(!zsb->isactive()) setstate(std::ios_base::failbit);
  };
  void close(){ 
    zsb->close(); 
    clear(); // clear error state
  };
};

#ifdef IZFSTREAM_DEBUG
int main(int argc, char* argv[]){
  std::string a;
  
  izfstream s;
  s.open(argv[1]);
  if(!s){
    std::cout << "ah!\n";
  }
  while( s.good() ){
    getline(s, a);
    std::cout << a << std::endl;
  }
  s.close();
  return 0;
}
#endif

#define IFSTREAM izfstream

#else
#  define IFSTREAM ifstream
#endif

#endif
