#include "plotters.h"
#include "gmv.h"
#include "gmv_base.h"
#include "colormap.h"
#include "text.h"
#include "util.h"

#define DOTPLOT_AXIS_THICKNESS 2
#define SOURCE_MARKER_HEIGHT (status.font_size_gene)

#define ANCHOR_FWD 1
#define ANCHOR_BWD 0

void swap(double &a, double &b){
  double c = b;
  b = a;
  a = c;
}

void swap(bool &a, bool &b){
  bool c = b;
  b = a;
  a = c;
}

double get_anc_alpha(unsigned i){
  if(mr.tfidf.empty() || status.alpha_by == ALPHA_BY_CONST) return status.alpha;
  double lo = mr.tfidf.histo.get_low();
  double hi = mr.tfidf.histo.get_high();
  return log10(mr.tfidf[i] - lo) * (status.alpha - status.alpha_low) / log10(hi-lo) + status.alpha_low;
}

void disp_sources(cairo_t* cr, int seq, double basewidth, 
		  double x, double y, double y1, double y2, 
		  double r, double g, double b){
  if(!status.disp_source) return;
  seqlen_t offset = mr.get_offset_t(seq);

  seqlen_t zf = status.plot_type==PLOT_SUPERZOOM ? status.superzoom_from : 0;
  seqlen_t zt = status.plot_type==PLOT_SUPERZOOM ? status.superzoom_to :
    status.plot_type==PLOT_DOT ? mr.getbs(0)->totalbases() : 
    mr.maxbases(status.toggle_seq);

  GRand* dice = g_rand_new();

  for(unsigned i=0; i<mr.sources_size(status.seq_lookup(seq)); i++){
    seq_source src = mr.get_source(status.seq_lookup(seq), i);
    cairo_set_source_rgba(cr, 0, 0, 0, 0.9);

    seqlen_t f = src.begin + offset - zf;
    seqlen_t t = src.end   + offset - zf;

    // rough clip
    if(t<0)     continue;
    if(f>zt-zf) continue;

    // clip
    bool l_clipped = false;
    bool r_clipped = false;

    if(f<0)     { l_clipped=true; f=0; }
    if(t>zt-zf) { r_clipped=true; t=zt-zf; }

    double width = (t-f)*basewidth;
    if(width<0) width=0-width; // because basewidth can be negative

    // draw text
    double posy  = y+THICKNESS_SEQ/2;
    double posx = x+basewidth*(f+t)/2;

    if( width>text_width(cr, FONT_CHR, SOURCE_MARKER_HEIGHT, src.getstr()) )
      draw_text(cr, FONT_CHR, SOURCE_MARKER_HEIGHT, 
		'c','t', posx, posy, src.getstr());
    else if (width>text_width(cr, FONT_CHR, SOURCE_MARKER_HEIGHT, src.getstr_short()))
      draw_text(cr, FONT_CHR, SOURCE_MARKER_HEIGHT, 
		'c','t', posx, posy, src.getstr_short());

      
    // prepare to mark
    int marker_height=SOURCE_MARKER_HEIGHT;
    int marker_width =SOURCE_MARKER_HEIGHT;

    if(basewidth<0) marker_width=0-marker_width;

    if( width < SOURCE_MARKER_HEIGHT*2 ) marker_width/=2;
    if( width < SOURCE_MARKER_HEIGHT   ) marker_width/=2;
    if( width < SOURCE_MARKER_HEIGHT/2 ) marker_width/=2;

    // small contigs are drawn with some probability
    if( width >= 1 || (width < 1 && 
		       g_rand_int_range(dice, 0, (int)(2/width))==1)){
      // draw left marker
      if(!l_clipped){
	cairo_set_source_rgba(cr, r, g, b, 0.5);
	
	// marker
	cairo_set_line_width(cr, 0);
	cairo_move_to(cr, x+basewidth*f, posy);
	cairo_rel_line_to(cr,  marker_width, marker_height/2);
	cairo_rel_line_to(cr, -marker_width, marker_height/2);
	cairo_close_path(cr);
	cairo_fill(cr);

	// line (if required)
	if(y1!=0){
	  cairo_set_line_width(cr, 1);

	  cairo_move_to(cr, x+basewidth*f, y1);
	  cairo_line_to(cr, x+basewidth*f, y2);
	  cairo_stroke(cr);
	}
      }

      if(!r_clipped){
	cairo_set_source_rgba(cr, r, g, b, 0.5);

	// marker
	cairo_set_line_width(cr, 0);
	cairo_move_to(cr, x+basewidth*t, posy);
	cairo_rel_line_to(cr, -marker_width, marker_height/2);
	cairo_rel_line_to(cr,  marker_width, marker_height/2);
	cairo_close_path(cr);
	cairo_fill(cr);

	// line (if required)
	if(y1!=0){
	  cairo_set_line_width(cr, 1);

	  cairo_move_to(cr, x+basewidth*t, y1);
	  cairo_line_to(cr, x+basewidth*t, y2);
	  cairo_stroke(cr);
	}
      }
    }
  }  

  g_rand_free(dice);
}

void disp_sources(cairo_t* cr, int seq, double basewidth, 
		  double x, double y, double y1, double y2){
  disp_sources(cr, seq, basewidth, x, y, y1, y2, 0, 0, 0);
}

void disp_chrom(cairo_t* cr, double seqx1, double seqx2, double seqy, 
		int seqno, double basewidth, seqlen_t ticks_interval,
		void(*setfrac)(double)){
  results *r = &mr;

  int realseqno = status.seq_lookup(seqno);

  seqlen_t from = status.superzoom_from;
  seqlen_t to   = status.superzoom_to;

  double x1 = seqx1;
  double x2 = x1;

  seqlen_t seqhead = mr.get_offset_t(seqno);
  seqlen_t seqlen  = mr.getbs(seqno)->totalbases();

  // boundary check and draw red arrow if out of range
  if ( seqhead + seqlen  < from ){
    cairo_set_source_rgb(cr, 1, 0, 0);
    cairo_set_line_width(cr, 0);
    cairo_move_to(cr, seqx1, seqy);
    cairo_rel_line_to(cr, THICKNESS_CHR, -THICKNESS_CHR/2);
    cairo_rel_line_to(cr, 0, THICKNESS_CHR);
    cairo_close_path(cr);
    cairo_fill(cr);

    cairo_set_line_width(cr, THICKNESS_CHR);
    for(int i=0; i<3; i++){
      cairo_set_source_rgb(cr, 1, 0, 0);
      cairo_move_to(cr, seqx1+(1.25+(i*0.75))*THICKNESS_CHR, seqy);
      cairo_rel_line_to(cr, THICKNESS_CHR*0.5, 0);
      cairo_stroke(cr);
    }
    return;
  }

  if ( to < seqhead ){
    cairo_set_source_rgb(cr, 1, 0, 0);
    cairo_set_line_width(cr, 0);
    cairo_move_to(cr, seqx2, seqy);
    cairo_rel_line_to(cr, -THICKNESS_CHR, -THICKNESS_CHR/2);
    cairo_rel_line_to(cr, 0, THICKNESS_CHR);
    cairo_close_path(cr);
    cairo_fill(cr);

    cairo_set_line_width(cr, THICKNESS_CHR);
    for(int i=0; i<3; i++){
      cairo_set_source_rgb(cr, 1, 0, 0);
      cairo_move_to(cr, seqx2-(1.25+(i*0.75))*THICKNESS_CHR, seqy);
      cairo_rel_line_to(cr,    -THICKNESS_CHR*0.5, 0);
      cairo_stroke(cr);
    }

    return;
  }

  // determine position
  if(from < seqhead)
    x1 += (seqhead-from)*basewidth;

  if(seqlen+seqhead < to)
    x2 = seqx1 + basewidth*(seqlen+seqhead-from);
  else
    x2 = seqx1 + basewidth*(to-from);

  // sequence body
  cairo_set_source_rgb(cr, 0, 0, 0);

  cairo_move_to (cr, x1, seqy);
  cairo_line_to (cr, x2, seqy);
  cairo_set_line_width (cr, THICKNESS_CHR);
  cairo_stroke (cr);


  // small sequence ticks
  if(status.disp_seqticks){
    string prev_label = "";
    seqlen_t prev_pos = 0;
    for(seqlen_t i=mr.get_offset_t(seqno);
	i<=   mr.get_offset_t(seqno) + mr.getbs(seqno)->totalbases(); 
	i+=ticks_interval){
      seqlen_t j = i-mr.get_offset_t(seqno); 
      seqlen_t k = i-status.superzoom_from;  

      if ( i < status.superzoom_from || status.superzoom_to < i ) continue;
      if ( j < 0 ) continue;

      double tick_offset = - THICKNESS_SEQ/2;

      cairo_set_source_rgb(cr, 1, 0, 0);
      cairo_set_line_width(cr, 0);
      cairo_move_to(cr, seqx1+k*basewidth, seqy + tick_offset);
      cairo_rel_line_to(cr, -3, -6);
      cairo_rel_line_to(cr, 6, 0);
      cairo_close_path(cr);
      //    cairo_stroke(cr);
      cairo_fill(cr);

      cairo_set_source_rgb(cr, 0, 0, 0);
      draw_text(cr, FONT_CHR, 8, 'l', 'B', 
		seqx1+k*basewidth+4, seqy + tick_offset - 1,
		ticklabel(j, prev_label, prev_pos).c_str());
    }
  }

  assert(cout << "disp_chrom(): from " << from << " to " << to <<endl);
  // unknown region (gray)
  if(status.plot_U){
    //    assert(cout << "disp_chrom(): Plotting unknown regions on seq#" << seqno << endl);
    cairo_set_source_rgb(cr, 0.5, 0.5, 0.5);
    cairo_set_line_width (cr, status.plot_R ? THICKNESS_CHR/2 : THICKNESS_CHR);
    seqlen_t chr_offset = mr.get_offset_t(seqno);
    double offset = status.plot_R ? THICKNESS_CHR/4 : 0;
    for (int j=0; j<(int)r->getbs(seqno)->us.size(); j++){
      seqlen_t start = r->getbs(seqno)->us[j];
      seqlen_t end   = r->getbs(seqno)->ue[j]+1;

      seqlen_t seg_offset = mr.get_segment_offset(realseqno, start, end);
      start += seg_offset;
      end   += seg_offset;

      start+= chr_offset;
      end  += chr_offset;

      if (end <= from || to <= start) continue;
      if (start < from) start = from;
      if (to < end) end = to;

      cairo_move_to(cr, seqx1 + (start-from)*basewidth, seqy + offset);
      cairo_rel_line_to(cr, (end-start)*basewidth, 0);
      cairo_stroke(cr);

      if(setfrac!=NULL) setfrac( ((double)j/(2*r->getbs(seqno)->us.size())) );
    }
  }

  // repeat region (light blue)
  if(status.plot_R){
    //    assert(cout << "disp_chrom(): Plotting repeat regions on seq#" << seqno << endl);
    cairo_set_source_rgb(cr, 0.7, 0.7, 1);
    cairo_set_line_width (cr, status.plot_U ? THICKNESS_CHR/2 : THICKNESS_CHR);
    seqlen_t chr_offset = mr.get_offset_t(seqno);
    double offset = status.plot_U ? -THICKNESS_CHR/4 : 0;
    for (int j=0; j<(int)r->getbs(seqno)->rs.size(); j++){
      seqlen_t start = r->getbs(seqno)->rs[j];
      seqlen_t end   = r->getbs(seqno)->re[j]+1;

      seqlen_t seg_offset = mr.get_segment_offset(realseqno, start, end);
      start += seg_offset;
      end   += seg_offset;

      start+= chr_offset;
      end  += chr_offset;

      if (end <= from || to <= start) continue;
      if (start < from) start = from;
      if (to < end) end = to;

      cairo_move_to(cr, seqx1 + (start-from)*basewidth, seqy + offset);
      cairo_rel_line_to(cr, (end-start)*basewidth, 0);
      cairo_stroke(cr);

      if(setfrac!=NULL) setfrac( ((double)j/(2*r->getbs(seqno)->rs.size())) + 0.33);
    }
  }

  // ORF markers (RGB)
  if(status.plot_O){
    double offset = THICKNESS_CHR/2 + THICKNESS_ORF; // offset to top / bottom

    // ORF background (gray)
    cairo_set_source_rgb(cr, 0.7, 0.7, 0.7);
    cairo_set_line_width(cr, THICKNESS_ORF);

    cairo_move_to(cr, x1, seqy - offset + THICKNESS_ORF/2);
    cairo_line_to(cr, x2, seqy - offset + THICKNESS_ORF/2);
    cairo_stroke(cr);

    cairo_move_to(cr, x1, seqy + offset - THICKNESS_ORF/2);
    cairo_line_to(cr, x2, seqy + offset - THICKNESS_ORF/2);
    cairo_stroke(cr);

    // and ORFs
    seqlen_t chr_offset = mr.get_offset_t(seqno);

    for (int f=0; f<3; f++) {
      int sfi, sri, efi, eri;

      // color by frame
      cairo_set_source_rgb(cr, 
			   (f==0) ? 0.9 : 0.1, 
			   (f==1) ? 0.9 : 0.1,
			   (f==2) ? 0.9 : 0.1 );
      double orfmark_width = THICKNESS_ORF/3;

      // forward ORFs
      efi=0;
      for(sfi=0; sfi<(int)r->getbs(seqno)->startF[f].size(); sfi++){
	seqlen_t start = r->getbs(seqno)->startF[f].at(sfi);
	if (to <= start+chr_offset) break; // no more ORFs on screen

	seqlen_t end = 0;
	for(efi=efi; efi < (int)r->getbs(seqno)->stopF[f].size(); efi++){
	  end = r->getbs(seqno)->stopF[f].at(efi);
	  if(start < end) break;
	}
	
	if (end-start <= (seqlen_t)status.min_orf) continue;

	seqlen_t seg_offset = mr.get_segment_offset(realseqno, start, end);
	start += seg_offset;
	end   += seg_offset;

	start+= chr_offset;
	end  += chr_offset;
	
	if (end <= from) continue;
	if (start < from) start = from;
	if (to < end) end = to;

	if( abs(end-start)*basewidth < orfmark_width ){
	  // ORF is shorter than width
	  cairo_set_line_width(cr, orfmark_width);
	  cairo_move_to(cr, seqx1 + (start-from)*basewidth, 
			seqy - offset + f*orfmark_width + orfmark_width/2);
	  cairo_rel_line_to(cr, (end-start)*basewidth,  0);
	  cairo_stroke(cr);
	} else {
	  // long ORF
	  cairo_set_line_width(cr, 0.001);
	  cairo_move_to (cr, seqx1 + (start-from)*basewidth, 
			 seqy - offset + f*orfmark_width);
	  cairo_rel_line_to(cr, 0, orfmark_width);
	  cairo_rel_line_to(cr, (end-start)*basewidth, 0);
	  cairo_rel_line_to(cr,  -orfmark_width, -orfmark_width);
	  cairo_close_path(cr);
	  cairo_fill(cr);
	}

	if(setfrac!=NULL) setfrac( f*0.11 
				   + 0.055*((double)sfi/r->getbs(seqno)->startF[f].size())
				   + 0.66 );
	
      }

      // backward ORFs
      eri=(int)r->getbs(seqno)->stopR[f].size()-1;
      int sriR=0;
      for(sri=r->getbs(seqno)->startR[f].size()-1; sri>=0; sri--){
	sriR++;
	seqlen_t start = r->getbs(seqno)->startR[f].at(sri);
	if (start+chr_offset <= from) break; // no more ORFs on screen

	seqlen_t end = 0;
	for(eri=eri; eri>=0 ; eri--){
	  end = r->getbs(seqno)->stopR[f].at(eri);
	  if(start > end) break;
	}
	
	if (start-end <= (seqlen_t)status.min_orf) continue;

	start+= chr_offset;
	end  += chr_offset;
	
	if (to <= end) continue;
	if (end < from) end = from;
	if (to < start) start = to;

	if( abs(end-start)*basewidth < orfmark_width ){
	  // ORF is short
	  cairo_set_line_width(cr, orfmark_width);
	  cairo_move_to (cr, seqx1 + (start-from)*basewidth, 
			 seqy + offset - f*orfmark_width - orfmark_width/2);
	  cairo_rel_line_to(cr, (end-start)*basewidth, 0);
	  cairo_stroke(cr);
	} else {
	  // ORF is enough long
	  cairo_set_line_width(cr, 0.001);
	  cairo_move_to (cr, seqx1 + (start-from)*basewidth, 
			 seqy + offset - f*orfmark_width);
	  cairo_rel_line_to(cr, 0, -orfmark_width);
	  cairo_rel_line_to(cr, (end-start)*basewidth + orfmark_width, 0);
	  cairo_rel_line_to(cr,  -orfmark_width, orfmark_width);
	  cairo_close_path(cr);
	  cairo_fill(cr);
	}

	if(setfrac!=NULL) setfrac( f*0.11 
				   + 0.055*((double)sriR/r->getbs(seqno)->startR[f].size()) 
				   + 0.66 );
      }
    }
  }
}

void draw_dot_plot(cairo_t *cr, void* useless_pointer){
  mkinterval itv;
  results *r = &mr;

  status.superzoom_from = 0;
  status.superzoom_to = r->maxbases(status.toggle_seq); 

  // determine first seq
  int seqno = mr.nseqs(status.toggle_seq);

  mr.except(0);

  // determine size of plot area
  int margin_t = ( status.margin_tb );
  int margin_b = (int)( status.margin_tb + text_height(cr, FONT_CHR, "100Mb") 
			+ THICKNESS_SEQ + status.margin_gn );
  int margin_r = ( status.margin_lr );
  int margin_l = (int)( status.margin_lr + text_height(cr, FONT_CHR, "100Mb")
			+ (seqno-2)*SOURCE_MARKER_HEIGHT
			+ (seqno-1)*THICKNESS_SEQ + status.margin_gn 
			+ (seqno-2)*MARGIN_SEQ );

  double basewidth   = (double)(width()  - margin_l - margin_r) / mr.getbs(0)->totalbases();
  double baseheight  = (double)(height() - margin_t - margin_b) / mr.maxbases(status.toggle_seq);

  // draw X ticks

  seqlen_t ticks_interval = get_ticks_interval(r->getbs(0)->totalbases());

  string prev_label = "";
  seqlen_t prev_pos = 0;
  for(int i=0; i<=r->getbs(0)->totalbases()/ticks_interval; i++){
    cairo_set_source_rgb(cr, 0.5, 0.5, 0.5);
    cairo_move_to(cr, margin_l+i*ticks_interval*basewidth, height()-margin_b-TICKS_OFFSET);
    cairo_line_to(cr, margin_l+i*ticks_interval*basewidth, height()-margin_b+TICKS_OFFSET);
    cairo_set_line_width(cr, THICKNESS_TICK);
    cairo_stroke(cr);

    cairo_set_source_rgb(cr, 0, 0, 0);

    draw_text(cr, FONT_CHR, 'c', 'c', 
	      margin_l + i*ticks_interval*basewidth, height()-margin_b+TICKS_OFFSET, 
	      ticklabel(ticks_interval*i, prev_label, prev_pos).c_str());
  }

  // draw Y ticks
  ticks_interval = get_ticks_interval(mr.maxbases(status.toggle_seq));

  prev_label = "";
  prev_pos = 0;
  for(int i=0; i<=mr.maxbases(status.toggle_seq)/ticks_interval; i++){
    cairo_set_source_rgb(cr, 0.5, 0.5, 0.5);
    cairo_move_to(cr, margin_l-TICKS_OFFSET, height()-margin_b-i*ticks_interval*baseheight);
    cairo_line_to(cr, margin_l+TICKS_OFFSET, height()-margin_b-i*ticks_interval*baseheight);
    cairo_set_line_width(cr, THICKNESS_TICK);
    cairo_stroke(cr);

    cairo_save(cr);
    cairo_rotate(cr, M_PI/2);
    cairo_set_source_rgb(cr, 0, 0, 0);

    draw_text(cr, FONT_CHR, 'c', 'c', 
	      height()-margin_b-i*ticks_interval*baseheight, TICKS_OFFSET-margin_l,
	      ticklabel(ticks_interval*i, prev_label, prev_pos).c_str());
    cairo_restore(cr);
  }

  // draw X axis
  cairo_set_source_rgb(cr, 0, 0, 0);
  cairo_move_to(cr, margin_l, height()-margin_b);
  cairo_line_to(cr, width()-margin_r, height()-margin_b);
  cairo_set_line_width(cr, DOTPLOT_AXIS_THICKNESS);
  cairo_stroke(cr);

  // draw Y axis
  cairo_set_source_rgb(cr, 0, 0, 0);
  cairo_move_to(cr, margin_l, margin_t);
  cairo_line_to(cr, margin_l, height()-margin_b);
  cairo_set_line_width(cr, DOTPLOT_AXIS_THICKNESS);
  cairo_stroke(cr);

  // ------------------------------
  // color for anchors
  vector<double> fr, fg, fb;
  vector<double> br, bg, bb;

  for(int i=0; i<seqno-1; i++){
    double hue= (double)(300-status.huediff_fb)*i/seqno;
    double r, g, b;

    // forward colors
    HSVtoRGB( &r, &g, &b, hue, status.saturation, status.fwdvalue );
    fr.push_back(r);
    fg.push_back(g);
    fb.push_back(b);

    // backward colors
    HSVtoRGB( &r, &g, &b, hue+status.huediff_fb,  status.saturation, status.bwdvalue );
    br.push_back(r);
    bg.push_back(g);
    bb.push_back(b);
  }
  // ------------------------------
  // plot chromosomes
  pbar_reset("Plotting chromosomes...");

  for(int i=0; i<seqno; i++){
    //    assert(cout << "draw_dot_plot(): Plotting seq#" << i << endl);

    // fraction_* is about progress bar
    double fraction_seq  =  ((double)1/seqno);
    double fraction_zero =  fraction_seq * i;   // overall at 0% for this seq
    pbar_set_task(fraction_zero, fraction_zero + fraction_seq);

    if(i==0){
      double seqy  = height() - status.margin_tb;
      double seqx1 = margin_l; 
      double seqx2 = width()-margin_r;

      disp_chrom(cr, seqx1, seqx2, seqy, i, basewidth, 
		 ticks_interval, pbar_set_fraction);
      
      disp_sources(cr, i, basewidth, seqx1, seqy, 
		   status.margin_tb, height()-margin_b);

      // chrom name
      cairo_set_source_rgb(cr, 1, 1, 1);
      string chr_name;
      if(status.disp_seqno) chr_name=("#" + itos(status.seq_lookup(i)+1)+ ": ");
      chr_name = chr_name + r->getbs(i)->name;

      int fs = FONT_SIZE;
      if ( fs > THICKNESS_SEQ ) fs=(int)THICKNESS_SEQ;
      draw_text(cr, FONT_CHR, fs, 'c', 'c', 
		(width()-margin_r + margin_l)/2, seqy,
		chr_name.c_str());
    } else {
      cairo_save(cr);
      cairo_rotate(cr, M_PI/2);

      double seqx1 = height()-margin_b;
      double seqx2 = margin_b;
      double seqy = -status.margin_lr - (SOURCE_MARKER_HEIGHT+THICKNESS_SEQ+MARGIN_SEQ)*(i-1);

      // chromosome band
      disp_chrom(cr, seqx1, seqx2, seqy, i, -baseheight, 
		 ticks_interval, pbar_set_fraction);

      disp_sources(cr, i, -baseheight, seqx1, seqy, 
		   -width()+status.margin_lr, -margin_l,
		   fr[i-1], fg[i-1], fb[i-1]);

      // chromosome name
      cairo_set_source_rgb(cr, 1, 1, 1);
      string chr_name;
      if(status.disp_seqno) chr_name=("#" + itos(status.seq_lookup(i)+1)+ ": ");
      chr_name = chr_name + r->getbs(i)->name;

      int fs = FONT_SIZE;
      if ( fs > THICKNESS_SEQ ) fs=(int)THICKNESS_SEQ;
      draw_text(cr, FONT_CHR, fs, 'c', 'c', 
		(height()-margin_b + margin_t)/2, seqy,
		chr_name.c_str());

      // color label
      cairo_set_source_rgb(cr, fr[i-1], fg[i-1], fb[i-1]); 
      draw_text(cr, FONT_CHR, 'r', 'c', 
		height()-margin_b + text_width(cr, FONT_CHR, "F "),  seqy,
		"F");

      cairo_set_source_rgb(cr, br[i-1], bg[i-1], bb[i-1]); 
      draw_text(cr, FONT_CHR, 'l', 'c', 
		height()-margin_b + text_width(cr, FONT_CHR, "F "),  seqy,
		" B");

      cairo_restore(cr);
    }
  }

  // ------------------------------
  // plot anchors
  //  assert(cout << "draw_dot_plot(): Plotting anchors..." << endl);
  pbar_reset("Plotting anchors...");

  itv.setval(r->arows());
  for (unsigned int i=0; i<r->arows(); i++){
    if (!mr.anchorFilter(i)) continue;
    double alpha = get_anc_alpha(i);

    for (unsigned int col=0; col<mr.nseqs(status.toggle_seq)*3-3; col+=3){
      int s=col/3; 

      if ( !status.plot_fwd && r->raa(i,col+5)==ANCHOR_FWD ) continue;
      if ( !status.plot_bwd && r->raa(i,col+5)==ANCHOR_BWD ) continue;

      double offset = basewidth/2;

      seqlen_t x1 = r->raa_physical(i,0);
      seqlen_t x2 = r->raa_physical(i,1);
      seqlen_t y1 = r->raa(i,col+3);
      seqlen_t y2 = r->raa(i,col+4);
      if(x1==x2 || y1==y2) continue;

      double xseq_s = (double)x1 * basewidth + margin_l + offset;
      double xseq_e = (double)x2 * basewidth + margin_l + offset;
      double yseq_s = -(double)y1 * baseheight + height()-margin_b - offset;
      double yseq_e = -(double)y2 * baseheight + height()-margin_b - offset ;
      int xseq_d = r->raa(i, 2);
      int yseq_d = r->raa(i, col+5);   // 1:+ 0:-

      if(yseq_d!=xseq_d){
	// reverse
	cairo_set_source_rgba(cr, br[s], bg[s], bb[s], alpha); 
	swap(yseq_s, yseq_e);
      } else
	cairo_set_source_rgba(cr, fr[s], fg[s], fb[s], alpha);

      cairo_move_to(cr, xseq_s, yseq_s);
      cairo_line_to(cr, xseq_e, yseq_e);
      cairo_set_line_width(cr, 1);
      cairo_stroke(cr);

      if(status.dot_radius != 0){
	cairo_arc(cr, xseq_s, yseq_s, status.dot_radius, 0, 2*M_PI);
	cairo_set_line_width(cr, 1);
	cairo_stroke(cr);
      }

      if(status.dot_radius_end != 0){
	cairo_arc(cr, xseq_e, yseq_e, status.dot_radius_end, 0, 2*M_PI);
	cairo_set_line_width(cr, 1);
	cairo_stroke(cr);
      }
    }
    if(itv.check(i)) pbar_set_fraction( (double)i/r->arows() );
  }

  mr.restore_except();
}

double endcap(bool last, bool upper, double seqdist){
  if( upper && !last) return(status.chrlink_endcap_na * seqdist);
  return(status.chrlink_endcap * seqdist);
}

// ======================================================================
// CHROM LINKS

// ----------------------------------------------------------------------
// multithreaded anchor plotter

void draw_chrom_links_anchors(draw_chrom_links_anchors_t* params){
  cairo_t* cr = params->cr;
  seqlen_t from = params->from;
  seqlen_t to = params->to;
  double basewidth = params->basewidth;
  int thread_id = params->thread_id;
  int threads = params->threads;
  unsigned colorbase = params->colorbase;
  int margin_t = params->margin_t;
  double max_label_width = params->max_label_width;
  double seqdist = params->seqdist;
  seqlen_t minlen_anchor = params->minlen_anchor;
  seqlen_t maxlen_anchor = params->maxlen_anchor;

  mkinterval itv;

  unsigned anc_per_thread = mr.arows()/threads;
  unsigned anc_from = thread_id * anc_per_thread;
  unsigned anc_to = ( thread_id == threads-1 ) ? mr.arows() : 
    (thread_id+1) * anc_per_thread;

  assert(cout << "thread #" << thread_id << "/" << threads << " plot anc #" << anc_from << "-" << anc_to << "\n");

  // clip extents
  double c_x1, c_x2, c_y1, c_y2;
  cairo_clip_extents(cr, &c_x1, &c_y1, &c_x2, &c_y2);

  itv.setval(anc_per_thread);
  for (unsigned i=anc_from; i<anc_to; i++){
    double fr, fg, fb, br, bg, bb;
    double line_r, line_g, line_b;
    double fill_r, fill_g, fill_b;
    double hue = 240;

    // tf-idf filter
    if (!mr.anchorFilter(i)) continue;

    // may want to find alternative sequence for color assignment
    int alt_colorbase = colorbase;
    if( mr.raa_physical(i,colorbase*3) == mr.raa_physical(i,colorbase*3+1) &&
	status.rift0_use_alt_seq ){
      for (unsigned col=0; col<mr.acols(); col++){
	if(mr.raa_physical(i,col*3) != mr.raa_physical(i,col*3+1)){
	  alt_colorbase = col;
	  break;
	}
      }
    }

    // color
    switch(status.color_by){
    case 0:
    case COLOR_BY_POSITION:
      hue *= abs((mr.raa_physical(i,alt_colorbase*3) +
		  mr.raa_physical(i,alt_colorbase*3+1))/2); 
      hue /= mr.getbs(alt_colorbase)->totalbases();
      break;
    case COLOR_BY_ANCHORLEN:
      hue *= ((double)(abs(mr.raa_physical(i, alt_colorbase*3+1)-
			   mr.raa_physical(i, alt_colorbase*3))-minlen_anchor)/
	      (maxlen_anchor-minlen_anchor));
      assert(cout << "max-min anchor "<<maxlen_anchor-minlen_anchor << endl);
      //      if(hue> 200) cout << hue << " ";
      hue = 240-hue;
      break;
    case COLOR_BY_TFIDF:
      double dist=(log10(mr.tfidf_color_scoremax) - log10(mr.tfidf[i]));
      if (dist<0) dist=0;
      hue *= (dist / (0.00000000000001+log10(mr.tfidf_color_scoremax)-log10(mr.tfidf.getmin())));
      break;
    }

    HSVtoRGB( &fr, &fg, &fb, hue,  status.saturation, status.fwdvalue );
    HSVtoRGB( &br, &bg, &bb, hue,  status.saturation, status.bwdvalue );
    double alpha = get_anc_alpha(i);

    bool all_in_range = true;
    int visible_seqs = mr.nseqs(status.toggle_seq);

    bool has_rifts = false;
    for (int col=0; col<(visible_seqs-1)*3; col+=3)
      if(mr.raa(i, col) == mr.raa(i, col+1)){ has_rifts = true; break; }

    for (int col=0; col<(visible_seqs-1)*3; col+=3){

      // scan next target line
      int nextcol = col+3;
      
      if (nextcol>=visible_seqs*3 || nextcol<0) continue;
	  
      // ------------------------------
      // target columns & positions

      double seqx1 = status.margin_lr+max_label_width;

      seqlen_t seqA1 = abs(mr.raa(i,col)) - from;       // seq A begin
      seqlen_t seqA2 = abs(mr.raa(i,col+1)) - from;     // seq A end

      seqlen_t seqB1 = abs(mr.raa(i,nextcol)) - from;   // seq B begin
      seqlen_t seqB2 = abs(mr.raa(i,nextcol+1)) - from; // seq B end
	  
      bool seqA_skip = (seqA1 == seqA2);
      bool seqB_skip = (seqB1 == seqB2);
      bool need_stencil = seqB_skip;
      vector<double> stencil;

      bool upper_island_only = false;
      bool lower_island_only = false;

      if(seqA_skip){
	int prevcol = col-3;
	while(prevcol >= 0 && seqA_skip){
	  seqA_skip = (mr.raa(i,prevcol) == mr.raa(i,prevcol+1));
	  prevcol -= 3;
	}
	if(seqA_skip && mr.raa(i,nextcol)!=mr.raa(i,nextcol+1)) 
	  upper_island_only = true;
	else continue;
      }

      while(seqB_skip){
	nextcol += 3;
	if(nextcol>=visible_seqs*3) break;
	seqB1 = abs(mr.raa(i,nextcol)) - from;  
	seqB2 = abs(mr.raa(i,nextcol+1)) - from;
	seqB_skip = (seqB1 == seqB2);
      }
      if(seqB_skip) lower_island_only = true;
      
      // ------------------------------
      // meanA/B for rough clipping & drawing lines 
      seqlen_t range = to-from;
      double meanA = (seqA1+seqA2)/2;
      double meanB = (seqB1+seqB2)/2;

      all_in_range = (all_in_range && 
		      (( meanA >= 0 && meanA <= range &&
			 meanB >= 0 && meanB <= range) ||
		       ( meanA >= 0 && meanA <= range && seqB1 == seqB2) ||
		       ( meanB >= 0 && meanB <= range && seqA1 == seqA2)));

      // ------------------------------
      // direction
      bool isForward = (! (upper_island_only||lower_island_only)) ? 
	( mr.raa(i,col+2) == mr.raa(i,nextcol+2) ) : true;

      // skip if [forward|backward] only
      if ( !status.plot_fwd && mr.raa(i,nextcol+2)==1 ) continue;
      if ( !status.plot_bwd && mr.raa(i,nextcol+2)==0 ) continue;
     
      // color for forward / backward
      if(status.color_by == 0 ||  status.color_by == COLOR_BY_POSITION){
	if( (upper_island_only && !status.rift0_use_alt_seq) ||
	    (has_rifts && status.rift0_col_for_all) ){ 
	  // default color for rift on seq 0 
	  fr=br=status.rift0_col[0];
	  fg=bg=status.rift0_col[1];
	  fb=bb=status.rift0_col[2];
	}
      }
      line_r = isForward ? fr : br; fill_r = line_r*status.fill_stroke_ratio;
      line_g = isForward ? fg : bg; fill_g = line_g*status.fill_stroke_ratio;
      line_b = isForward ? fb : bb; fill_b = line_b*status.fill_stroke_ratio;
      
      // adjust position
      if( !isForward ) swap(seqB1, seqB2);

      if(seqA1<seqA2) seqA2++; else seqA1++;
      if(seqB1<seqB2) seqB2++; else seqB1++;


      // ------------------------------
      // range & clipping stuff

      // rough clip (mandatory: skip if everything is out of viewport)
      if( (seqA1<=0 && seqA2<=0 && seqB1<=0 && seqB2<=0) ||
	    (seqA1>=range && seqA2>=range && seqB1>=range && seqB2>=range) ) 
	continue;

      // hard clipping (optional: skip if the line just go across viewport)
      if ( status.clip_hard && 
	   ( (seqA1<=0 && seqA2<=0 && seqB1>=range && seqB2>=range) ||
	     (seqB1<=0 && seqB2<=0 && seqA1>=range && seqA2>=range) ) )
	continue;

      // ------------------------------
      // vertical position

      // basic calculations
      bool lastseq = (nextcol/3 == (int)mr.nseqs(status.toggle_seq)-1);
      double endcap_from = endcap(lastseq, false, seqdist);
      double endcap_to   = endcap(lastseq, true,  seqdist);

      endcap_from += THICKNESS_SEQ/2;
      endcap_to   += THICKNESS_SEQ/2;

      double seqy       = margin_t+seqdist*(col/3);
      double fromY_base = seqy + THICKNESS_SEQ/2;
      double fromY      = seqy + endcap_from;
      double toY        = seqy + (nextcol-col)/3*seqdist - endcap_to;
      double toY_base   = seqy + (nextcol-col)/3*seqdist - THICKNESS_SEQ/2;

//       cout << "Ypos: " 
// 	   << fromY_base <<", " 
// 	   << fromY <<", "
// 	   << toY   <<", "
// 	   << toY_base <<endl;

      // stencil stuff
      stencil.clear();
      for(int c=col+3; c<nextcol; c+=3){
	stencil.push_back(seqy + ((c-col)/3-status.chrlink_endcap_na)*seqdist);
	stencil.push_back(seqy + ((c-col)/3+status.chrlink_endcap)*seqdist);
      }

      cairo_pattern_t* line_pattern = NULL;
      cairo_pattern_t* fill_pattern = NULL;

      double linewidth = basewidth * abs(seqA1 - seqA2);

      if(need_stencil || upper_island_only || lower_island_only){
	cairo_save(cr);
	cairo_reset_clip(cr);

	// create fill pattern
	if (linewidth >= status.max_thickness){
	  if(status.rift_pattern_fill){
	    if(!upper_island_only && !lower_island_only){
	      // horizontal stripe
	      cairo_push_group(cr);
	      cairo_set_source_rgba(cr, fill_r, fill_g, fill_b, alpha);
	      cairo_set_line_width(cr, 2);
	      for(double l=c_y1; l<c_y2; l+=4){
		cairo_move_to(cr, c_x1, l); cairo_line_to(cr, c_x2, l); }
	      cairo_stroke(cr);
	      fill_pattern = cairo_pop_group(cr);

	      cairo_push_group(cr);
	      cairo_set_source_rgba(cr, line_r, line_g, line_b, alpha);
	      cairo_set_line_width(cr, 1);
	      for(double l=c_y1+2; l<c_y2; l+=4){
		cairo_move_to(cr, c_x1, l); cairo_line_to(cr, c_x2, l); }
	      cairo_stroke(cr);
	      line_pattern = cairo_pop_group(cr);
	    } else {
	      // vertical stripe
	      cairo_push_group(cr);
	      cairo_set_source_rgba(cr, fill_r, fill_g, fill_b, alpha);
	      cairo_set_line_width(cr, 2);
	      for(double l=c_x1; l<c_x2; l+=4){
		cairo_move_to(cr, l, c_y1); cairo_line_to(cr, l, c_y2); }
	      cairo_stroke(cr);
	      fill_pattern = cairo_pop_group(cr);

	      cairo_push_group(cr);
	      cairo_set_source_rgba(cr, line_r, line_g, line_b, alpha);
	      cairo_set_line_width(cr, 1);
	      for(double l=c_x1+2; l<c_x2; l+=4){
		cairo_move_to(cr, l, c_y1); cairo_line_to(cr, l, c_y2); }
	      cairo_stroke(cr);
	      line_pattern = cairo_pop_group(cr);
	    }
	  }
	}

	// set new clipping region (FIXME: this will be optional)
	if(status.rift_stencil && need_stencil){
	  cairo_move_to(cr, c_x1, c_y1);
	  cairo_line_to(cr, c_x2, c_y1);
	
	  for(unsigned si=0; si<stencil.size(); si+=2){
	    cairo_line_to(cr, c_x2, stencil[si]);
	    cairo_line_to(cr, c_x1, stencil[si]);
	    cairo_line_to(cr, c_x1, stencil[si+1]);
	    cairo_line_to(cr, c_x2, stencil[si+1]);
	  }

	  cairo_line_to(cr, c_x2, c_y2);
	  cairo_line_to(cr, c_x1, c_y2);
	  cairo_close_path(cr);
	  cairo_clip(cr);
	  cairo_new_path(cr);
	} else {
	  // must restore clip!
	  cairo_rectangle(cr, c_x1, c_y1, fabs(c_x2-c_x1), fabs(c_y2-c_y1));
	  cairo_clip(cr);
	  cairo_new_path(cr);
	}
      }
      
      // line width
      if (linewidth < status.max_thickness){
	// line is enough thin

	if (linewidth < status.min_thickness) 
	  linewidth=status.min_thickness;

	cairo_set_line_width(cr, linewidth);

	double fromX = seqx1 + meanA*basewidth;
	double toX = seqx1 + meanB*basewidth;

	// draw a line
	cairo_set_source_rgba(cr, line_r, line_g, line_b, alpha);

	if(upper_island_only){
	  cairo_move_to(cr, toX, toY);
	  cairo_line_to(cr, toX, toY_base);
	} 
	if(lower_island_only){
	  cairo_move_to(cr, fromX, fromY_base);
	  cairo_line_to(cr, fromX, fromY);
	}
	if(!upper_island_only && !lower_island_only){
	  cairo_move_to(cr, fromX, fromY_base);
	  cairo_line_to(cr, fromX, fromY);
	  cairo_line_to(cr, toX, toY);
	  cairo_line_to(cr, toX, toY_base);
	}
	cairo_stroke(cr);
      } else { 
	// line too thick: 
	cairo_set_line_width(cr, 1);

	if(upper_island_only || lower_island_only){
	  if(upper_island_only){
	    cairo_move_to(cr, seqx1 + seqB2*basewidth, toY);
	    cairo_line_to(cr, seqx1 + seqB2*basewidth, toY_base);
	    cairo_line_to(cr, seqx1 + seqB1*basewidth, toY_base);
	    cairo_line_to(cr, seqx1 + seqB1*basewidth, toY);
	  } else {
	    cairo_move_to(cr, seqx1 + seqA1*basewidth, fromY);
	    cairo_line_to(cr, seqx1 + seqA1*basewidth, fromY_base);
	    cairo_line_to(cr, seqx1 + seqA2*basewidth, fromY_base);
	    cairo_line_to(cr, seqx1 + seqA2*basewidth, fromY);
	  }
	  cairo_close_path(cr);
	  if (fill_pattern != NULL) cairo_set_source(cr, fill_pattern);
	  else cairo_set_source_rgba(cr, fill_r, fill_g, fill_b, alpha);
	  cairo_fill_preserve(cr);

	  if (line_pattern != NULL) cairo_set_source(cr, line_pattern);
	  else cairo_set_source_rgba(cr, line_r, line_g, line_b, alpha);
	  cairo_stroke(cr);
	} else {
	  for(int pass=0; pass<4; pass++){
	    double xa1 = seqx1 + seqA1*basewidth;
	    double xa2 = seqx1 + seqA2*basewidth;
	    double xb1 = seqx1 + seqB1*basewidth;
	    double xb2 = seqx1 + seqB2*basewidth;

	    switch(pass){
	    case 0:
	      cairo_move_to(cr, xa1, fromY); cairo_line_to(cr, xa1, fromY_base);
	      cairo_line_to(cr, xa2, fromY_base); cairo_line_to(cr, xa2, fromY);

	      cairo_line_to(cr, xb2, toY);     cairo_line_to(cr, xb2, toY_base);
	      cairo_line_to(cr, xb1, toY_base);     cairo_line_to(cr, xb1, toY);
	      break;
	      
	    case 1:
	      cairo_move_to(cr, xa1, fromY); cairo_line_to(cr, xa1, fromY_base);
	      cairo_line_to(cr, xa2, fromY_base); cairo_line_to(cr, xa2, fromY);
	      break;
	      
	    case 2:
	      cairo_line_to(cr, xb2, toY);   cairo_line_to(cr, xb2, toY_base);
	      cairo_line_to(cr, xb1, toY_base);   cairo_line_to(cr, xb1, toY); 
	      break;
	      
	    case 3:
	      cairo_move_to(cr, xa1, fromY);  cairo_line_to(cr, xa2, fromY);
	      cairo_line_to(cr, xb2, toY);    cairo_line_to(cr, xb1, toY); 
	      break;
	    }

	    cairo_close_path(cr);

	    switch(pass){
	    case 0: // stroke only
	      if (line_pattern != NULL) cairo_set_source(cr, line_pattern);
	      else cairo_set_source_rgba(cr, line_r, line_g, line_b, alpha);
	      cairo_stroke(cr);
	      break;

	    case 1: case 2: // fill islands
	      cairo_set_source_rgba(cr, fill_r, fill_g, fill_b, alpha);
	      cairo_fill(cr);
	      break;

	    case 3: // fill other region
	      if (fill_pattern != NULL) cairo_set_source(cr, fill_pattern);
	      else cairo_set_source_rgba(cr, fill_r, fill_g, fill_b, alpha);
	      cairo_fill(cr);
	      break;
	    }
	  }
	}

      } // end "else" for line too thick

      // cleanups
      if(need_stencil || upper_island_only || lower_island_only) cairo_restore(cr);
      if(fill_pattern != NULL) cairo_pattern_destroy(fill_pattern);
      if(line_pattern != NULL) cairo_pattern_destroy(line_pattern);
    } // end "for" for all cols
    
    if(all_in_range){ // remove visible ancs from justification list
      status.in_range_anchors.insert(i);
      if(status.dc_counter != 0) status.overwrap_ancs.erase(i);
    }
    if(itv.check(i)) pbar_set_fraction(thread_id, (double)(i-anc_from)/mr.arows());
  }

}

void draw_chrom_links_annotations(draw_chrom_links_anchors_t* params){
  cairo_t* cr = params->cr;
  double max_label_width = params->max_label_width;
  int margin_t = params->margin_t;
  double basewidth = params->basewidth;
  double seqdist = params->seqdist;
  unsigned numOfAnnotations = params->numOfAnnotations;
  int thread_id = params->thread_id;
  int threads = params->threads;

  double margin_l = status.margin_lr+max_label_width;
  
  cairo_rectangle(cr, margin_l, margin_t, 
		  width()-margin_l-status.margin_lr, 
		  height()-margin_t);
  cairo_clip(cr);
  cairo_new_path(cr);

  unsigned numOfAnnotated = 0;

  gene_markers_type thread_gene_markers;

  seqlen_t zf = status.plot_type==PLOT_SUPERZOOM ? status.superzoom_from : 0;
  seqlen_t zt = status.plot_type==PLOT_SUPERZOOM ? status.superzoom_to : mr.maxbases(status.toggle_seq);

  double seqx1 = status.margin_lr+max_label_width;

  int genewidth   = status.genewidth;
  int genepadding = status.genepadding;
  int genemargin  = status.genemargin;

  if( mr.has_hotspots() && status.show_hotspots ) genemargin += genewidth;

  for (unsigned int seq=0; seq<mr.nseqs(status.toggle_seq); seq++){
    unsigned tracks = mr.annotations[status.seq_lookup(seq)].size();
    if(!status.show_gene_marker || tracks==0) continue;

    // ------------------------------
    // assign bands on each track

    bool lastseq = (seq == mr.nseqs(status.toggle_seq)-1);
    unsigned bands = (int)((endcap(lastseq, lastseq, seqdist)-
			    genemargin-THICKNESS_SEQ/2) /
			   (genewidth+genepadding));
    if (bands==0) bands=1;

    vector <int> bandAlloc;
    bandAlloc.clear();

    // i want more sophisticated allocation policy, but I have no idea...
    if(bands<tracks) bands=tracks;

    for(unsigned int t=0; t<=tracks; t++)
      bandAlloc.push_back(t*(bands/tracks));

    // ------------------------------
    // track names
    if(thread_id == 0){
      for(unsigned int t=0; t<tracks; t++){
	assert(cout << "drawing seq(geom)=" << seq << " seq(original)=" << status.seq_lookup(seq) << " track=" << t << endl);
      
	double posY = margin_t + seqdist*seq;
	double posY_offset = genemargin - genepadding + THICKNESS_SEQ/2 +
	  (bandAlloc[t] + bandAlloc[t+1])*(genepadding+genewidth)/2+genewidth/2;
	if(!lastseq) posY += posY_offset;
	else         posY -= posY_offset;
	
	string tname = mr.annotations[status.seq_lookup(seq)].at(t)->get_name();

	if(tname.find_last_of("/") != string::npos)
	  tname = tname.substr(tname.find_last_of("/")+1);
	if(tname.find_last_of("\\") != string::npos)
	  tname = tname.substr(tname.find_last_of("\\")+1);

	cairo_set_source_rgb(cr, 0, 0, 0);

	draw_text(cr, FONT_CHR, status.font_size_gene, 'r', 'c', 
		  //	seqx1-7, posY, tname.c_str());
		  seqx1-7, posY, tname, status.max_width_gn);

	double trackWidth = ((bandAlloc[t+1]-bandAlloc[t])*
			     (genepadding+genewidth)
			     -genepadding)-2;

	cairo_set_line_width(cr, 0.5);
	cairo_move_to(cr, seqx1-3, posY-trackWidth/2);
	cairo_rel_line_to(cr, -2, 0);
	cairo_rel_line_to(cr, 0, trackWidth);
	cairo_rel_line_to(cr, 2, 0);
	cairo_stroke(cr);

	assert(cout << "track #" << t << " label Y=" 
	       << posY-trackWidth/2 << "-" << posY+trackWidth << endl);
      }
    }

    // ------------------------------
    // plot each track
    for(unsigned int annoTrack=0; annoTrack < tracks; annoTrack++){
      annotationTrack* thistrack = mr.annotations[status.seq_lookup(seq)].at(annoTrack);

      // parameters to draw score bars
      // (these ones for gene markers will be recalculated later)
      int nbands = bandAlloc[annoTrack+1]-bandAlloc[annoTrack];
      double bandwidth = nbands*genewidth + (nbands-1)*genepadding - 2; 

      double seqy  = margin_t+seqdist*seq;
      double seqy_offset = THICKNESS_SEQ/2+
	bandAlloc[annoTrack]*(genewidth+(genepadding))+genemargin+genewidth/2 + bandwidth/2;

      if(!lastseq) seqy += seqy_offset;
      else         seqy -= seqy_offset;

      double marker_top = seqy-bandwidth/2;
      double marker_bottom = seqy+bandwidth/2;

      assert(cout << "track #" << annoTrack << " marker Y=" 
	     << marker_top << "-" << marker_bottom << endl);

      double max = thistrack->maxScore();
      double min = thistrack->minScore();
      double scorePerPixel = (max-min) / (marker_bottom-marker_top);

      int bandbase = bandAlloc[annoTrack];
      int bands_thistrack = bandAlloc[annoTrack+1] - bandAlloc[annoTrack];

      // divide the job
      unsigned genes = mr.annotations[status.seq_lookup(seq)].at(annoTrack)->size();
      unsigned gene_per_thread = genes/threads;
      unsigned gene_from = thread_id * gene_per_thread;
      unsigned gene_to = ( thread_id == threads-1 ) ? genes :
	(thread_id+1) * gene_per_thread;

      for(unsigned int a=gene_from; a<gene_to; a++){
	annotation g = mr.rat(thistrack->at(a));

	// filter
	if(!mr.geneFilter(thistrack, a)) continue;

	// ok, draw it.
	int band = a % bands_thistrack + bandbase;

	seqlen_t f = g.begin + mr.get_offset_t(seq);
	seqlen_t t = g.end   + mr.get_offset_t(seq);

	// Rough clip
	if(zt < f || zf > t) continue;

	// clip
	f -= zf; t -= zf;
	if(f<0) f = 0;
	if(t>zt-zf) t=zt-zf;

	// show gene name or not
	bool show_gene_name=false;

	switch(status.gene_name_onoff_method){
	case DISP_NAME_ON:
	  show_gene_name = true; break;

	case DISP_NAME_OFF:
	  show_gene_name = false; break;

	case 0:
	case DISP_NAME_AUTO:
	default:
	  if(basewidth >= (double)1/(double)status.show_gene_name_threshold)
	    show_gene_name = true; 
	  else 
	    show_gene_name = false; 
	}

	// ------------------------------
	// draw it

	double marker_from = seqx1+basewidth*f;
	double marker_to   = seqx1+basewidth*t;

	if(!thistrack->hasScore()){
	  // usual gene markers (with annotations)

	  // recalculate params
	  seqy  = margin_t+seqdist*seq;
	  seqy_offset = THICKNESS_SEQ/2+
	    band*(genewidth+(genepadding))+genemargin+genewidth;
	  
	  if(!lastseq) seqy += seqy_offset;
	  else         seqy -= seqy_offset;

	  marker_top = seqy-genewidth/2;
	  marker_bottom = seqy+genewidth/2;

	  // and draw
	  int anno_col_base = 0;
	  if(status.anno_col_map.find(g.atype) != status.anno_col_map.end()){
	    anno_col_base = status.anno_col_map[g.atype];
	  } else anno_col_base = status.anno_col_map[""];

	  cairo_set_source_rgba(cr, 
				status.anno_col[anno_col_base].r(), 
				status.anno_col[anno_col_base].g(), 
				status.anno_col[anno_col_base].b(), 
				status.anno_col[anno_col_base].a());

	  double mf = marker_from;
	  double mt = marker_to;

	  size_t exons = g.segments.size();
	  size_t exon_count = 0;
	  if( !thistrack->is_hidden_feature(g.atype) ){
	    do { // run at least once; repeats if we have exons
	      if(exons != 0){
		if(exon_count == 0){ 
		  // center line for genes with exon/intron
		  cairo_set_line_width (cr, 2); 
		  cairo_move_to(cr, mf, seqy);
		  cairo_line_to(cr, mt, seqy);
		  cairo_stroke(cr);
		}

		mf = seqx1 + basewidth * ( g.segments[exon_count].begin
					   + mr.get_offset_t(seq) - zf);
		mt = seqx1 + basewidth * ( g.segments[exon_count].end
					   + mr.get_offset_t(seq) - zf);
	      
		if( thistrack->is_hidden_feature(g.segments[exon_count].type))
		  continue;
	      }

	      if(fabs(mt-mf) < genewidth/4){
		// no explicit direction for short genes
		cairo_set_line_width (cr, genewidth); 
		cairo_move_to(cr, mf, seqy);
		cairo_line_to(cr, mt, seqy);
		cairo_stroke(cr);
	      } else {
		// enough long
		cairo_set_line_width (cr, 0.1); 
		if(g.reverse){
		  // arrowhead on head: gene on reverse strand
		  cairo_move_to(cr, mf, seqy-genewidth/4);
		  cairo_line_to(cr, mf+genewidth/4, seqy-genewidth/2);
		  cairo_line_to(cr, mt, seqy-genewidth/2);
		  cairo_line_to(cr, mt, seqy+genewidth/2);
		  cairo_line_to(cr, mf+genewidth/4, seqy+genewidth/2);
		  cairo_line_to(cr, mf, seqy+genewidth/4);
		} else {
		  // arrowhead on tail: gene on forward strand
		  cairo_move_to(cr, mf, seqy-genewidth/2);
		  cairo_line_to(cr, mt-genewidth/4, seqy-genewidth/2);
		  cairo_line_to(cr, mt, seqy-genewidth/4);
		  cairo_line_to(cr, mt, seqy+genewidth/4);
		  cairo_line_to(cr, mt-genewidth/4, seqy+genewidth/2);
		  cairo_line_to(cr, mf, seqy+genewidth/2);
		}
		cairo_close_path(cr);
		cairo_fill(cr);
	      }
	    
	    } while(++exon_count < exons);

	    if(marker_to-marker_from < 1) marker_to += 1; // without this, thin 
	    // markers will be ignored
	    // by check_gene_marker_existence()

	    rectangle r = rectangle(marker_from, marker_top,
				    marker_to, marker_bottom);
	    annotationIndex ai = annotationIndex(status.seq_lookup(seq), 
						 annoTrack, a);

	    thread_gene_markers.insert(gene_marker_type(r, ai));

	    if(show_gene_name){
	      cairo_set_source_rgba(cr, 0, 0, 0, 0.8);
	      draw_text(cr, FONT_CHR, status.font_size_gene, 'l', 'c', 
			seqx1+basewidth*f, seqy, g.getname(status.name_by));
	    }
	  }
	} else {
	  // score
	  double score = g.score;

	  if(score < 0) cairo_set_source_rgba(cr, 1, 0, 0, 0.5);
	  else cairo_set_source_rgba(cr, 0, 0, 0, 0.5);

	  double marker_y = marker_bottom - (score-min)/scorePerPixel;
	  double marker_y2 = marker_y; // this will be 0 level in block style

	  if(thistrack->plotStyle() == 0){
	    // style is arrow

	    double lw = 1; 
	    if(thistrack->emphasizePlot()){
	      lw = 2;
	      marker_from -= ( (!g.reverse) ? 1 : -1 );
	      marker_to   += ( (!g.reverse) ? 1 : -1 );
	    }
	    
	    cairo_set_line_width(cr, lw);
	    cairo_move_to(cr, marker_from, marker_y);
	    cairo_line_to(cr, marker_to,   marker_y);
	    cairo_stroke(cr);
	    
	    // arrow head
	    if(marker_to - marker_from > 5){
	      if(!g.reverse){
		// fwd strand
		cairo_move_to(cr, marker_to,
			      marker_bottom - (score-min)/scorePerPixel);
		cairo_rel_line_to(cr, -2*lw, -lw);
		cairo_rel_line_to(cr, 0, lw);
	      } else {
		// bwd strand
		cairo_move_to(cr, marker_from, 
			      marker_bottom - (score-min)/scorePerPixel);
		cairo_rel_line_to(cr, 2*lw, -lw);
		cairo_rel_line_to(cr, 0, lw);
	      }
	      cairo_close_path(cr);
	      cairo_fill_preserve(cr);
	      cairo_stroke(cr);
	    }
	  } else {
	    // style is block
	    marker_y2 = marker_bottom - (0-min)/scorePerPixel;
	    
	    if( marker_to - marker_from > 5 &&
		fabs( marker_y - marker_y2 ) > 3 ){
	      // enough long to draw head
	      
	      double yoffset = ( score>0 ) ? 2 : -2;

	      if(!g.reverse){
		// fwd strand
		cairo_move_to(cr, marker_from, marker_y2);
		cairo_line_to(cr, marker_from, marker_y);
		cairo_line_to(cr, marker_to-2, marker_y);
		cairo_line_to(cr, marker_to,   marker_y + yoffset);
		cairo_line_to(cr, marker_to,   marker_y2);
	      } else {
		// bwd strand
		cairo_move_to(cr, marker_to, marker_y2);
		cairo_line_to(cr, marker_to, marker_y);
		cairo_line_to(cr, marker_from+2, marker_y);
		cairo_line_to(cr, marker_from, marker_y + yoffset);
		cairo_line_to(cr, marker_from, marker_y2);
	      }

	    } else {
	      // not enough long: just draw a box
	      cairo_move_to(cr, marker_from, marker_y);
	      cairo_line_to(cr, marker_to, marker_y);
	      cairo_line_to(cr, marker_to, marker_y2);
	      cairo_line_to(cr, marker_from, marker_y2);
	    }

	    cairo_close_path(cr);
	    cairo_fill_preserve(cr);
	    cairo_stroke(cr);
	  }

	  if(thistrack->showName()){
	    cairo_set_source_rgba(cr, 0, 0, 0, 0.8);
	    draw_text(cr, FONT_CHR, status.font_size_gene, 'l', 'b', 
		      marker_from, marker_y2, g.getname(status.name_by));
	  }

	  rectangle r = rectangle(marker_from, marker_y2-1,
				  marker_to, marker_y+1);
	  
	  annotationIndex ai = annotationIndex(status.seq_lookup(seq), 
					       annoTrack, a);
	  thread_gene_markers.insert(gene_marker_type(r, ai));

	}

	numOfAnnotated++;
	pbar_set_fraction(thread_id, (double)numOfAnnotated/numOfAnnotations );
	//	status.prog->fraction( (double)(i + (double)a/thistrack->size())/mr.nseqs(status.toggle_seq) );
      }
    }
  }

  g_mutex_lock(status.render_mutex);
  for(gene_markers_type::iterator i=thread_gene_markers.begin();
      i!=thread_gene_markers.end(); i++){
    status.gene_markers.insert(*i);
  }

  g_mutex_unlock(status.render_mutex);
}

// ----------------------------------------------------------------------

void draw_chrom_links(cairo_t *cr, void *useless_pointer)
{
  status.gene_markers.clear();
  status.in_range_anchors.clear();
  status.apply_defaults();

  if(status.plot_type==PLOT_CHROM_LINKS){ 
    status.superzoom_from = 0;
    status.superzoom_to = mr.maxbases(status.toggle_seq); }

  seqlen_t from = status.superzoom_from;
  seqlen_t to   = status.superzoom_to;

  //  assert(cout << "draw_chrom_links(): start with cr at: " << (void*)cr << endl);

  // find the longest name
  int max_label_width=0;
  for(unsigned int i=0; i<mr.nseqs(status.toggle_seq); i++){
    int w;
    if(status.disp_seqno) 
      w = text_width(cr, FONT_CHR, "#" + itos(status.seq_lookup(i)+1) + ": " 
		     + mr.getbs(i)->name);
    else 
      w = text_width(cr, FONT_CHR, mr.getbs(i)->name);

    if (max_label_width<w) max_label_width=w;
  }
  if (max_label_width > status.max_width_gn) max_label_width = status.max_width_gn;

  // extend top-margin for ticks label
  int margin_t = status.margin_tb + text_height(cr, FONT_CHR, "100Mb");
  int margin_b = status.margin_tb;

  // calculate size
  int seqwidth  = width()  - (2*status.margin_lr) - max_label_width;
  int seqheight = height() - (margin_t + margin_b);
  double basewidth = (double)seqwidth/mr.maxbases(status.toggle_seq);
  double seqdist   = (double)seqheight/(mr.nseqs(status.toggle_seq)-1);

  if (status.plot_type == PLOT_SUPERZOOM)
    basewidth = (double)seqwidth/(status.superzoom_to - status.superzoom_from);

  // put ticks
  //  assert(cout << "draw_chrom_links(): Plotting ticks..." << endl);

  seqlen_t ticks_interval = get_ticks_interval(mr.maxbases(status.toggle_seq));
  if(status.plot_type == PLOT_SUPERZOOM) 
    ticks_interval = get_ticks_interval(status.superzoom_to-status.superzoom_from);

  if (ticks_interval==0) ticks_interval=1;

  assert(cout << "ticks interval is " << ticks_interval << endl);

  seqlen_t prev_pos = 0;
  string prev_label = "";
  for(seqlen_t i=0; i<=mr.maxbases(status.toggle_seq)/ticks_interval; i++){
    if(status.plot_type == PLOT_SUPERZOOM &&
       ( i*ticks_interval < status.superzoom_from ||
	 status.superzoom_to < i*ticks_interval) ) continue;

    cairo_set_source_rgb(cr, 0.5, 0.5, 0.5);
    cairo_move_to(cr, 
		  status.margin_lr + max_label_width + basewidth*
		  (ticks_interval*i - status.superzoom_from),
		  margin_t - THICKNESS_SEQ/2 - TICKS_OFFSET);
    cairo_rel_line_to(cr, 0, seqdist*(mr.nseqs(status.toggle_seq)-1) + THICKNESS_SEQ*1.5 + TICKS_OFFSET );
    cairo_set_line_width(cr, THICKNESS_TICK);
    cairo_stroke(cr);

    cairo_set_source_rgb(cr, 0, 0, 0);
    draw_text(cr, FONT_CHR, 'c', 'B', 
	      status.margin_lr + max_label_width + basewidth*
	      (ticks_interval*i - status.superzoom_from), 
	      margin_t - THICKNESS_SEQ/2 - TICKS_OFFSET,
	      ticklabel(ticks_interval*i, prev_label, prev_pos).c_str());
  }

  // plot base range
  if(status.disp_range){
    seqlen_t maxbases = mr.maxbases(status.toggle_seq);
    seqlen_t from = status.superzoom_from;
    seqlen_t to   = status.superzoom_to;


    cairo_set_source_rgb(cr, 0, 0, 0);
    draw_text(cr, FONT_CHR, 'l', 't',
	      0, 0, string(itos_comma(status.superzoom_from)+"bp-"+
			   itos_comma(status.superzoom_to)+"bp of "+
			   itos_comma(maxbases)+"bp ("+
			   itos(((double)(to-from)*100/maxbases), 2)+"%)").c_str());
  }

  if(status.plot_type == PLOT_CHROM_LINKS){
    status.chrom_left = (int)(status.margin_lr+max_label_width);
    status.chrom_right = (int)(width()-status.margin_lr);
  }

  // set global basewidth
  status.basewidth = basewidth;
  assert(cout << "basewidth = " << basewidth << endl);

  // ------------------------------
  // set clickable zone

  status.clickable_y.clear();
  for(unsigned int i=0; i<mr.nseqs(status.toggle_seq); i+=1){
    bool lastseq  = (i == mr.nseqs(status.toggle_seq)-1);
    double seqy  = margin_t+seqdist*i;
    double endcap_from = endcap(lastseq, false, seqdist);
    double endcap_to   = endcap(lastseq, true,  seqdist);

    int from1 = (int)seqy;
    int to1   = (int)(from1+endcap_from);
    to1 += (int)(THICKNESS_SEQ/2);

    int to2   = (int)(seqy+seqdist);
    int from2 = (int)(to2-endcap_to);
    from2   -= (int)(THICKNESS_SEQ/2);

    status.clickable_y.push_back(pair<int,int>(from1, to1));
    status.clickable_y.push_back(pair<int,int>(from2, to2));
  }

  // ------------------------------
  // total # of annotations: prescan annotation tracks

  pbar_reset("Calculating anchor scores from expression data...");

  unsigned numOfAnnotations=0;

  vector<annotationTrack*>* filterTracks = &(mr.filterTracks);
  vector<filterTrack> filterTracksToCalc;

  filterTracks->clear();

  for (unsigned int i=0; i<mr.nseqs(status.toggle_seq); i++){
    for(unsigned int j=0; j<mr.annotations[status.seq_lookup(i)].size(); j++){
      annotationTrack* thistrack =  mr.annotations[status.seq_lookup(i)].at(j);
      numOfAnnotations += thistrack->size();

      assert(cout << "draw_chrom_links(): s" << i << " t" << j
	     << "(" << thistrack->get_name() << ")\t has: " 
	     << ( thistrack->isTraceable()  ? "traceability, " : "")
	     << ( thistrack->hasScore()     ? "score, " : "")
	     << ( thistrack->filterActive() ? "score filter " : "") 
	     << endl);

      if(thistrack->filterActive()){
	filterTracks->push_back(thistrack);
	if(!thistrack->hasAncScores()) filterTracksToCalc.push_back(filterTrack(thistrack,i));
      }
    }
  }

  for(unsigned i=0; i<filterTracksToCalc.size(); i++){
    pbar_set_task((double)i/filterTracksToCalc.size(),
		  (double)(i+1)/filterTracksToCalc.size());
    filterTracksToCalc[i].first->
      calcAncScores(filterTracksToCalc[i].second, pbar_set_fraction);
  }

  // ------------------------------
  pbar_reset("Calculating gene scores from expression data...");

  filterTracksToCalc.clear();
  for (unsigned int i=0; i<mr.nseqs(status.toggle_seq); i++){
    for(unsigned int j=0; j<mr.annotations[status.seq_lookup(i)].size(); j++){
      annotationTrack* thistrack =  mr.annotations[status.seq_lookup(i)].at(j);

      if(thistrack->filterActive()){
	if(!thistrack->checkAnnotExpScores(i)) 
	  filterTracksToCalc.push_back(filterTrack(thistrack,i));
      }
    }
  }

  for(unsigned i=0; i<filterTracksToCalc.size(); i++){
    pbar_set_task((double)i/filterTracksToCalc.size(),
		  (double)(i+1)/filterTracksToCalc.size());
    filterTracksToCalc[i].first->
      calcAnnotExpScores(filterTracksToCalc[i].second, pbar_set_fraction);
  }

  // ------------------------------

  pbar_reset("Applying anchors and their scores on annotation data...");

  filterTracksToCalc.clear();
  for (unsigned int i=0; i<mr.nseqs(status.toggle_seq); i++){
    for(unsigned int j=0; j<mr.annotations[status.seq_lookup(i)].size(); j++){
      annotationTrack* thistrack =  mr.annotations[status.seq_lookup(i)].at(j);
      if( (status.hide_anchored_gene || status.hide_unanchored_gene) &&
	  thistrack->isFilterable() )
	filterTracksToCalc.push_back(filterTrack(thistrack,i));
    }
  }

  for(unsigned i=0; i<filterTracksToCalc.size(); i++){
    pbar_set_task((double)i/filterTracksToCalc.size(),
		  (double)(i+1)/filterTracksToCalc.size());
    filterTracksToCalc[i].first->
      calcAnnotScores(filterTracksToCalc[i].second, pbar_set_fraction);
  }

  // ------------------------------
  // plot anchors

  assert(cout << "draw_chrom_links(): Plotting anchors..." << endl);

  // find the "base" sequence for coloring
  unsigned int colorbase = 0;

  // find the longest anchor
  seqlen_t maxlen_anchor=0;
  seqlen_t minlen_anchor=mr.maxbases_physical(status.toggle_seq);
  for (unsigned int a=0; a<mr.arows(); a++){
    seqlen_t l=abs(mr.raa_physical(a,colorbase*3+1)-
		   mr.raa_physical(a,colorbase*3));
    if(maxlen_anchor<l) maxlen_anchor=l;
    if(minlen_anchor>l) minlen_anchor=l;
  }
  if(minlen_anchor==maxlen_anchor) maxlen_anchor++;

  // ----------------------------------------------------------------------
  // multithreaded anchor / annotation plotting

  int threads = num_cpus(); // FIXME! 
  cairo_surface_t* surface;
  vector<cairo_surface_t*> surface_p;
  vector<cairo_t*> cr_p;
  draw_chrom_links_anchors_t* plotter_params = (draw_chrom_links_anchors_t*)malloc(sizeof(draw_chrom_links_anchors_t)*threads); // this mustn't be a vector, since it mustn't be rellocated 
  vector<GThread*> plotter_threads;

  for(unsigned i=0; i<2; i++){
    if(i==0) pbar_reset("Plotting anchors...");
    if(i==1) pbar_reset("Annotating...");
    pbar_set_threads(threads);

    surface_p.clear(); cr_p.clear(); 
    plotter_threads.clear();

    surface = cairo_get_target(cr);
    for(int t=0; t<threads; t++){
      surface_p.push_back(cairo_surface_create_similar(surface,
						       CAIRO_CONTENT_COLOR_ALPHA,
						       width(), height()));
      cr_p.push_back(cairo_create(surface_p[t]));
    }

    // prepare for threads
    for(int t=threads-1; t>=0; t--){

      // set clipping zone
      if(i==0){
	cairo_rectangle(cr_p[t], status.margin_lr+max_label_width, 
			margin_t, seqwidth, seqheight);
	cairo_clip(cr_p[t]);
	cairo_new_path(cr_p[t]);
      } else { // no clip for annotation
	cairo_reset_clip(cr_p[t]);
      }
    
      draw_chrom_links_anchors_t* p = &plotter_params[t];
      p->cr = cr_p[t];
      p->from = from;
      p->to = to;
      p->basewidth = basewidth;
      p->thread_id = t;
      p->threads = threads;
      p->colorbase = colorbase;
      p->margin_t = margin_t;
      p->max_label_width = max_label_width;
      p->seqdist = seqdist;
      p->minlen_anchor = minlen_anchor;
      p->maxlen_anchor = maxlen_anchor;
      p->numOfAnnotations = numOfAnnotations;

      if (t==0){
	// no thread creation for #0
	if(i==0) draw_chrom_links_anchors(p);
	else     draw_chrom_links_annotations(p);
	plotter_threads.push_back(NULL);
      } else {
	GThread* th;
	th = ( (i==0) ? 
	       g_thread_create((GThreadFunc)draw_chrom_links_anchors,
			       (gpointer)p, true, NULL) :
	       g_thread_create((GThreadFunc)draw_chrom_links_annotations,
			       (gpointer)p, true, NULL) );
	plotter_threads.push_back(th);
      }
    }

    // cleanup threads
    for(int t=0; t<threads; t++){
      cairo_pattern_t* pattern;
      cairo_matrix_t m;

      if(plotter_threads[t] != NULL) g_thread_join(plotter_threads[t]);
    
      pattern = cairo_pattern_create_for_surface(surface_p[t]);
      cairo_matrix_init_identity(&m);
      // m.x0 = 0; m.y0 = 0; // offset 0
      cairo_pattern_set_matrix(pattern, &m);

      cairo_set_source(cr, pattern);
      cairo_rectangle(cr, 0, 0, width(), height());
      cairo_fill(cr);

      cairo_pattern_destroy(pattern);
      cairo_surface_destroy(surface_p[t]);
      cairo_destroy(cr_p[t]);
    }
  }
  free(plotter_params); // will be done after annotation

  // plot chromosomes & sources
  if(status.plot_O) pbar_reset("Drawing Chromosomes & ORFs...");
  else pbar_reset("Drawing Chromosomes...");

  for(unsigned int i=0; i<mr.nseqs(status.toggle_seq); i++){
    assert(cout << "draw_chrom_links(): Plotting seq#" << endl);
    
    // variables for progress bar
    double fraction_seq  =  ((double)1/(mr.nseqs(status.toggle_seq)));
    double fraction_zero =  fraction_seq * i;
    
    double seqy  = margin_t+seqdist*i;
    double seqx1 = status.margin_lr+max_label_width;
    double seqx2 = seqx1+seqwidth; 
    
    // name
    cairo_set_source_rgb(cr, 0, 0, 0);

    string chr_name;
    if(status.disp_seqno) chr_name=("#" + itos(status.seq_lookup(i)+1)+ ": ");
    chr_name = chr_name + mr.getbs(i)->name;
    draw_text(cr, FONT_CHR, 'r','c', seqx1-status.margin_gn, seqy, 
	      chr_name.c_str(), status.max_width_gn);

    pbar_set_task(fraction_zero, fraction_zero + fraction_seq);
    disp_chrom(cr, seqx1, seqx2, seqy, i, basewidth, 
	       ticks_interval, pbar_set_fraction);

    disp_sources(cr, i, basewidth, 
		 status.margin_lr+max_label_width, margin_t+seqdist*i, 0, 0);
  }

  // plot repeats
  if( mr.has_hotspots() && status.show_hotspots ){
    pbar_reset("Plotting repeats...");

    double x1 = status.margin_lr + max_label_width;
    double x2 = width()-status.margin_lr;
    seqlen_t pixrange = x2-x1;
    vector<double> repcount(pixrange);

    int method = status.hotspot_by;
    if(method==0) method=HOTSPOT_BY_DENSITY;

    for(unsigned s=0; s<mr.nseqs(status.toggle_seq); s++){
      pbar_set_task((double)s/mr.nseqs(status.toggle_seq),
		    (double)(s+1)/mr.nseqs(status.toggle_seq));
      for(int i=0; i<pixrange; i++) repcount[i]=0;

      bool lastseq = (s == mr.nseqs(status.toggle_seq)-1);
      unsigned ss = status.seq_lookup(s);
      double y  = margin_t+seqdist*s;

      double rep_width = status.genewidth/2;
      double true_rep_len = mr.hotspot_pattern_len() * basewidth;
      double rep_len = true_rep_len;
      if (rep_len < rep_width) rep_len = rep_width;

      y += (lastseq ? -1 : 1) * (status.genemargin + THICKNESS_SEQ/2 + rep_width*2 - status.genepadding);
      for(unsigned r=0; r<mr.hotspot_patterns(); r++){
	multimap<int, seqlen_t>::iterator i;

	// color
	double hue, red=1, green=0, blue=0;
	pair <int, int> valrange; // max, min
	switch(method){
	case HOTSPOT_BY_PATTERN: 
	  hue = 300 * (double)r / mr.hotspot_patterns(); break;
	case HOTSPOT_BY_NUMBER:
	  valrange = mr.hotspot_range_by_pattern(r);
	  hue = 300 * ((double)(mr.get_hotspot_by_pattern(r) 
				- valrange.first) / 
		       (valrange.second - valrange.first) ); break;
	case HOTSPOT_BY_COMBINATION:
	  valrange = mr.hotspot_range_by_combination(r);
	  hue = 300 * ( (double)(mr.get_hotspot_by_combination(r) 
				 - valrange.first) / 
			(valrange.second - valrange.first) ); break;
	default: // red
	  hue = 0; break;
	};
	HSVtoRGB(&red, &green, &blue, hue, 0.9, 0.9);
	cairo_set_source_rgb(cr, red, green, blue);
	cairo_set_line_width(cr, 1);

	// draw (or just count for density plot)
	for(hspotIte i=mr.hotspot_pos_lookup(r, ss).first;
	    i!=mr.hotspot_pos_lookup(r, ss).second; i++){
	  pair<bool, seqlen_t> hspot = mr.get_hotspot_pos(ss, i);

	  bool reverse = hspot.first;
	  seqlen_t pos = hspot.second - status.superzoom_from;
	  if(pos < 0 || pos > status.superzoom_to-status.superzoom_from) 
	    continue;

	  double xoffset = (reverse ? -1 : 1) * (rep_len - rep_width);
	  double yoffset = (reverse ? -1 : 1) * rep_width;

	  double x = pos*basewidth;
	  if(method != HOTSPOT_BY_DENSITY &&
	     method != HOTSPOT_BY_DENSITY_LOG){
	    x += status.margin_lr + max_label_width;
	    cairo_move_to(cr, x, y);
	    cairo_rel_line_to(cr,  xoffset, 0);
	    cairo_rel_line_to(cr,  yoffset, -yoffset/2);
	    cairo_rel_line_to(cr, -yoffset, -yoffset/2);
	    cairo_rel_line_to(cr, -xoffset, 0);
	    cairo_close_path(cr);
	    cairo_fill(cr);
	  } else { // HOSPOT_BY_DENSITY{,_LOG}: just count
	    int xx= (!reverse) ? (int)x : (int)(x-true_rep_len) ;
	    int xx_cnt = 0;
	    do {
	      repcount[xx + xx_cnt]+=1;
	      xx_cnt++;
	    } while(xx_cnt < true_rep_len);
	  } 
	}
	pbar_set_fraction((double)r/mr.hotspot_patterns());
      }

      if(method == HOTSPOT_BY_DENSITY ||
	 method == HOTSPOT_BY_DENSITY_LOG){

	// find max
	double min_rep=0, max_rep=0;
	for(vector<double>::iterator repite=repcount.begin(); 
	    repite!=repcount.end(); repite++){

	  if(method==HOTSPOT_BY_DENSITY_LOG) *repite=log10(*repite);
	  if(repite==repcount.begin()){ 
	    max_rep = *repite;
	    min_rep = *repite;
	  } else {
	    if( max_rep < *repite ) max_rep = *repite;
	    if( *repite < min_rep ) min_rep = *repite;
	  }
	}
	if(max_rep==min_rep) max_rep+=1;
	double amplitude = max_rep - min_rep;

	// draw them
	double y  = margin_t+seqdist*s;
	bool lastseq = (s == mr.nseqs(status.toggle_seq)-1);
	y += (lastseq ? -1 : 1) * (status.genemargin + THICKNESS_SEQ/2 + status.genewidth - status.genepadding);
	y += status.genewidth/2;

	for(int x=0; x<pixrange; x++){
	  if(repcount[x] != 0){
	    cairo_set_line_width(cr, 1);
	    cairo_set_source_rgb(cr, 1, 0, 0);
	    cairo_move_to(cr, x1 + x, y);
	    cairo_rel_line_to(cr, 0, -((repcount[x]-min_rep) * status.genewidth/amplitude));
	    cairo_stroke(cr);
	  }
	}
      }
    }
    pbar_set_task(0, 1);
  }
  assert(cout << "draw_chrom_links(): done with plotting!" << endl);
}

