
// Original code written by Takanori Maehara:
//    http://www.prefield.com/algorithm/container/segment_tree.html

using namespace std;

#ifndef _SEGMENT_TREE_H_
#define _SEGMENT_TREE_H_

#include <set>
#include <vector>
#include <algorithm>
#include <math.h>

#include "gmv_defs.h"
#include "util.h"

//#define seqlen_t int
//#include <iostream>

/* 
  usage: 
    step 1: insert intervals by insert(from, to, content);
    step 2: build tree by build_tree();
    step 3: query point by query(point, res); or range by query(from, to, res);
            you'll get results in set<content> res;
 */


template <typename T> class segment_tree {
  typedef seqlen_t position;
  //  typedef T content;

public:
  struct interval {
    position low, high;
    interval(position low, position high) : low(low), high(high) { };
  };

  template <typename TT> class posListItem {
  public:
    position l;
    TT c;
    posListItem(position l, TT c) : l(l), c(c) { };
    friend bool operator<(const posListItem<TT> a, const posListItem<TT> b){ 
      return a.l < b.l; };
    friend bool operator==(const posListItem<TT> a, const posListItem<TT> b){ 
      return a.l == b.l; };
  };

  void insert(position a, position b, T c){
    position A=abs(a);
    position B=abs(b);
    if(A>B) swap(A, B);

    ilist.push_back(pair<interval,T>(interval(A,B+1), c));
    lpos.push_back(posListItem<T>(A,c));
    pos.push_back(A);
    pos.push_back(B+1);
  };
  
  void build_tree(void(*setfrac)(double)){ 
    sort(pos.begin(), pos.end());
    if(setfrac!=NULL) setfrac(0.222);

    vector<position>::iterator end_pos = unique(pos.begin(), pos.end());
    pos.erase(end_pos, pos.end());
    if(setfrac!=NULL) setfrac(0.444);
    
    sort(lpos.begin(), lpos.end()); // don't apply unique on this!
    if(setfrac!=NULL) setfrac(0.666);

    root = build_tree(0, pos.size()-1);

    size_t cnt=0;
    for(typename vector<pair<interval,T > >::iterator i=ilist.begin();
      i!=ilist.end(); i++){
      insert(root, i->first, i->second);
      if(setfrac!=NULL) setfrac(cnt/(ilist.size()*3)+0.666);
    }
    pos.clear();
    ilist.clear();
  };

  void build_tree(){
    build_tree(NULL);
  };

  void query(position p, set<T> &out) const { query(root, p, out); };
  
  void query(position p, position q, set<T> &out) const {
    position P=abs(p);
    position Q=abs(q);
    if(P>Q) swap(P, Q);
    query(root, P, out); 
    query(root, Q, out); 

    typename vector<posListItem<T> >::const_iterator lo =
      lower_bound(lpos.begin(), lpos.end(), posListItem<T>(P, 0));
    typename vector<posListItem<T> >::const_iterator hi = 
      lower_bound(lpos.begin(), lpos.end(), posListItem<T>(Q, 0));
    for(typename vector<posListItem<T> >::const_iterator i=lo; i!=hi; i++)
      out.insert(i->c);
  };

  void clear(){
    pos.clear();
    ilist.clear();
    lpos.clear();
  }

private:
  vector< position > pos;
  vector< pair<interval,T > > ilist; // list of intervals 
  vector< posListItem<T> > lpos; // list of left ends

  struct node {
    vector<T> values;
    position B, E, M;
    node *left, *right;
  } *root;

  node *build_tree(position i, position j) {
    position m = (i+j)/2;
    node *t = new node;
    t->B = pos[i]; t->E = pos[j]; t->M = pos[m];
    t->left  = (i+1 < j ? build_tree(i, m) : NULL);
    t->right = (i+1 < j ? build_tree(m, j) : NULL);

    return t;
  }

  void insert(node *v, const interval& I, T c) {
    if (I.low <= v->B && v->E <= I.high) {
      v->values.push_back( c );
    } else {
      if (I.low  < v->M) insert(v->left , I, c);
      if (I.high > v->M) insert(v->right, I, c);
    }
  }

  void query(node *t, position p, set<T> &out) const {
    if (!t || p < t->B || p >= t->E) return;

    for(typename vector<T>::iterator i=t->values.begin(); 
  	i!=t->values.end(); i++) out.insert(*i);
    if (p < t->M) query(t->left, p, out);
    else          query(t->right, p, out);
  }
};

/* for test
int main(int argc, char* argv[]){
  segment_tree<int> t;

  t.insert(0, 7, 1);  
  t.insert(0, 5, 2);  
  t.insert(5, 15, 3); 
  t.insert(11, 12, 4);
  t.insert(20, 25, 5);

  t.build_tree();

  set<int> result;

  if(argc==2)
    t.query(atoi(argv[1]), result);
  else
    t.query(atoi(argv[1]), atoi(argv[2]), result);

  for(set<int>::iterator i=result.begin(); i!=result.end(); i++)
    cout << *i << endl;
  
  return 0;
}
*/

#endif

