#include <assert.h>

#include "text.h"
#include "gmv.h"
#include "gmv_base.h"
#include "gmv_defs.h"
#include "util.h"
#include <pango/pangocairo.h>

using namespace std;

#define WITH_PANGO   // pango/cairo seems to be a little bit buggy... 

// ----------------------------------------------------------------------
//  Text handling

pair<int,int> text_size(cairo_t *cr, const char *font, string text){
  return text_size(cr, font, FONT_SIZE, text);
}

pair<int,int> text_size(cairo_t *cr, const char *font, int fsize, string text){

#ifdef WITH_PANGO
  PangoLayout *textlayout;
  PangoFontDescription *desc;

  lock_gdk();
  textlayout = pango_cairo_create_layout (cr);
  
  pango_layout_set_text(textlayout, text.c_str(), -1);
  desc = pango_font_description_from_string(font);
  pango_font_description_set_size(desc, fsize*PANGO_SCALE);
  pango_layout_set_font_description(textlayout, desc);

  int width, height;
  pango_layout_get_size (textlayout, &width, &height);

  pango_font_description_free (desc);
  g_object_unref(textlayout);
  unlock_gdk();
  
  return pair<int,int>(width/PANGO_SCALE, height/PANGO_SCALE);

#else

  cairo_text_extents_t extents;

  lock_gdk();
  cairo_select_font_face(cr, font, CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_BOLD);
  cairo_set_font_size(cr, fsize);
  cairo_text_extents(cr, text.c_str(), &extents);

  int x = (int)extents.x_advance;
  int y = (int)extents.height;
  unlock_gdk();

  return pair<int,int>(x,y);
#endif
}


int text_width(cairo_t *cr, const char *font, string text){
  return text_width(cr, font, FONT_SIZE, text);
}

int text_width(cairo_t *cr, const char *font, int fsize, string text){
  return text_size(cr, font, fsize, text).first;
}

int text_height(cairo_t *cr, const char *font, string text){
  return text_height(cr, font, FONT_SIZE, text);
}

int text_height(cairo_t *cr, const char *font, int fsize, string text){
  return text_size(cr, font, fsize, text).second;
}

void draw_text(cairo_t *cr,  const char *font, char hpos, char vpos,
	       double x, double y, string text){
  draw_text(cr, font, FONT_SIZE, hpos, vpos, x, y, text, 0);
}

void draw_text(cairo_t *cr,  const char *font, char hpos, char vpos,
	       double x, double y, string text, int maxsize){
  draw_text(cr, font, FONT_SIZE, hpos, vpos, x, y, text, maxsize);
}

void draw_text(cairo_t *cr, const char *font, int size, char hpos, char vpos, 
	       double x, double y, string text){
  draw_text(cr, font, size, hpos, vpos, x, y, text, 0);
}

void draw_text(cairo_t *cr, const char *font, int size, char hpos, char vpos, 
	       double x, double y, string text, int maxsize){
  lock_gdk();

#ifdef WITH_PANGO
  PangoLayout *textlayout;
  PangoFontDescription *desc;

  textlayout = pango_cairo_create_layout (cr);
  
  pango_layout_set_text(textlayout, text.c_str(), -1);
  desc = pango_font_description_from_string (font);
  pango_font_description_set_size(desc, size*PANGO_SCALE*0.8);
  pango_font_description_set_weight(desc, PANGO_WEIGHT_NORMAL);
  pango_layout_set_font_description (textlayout, desc);

  int width, height;
  pango_layout_get_size(textlayout, &width, &height);

  if(maxsize != 0 && width > maxsize*PANGO_SCALE){
    pango_layout_set_ellipsize(textlayout, PANGO_ELLIPSIZE_END);
    pango_layout_set_width(textlayout, maxsize*PANGO_SCALE);
    width = maxsize*PANGO_SCALE;
  }

  double px, py;
  px = x;
  py = y;

  switch(hpos){
  case 'r': px -= (double)(width)/PANGO_SCALE; break;
  case 'c': px -= ((double)width/2)/PANGO_SCALE;
  }
  
  PangoFontMetrics* fm;
  switch(vpos){
  case 'b': 
    py -= (double)height/PANGO_SCALE; 
    break;
  case 'B': // baseline at Y
    fm =  pango_context_get_metrics(pango_layout_get_context(textlayout),
				    desc, pango_language_from_string("en"));
    py -= (double)(height - pango_font_metrics_get_descent(fm))/PANGO_SCALE;
    pango_font_metrics_unref(fm);
    break; 
  case 'c':
    py -= ((double)height/2)/PANGO_SCALE;
    break;
  }

  cairo_move_to(cr, px, py);

  pango_cairo_show_layout (cr, textlayout);
    
  pango_font_description_free(desc);
  g_object_unref(textlayout);

#else

  cairo_text_extents_t extents;

  cairo_select_font_face(cr, font, CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_BOLD);
  cairo_set_font_size(cr, size);
  cairo_text_extents(cr, text.c_str(), &extents);

  double width  = extents.x_advance;
  double height = extents.height;

  double px, py;
  px = x;
  py = y;

  switch(hpos){
  case 'r': px -= width; break;
  case 'c': px -= width/2;
  }
  
  switch(vpos){
  case 'B':
  case 'b': break;
  case 't': py += height; break;
  case 'c': py += height/2;
  }

  cairo_move_to(cr, px, py);
  cairo_show_text(cr, text.c_str());

#endif
  unlock_gdk();
}

string ticklabel(seqlen_t b, string& ps, seqlen_t& pb){
  string label = ticklabel(b);
  if (label == ps) label = string("+") + ticklabel(b-pb);
  else ps = label;
  pb = b;
    
  return label;
}

string ticklabel(seqlen_t b){
  ostringstream os;
  os << "";

  if(b < 1000)
    os << b << "b";
  else {
    if(b < 1000*1000)
      os << (double)b/(1000) << "kb";
    else {
      if(b < 1000*1000*1000)
	os << (double)b/(1000*1000) << "Mb";
      else
	os << (double)b/1000/1000/1000 << "Gb";
    }
  }
  return os.str();
}

seqlen_t get_ticks_interval(seqlen_t seql){
  seqlen_t i;
  for(i=10; i<=1000000000; i*=10)
    if (seql < i) break;
  
  i /= 10;
  if (seql < i*2) return i/5;
  if (seql < i*4) return i*2/5;
  if (seql < i*5) return i/2;
  return i;
}
